#include "ExplorationParam.h"
#include "Types.h"

char ExplorationParam::getOptionChar( Option option ){
  if( option == PERFORM_EXPLORATION )
    return 'E';

  if( option == PRIME )
    return 'P';

  if( option == LIK )
    return 'L';

  if( option == INPUT_REC )
    return 'H';

  if( option == POST_PROB )
    return 'Q';

  if( option == CHECK_STATE_SET )
    return 'C';

  if( option == MAX_DEPTH )
    return 'M';

  if( option == OUTPUT_STEP )
    return 'S';

  if( option == COMPUTE_MAPPING_PROB )
    return 'D';
}

string ExplorationParam::getOptionValue( string optionString , Option option ){
  string S;
  uint pos = optionString.find( getOptionChar( option ) );

  pos++;

  // Skip spaces
  while( isspace( optionString[ pos ] )  )
    pos++;

  // Retrieve value
  while( !isspace( optionString[ pos ] )  ){
    S += optionString[ pos ];
    pos++;
  }

  return S;
}

void ExplorationParam::initialize( const string& speciesTreeFile , const string& edgeValuesFile , const string& optionString , const string& geneTreeString , const string& realRecString ){
  _reconciliation = 0;
  
  // MaxDepth: default value is Diameter
  // Defined in ExplorationAlgo::Constructor
  _maxDepth = lastUint;
  _outputStep = 100;


  _realRecString = realRecString;


  // Species tree
  string speciesTree = readFirstLine( speciesTreeFile );
  _paramSpeciesTree.initialize( speciesTree );

  _paramGeneTree.initialize( _paramSpeciesTree , geneTreeString );

  // Birth Death Param
  _birthDeathParam.initialize( _paramSpeciesTree , edgeValuesFile , true );

  // Option
  _performExploration = charInString( optionString , getOptionChar( PERFORM_EXPLORATION ) );
  _prime              = charInString( optionString , getOptionChar( PRIME ) );
  _lik                = charInString( optionString , getOptionChar( LIK ) );
  _computeProb        = charInString( optionString , getOptionChar( POST_PROB ) );
  _checkStateSet      = charInString( optionString , getOptionChar( CHECK_STATE_SET ) );
  _computeMappingProb  = charInString( optionString , getOptionChar( COMPUTE_MAPPING_PROB ) );

  // Max depth
  if( charInString( optionString , getOptionChar( MAX_DEPTH ) ) )
    _maxDepth = stoi( getOptionValue( optionString , MAX_DEPTH ) );

  // Output steps
  if( charInString( optionString , getOptionChar( OUTPUT_STEP ) ) )
    _outputStep = stoi( getOptionValue( optionString , OUTPUT_STEP ) );

  // Define NB_OF_VALUATIONS (see TreePair:Types.h)
  if( _lik && !_computeProb )
    Types::define_nbOfValuations( 4 );

  if( _lik && _computeProb )
    Types::define_nbOfValuations( 5 );


  // Define Number of valuations constants
  // Temporary Exception
  Exception( Types::nbOfValuations() == 4 && !_lik , "ExplorationParam::ExplorationParam: NB_OF_VALUATIONS == 4 <----> likelihood option" );
}

ExplorationParam::ExplorationParam( const string& speciesTreeFile , const string& geneTreeFile , const string& edgeValuesFile , const string& optionString ):
  _resultsStream( resultsFileName().c_str() , std::ios::out ){

  _inputRec       = charInString( optionString , getOptionChar( INPUT_REC ) );

  // Gene tree
  string geneTreeString = readFirstLine( geneTreeFile );
  string realRecString;

  // Read the reconciliation in tree string format
  if( _inputRec ){
    ifstream geneTreeStream( geneTreeFile.c_str() , std::ios::in );
    while( !geneTreeStream.eof() ){
      string buffer;
      getline( geneTreeStream , buffer );
      if( !buffer.empty() && buffer[0] == '[' ){
	realRecString = buffer;
	break;
      }
    }
    geneTreeStream.close();
    Exception( realRecString.empty() , "ExplorationParam::ExplorationParam ---> invalid RandomGeneTree" );
  }

  initialize( speciesTreeFile , edgeValuesFile , optionString , geneTreeString , realRecString );
}

ExplorationParam::ExplorationParam( const string& speciesTreeFile , const string& edgeValuesFile , const string& optionString , const string& geneTreeString , const string& realRecString ):
  _resultsStream( resultsFileName().c_str() , std::ios::out ){

  _inputRec = charInString( optionString , getOptionChar( INPUT_REC ) );
  initialize( speciesTreeFile , edgeValuesFile , optionString , geneTreeString , realRecString );
}

ExplorationParam::~ExplorationParam(){ 
  _resultsStream.close(); 
  if( _reconciliation != 0 )
    delete _reconciliation;
}

string ExplorationParam::help(){
  string s = "\n\nOptions:";
  s += "\n" + ctos( getOptionChar( PRIME ) )                + "\t: Output reconciliation to primetv format";
  s += "\n" + ctos( getOptionChar( PERFORM_EXPLORATION ) )  + "\t: perform explorations";
  s += "\n" + ctos( getOptionChar( LIK ) )                  + "\t: compute likelihood";
  s += "\n" + ctos( getOptionChar( INPUT_REC ) )            + "\t: read and consider true reconciliation";
  s += "\n" + ctos( getOptionChar( POST_PROB ) )            + "\t: compute posterior probabilities";
  s += "\n" + ctos( getOptionChar( CHECK_STATE_SET ) )      + "\t: Check that the enumeration algorithm is consistant with the space set";
  s += "\n" + ctos( getOptionChar( MAX_DEPTH ) )            + "\t: Maximal depth";
  s += "\n" + ctos( getOptionChar( OUTPUT_STEP ) )          + "\t: Print to logfile each n steps";
  s += "\n" + ctos( getOptionChar( COMPUTE_MAPPING_PROB ) ) + "\t: Compute the prob for each mapping rec(u) = x";

  return s;
}


std::ostream& operator << (std::ostream& output, const ExplorationParam& explorationParam ){
  output<< "\nParam Species tree";
  output<< explorationParam.paramSpeciesTree();

  output<< "\nParam  Gene tree";
  output<< explorationParam.paramGeneTree();

  output<< "\nBirth and Death";
  output<< explorationParam.birthDeathParam();

  output<< "\nReal Rec String\n";
  output<< explorationParam.realRecString();

  return output;
}
