#include "ExplorationStat.h"
#include "Reconciliation.h"
#include "Types.h"
#include "ExplorationParam.h"

ExplorationStat::ExplorationStat( const TreePair& treePair ,  const ExplorationParam& explorationParam ):
  _treePair( treePair ) , _localCounts(Types::nbOfValuations(),0) , _explorationParam( explorationParam ),
  _minimaListVec( Types::nbOfValuations() ) , _recNameVector( Types::nbOfRecName() , 0 ),
  _likelihoodStat( treePair ){

  // Define RecName Vector
  _recNameVector[ MinRec ]  = const_cast<Reconciliation*> ( &( _treePair.extremeRec( Lowest ) ) );

  if( explorationParam.inputRec() )
    _recNameVector[ RealRec ] = const_cast<Reconciliation*> ( explorationParam.reconciliation() );
  
  for( uint val = 0; val < Types::nbOfValuations(); val++ )
    _bestValues.setWorstValues();

  _distribNb.resize( Types::nbOfValuations() );
  _distribNMC.resize( Types::nbOfValuations() );
  if( reconciliationNameIsDefined( RealRec ) )
    _distribDiffMapping.resize( Types::nbOfValuations() );

  _nbOfRec = 0;

  // Resize variance array
  _nmcDistDeviation_NbRec.resize( treePair.diameter()/2 + 1 , 0 );
  _sumDistDeviation = 0;

  // Resize nmcDist arrays
  _nmcDist_NbRec.resize( Types::nbOfRecName() );

  for( int i = 0; i < Types::nbOfRecName(); i++ ){
    _nmcDist_NbRec[i].resize( treePair.diameter() + 1 , 0 );
  }

  // Initialize Likelihood stat
  _likelihoodStat.initialize( explorationParam , explorationParam.lik() , this );
}

ExplorationStat::~ExplorationStat(){
  for( uint val = 0; val < Types::nbOfValuations(); val++ )
    for( list<Reconciliation*>::iterator rec = _minimaListVec[ val ].begin(); rec != _minimaListVec[ val ].end(); rec++ )
      delete *rec;

  if( (uint) (LikRec) < Types::nbOfRecName() )
    delete _recNameVector[ LikRec ];
}

void ExplorationStat::updateDistrib( vector< Map_Double_Uint >& distrib , uint val , uint value , const Reconciliation& reconciliation ){
  Map_Double_Uint& dist = distrib[ val ];

  double cost = reconciliation.costValuation( val );
  Map_Double_Uint::iterator itPair = dist.find( cost );
  if( itPair == dist.end() )
    dist.insert( pair< double, uint>( cost , value ) );
  else
    itPair->second += value;  
}

void ExplorationStat::updateLikRec( const Reconciliation& reconciliation ){
  if( _recNameVector[ LikRec ] == 0 ){
    _recNameVector[ LikRec ] = new Reconciliation( _treePair , reconciliation.recVector() );
    _recNameVector[ LikRec ]->valuationValuesUpdate().setValue( Lik , reconciliation.costValuation( Lik ) );
  }
  else{
    if( reconciliation.costValuation( Lik ) > _recNameVector[ LikRec ]->costValuation( Lik ) ){
      delete _recNameVector[ LikRec ];
      _recNameVector[ LikRec ] = new Reconciliation( _treePair , reconciliation.recVector() );
      _recNameVector[ LikRec ]->valuationValuesUpdate().setValue( Lik , reconciliation.costValuation( Lik ) );
    }
  }
}

void ExplorationStat::update( const Reconciliation& reconciliation , bool withMinimaList ){

  _nbOfRec++;

  _likelihoodStat.update( reconciliation );

  // Update NMC Distrib
  for( uint val = 0; val < Types::nbOfValuations(); val++ ){
    uint distnmc = computeNMCDist_WithNearestMinima( reconciliation , withMinimaList , val );
    updateDistrib( _distribNMC , val , distnmc , reconciliation );
  }


  // If minima list has already been computed, it is the same for the others distrib 
  if( withMinimaList )
    return;

  // Update Diff and Counts distributions
  for( uint val = 0; val < Types::nbOfValuations(); val++ ){
    updateDistrib( _distribNb  , val , 1       , reconciliation );
    if( reconciliationNameIsDefined( RealRec ) ){
      uint diffMapping = nbOfDiffMapping( reconciliation , * ( reconciliationName( RealRec ) ) );
      updateDistrib( _distribDiffMapping, val , diffMapping , reconciliation );
    }
  }


  // Update Global counts
  for( uint val = 0; val < Types::nbOfValuations(); val++ ){
    double cost = reconciliation.costValuation( val );
    if( _bestValues.hasWorstValue( val , cost ) ){
      _bestValues.setValue( val , cost );
      // Update Minima list
      for( list<Reconciliation*>::iterator rec = _minimaListVec[ val ].begin(); rec != _minimaListVec[ val ].end(); rec++ )
	delete *rec;
      _minimaListVec[ val ].clear();
      _minimaListVec[ val ].push_back( new Reconciliation( reconciliation ) );
    }
    else
      if( cost == _bestValues.value( val ) )
	_minimaListVec[ val ].push_back( new Reconciliation( reconciliation ) );
  }


  // Update Local counts
  if( reconciliation.localMinComputed() )
    for( uint val = 0; val < Types::nbOfValuations(); val++ )
      if( reconciliation.isLocalMin( val ) )
	_localCounts[ val ]++;

  ///////////////////////////////////
    // Arrays of size 3 x Diam
    // For all REC_NAME
    for( int i = 0; i < Types::nbOfRecName(); i++ ){
      uint nmcDist = reconciliation.valuationValues().nmcDist( (REC_NAME) (i) );

      // Variance distribution
      if( (REC_NAME) (i) == MinRec ){
	double deviation = fabs( (double) ( _treePair.diameter()/2 ) - nmcDist );
	// Causes an error "munmap_chunk" (May 4, 2009)
	//	_nmcDistDeviation_NbRec[ (int) ( floor(deviation) ) ]++;
	_sumDistDeviation += deviation;	
      }

      // Update nmc distances 
      _nmcDist_NbRec[ i ] [ nmcDist ]++;
    }
}


void ExplorationStat::clearDistribNMC(){
  for( uint val = 0; val < Types::nbOfValuations(); val++ )
    _distribNMC[ val ].clear();
}

bool ExplorationStat::moreThanOneMinima() const{
  for( uint val = 0; val < Types::nbOfValuations(); val++ )
    if( _minimaListVec[ val ].size() > 1 )
      return true;
}


uint ExplorationStat::computeNMCDist_WithNearestMinima( const Reconciliation& reconciliation , bool withMinimaList , uint val ) const{
  if( withMinimaList ){
    Exception( _minimaListVec[ val ].empty() , "ExplorationStat::computeNMCDist_WithNearestMinima ---> minima list is empty" );
    uint minDist = lastUint;
    for( list<Reconciliation*>::const_iterator rec = _minimaListVec[ val ].begin(); rec != _minimaListVec[ val ].end(); rec++ ){
      uint dist = distNMC( reconciliation , **rec );
      if( dist < minDist )
	minDist = dist;
    }
    return minDist;
  }
  else
    return distNMC( reconciliation , _treePair.extremeRec( Lowest ) );
}

void printDistributions_Format1(std::ostream& output , const ExplorationStat& explorationStat ){
  for( uint val = 0; val < Types::nbOfValuations(); val++ ){
    output<<"\n\n\n" << valuationToString( val ) <<":";

    const Map_Double_Uint& distNb   = explorationStat.distribNb()[val];
    const Map_Double_Uint& distDiff = explorationStat.distribDiffMapping()[val];
    const Map_Double_Uint& distNMC  = explorationStat.distribNMC()[val];

    // Print cost
    for( Map_Double_Uint::const_iterator itPair = distNb.begin(); itPair != distNb.end(); itPair++ )
      output <<"\t"<< dtos( itPair->first );

    output<<"\n\n#rec:";
    for( Map_Double_Uint::const_iterator itPair = distNb.begin(); itPair != distNb.end(); itPair++ )
      output <<"\t"<< itPair->second;

    output<<"\nNMC:";
    for( Map_Double_Uint::const_iterator itPairNMC = distNMC.begin(); itPairNMC != distNMC.end(); itPairNMC++ )
      output <<"\t"<< (double)(itPairNMC->second);

    if( explorationStat.reconciliationNameIsDefined( RealRec ) ){
      output<<"\nS.Dif:";
      for( Map_Double_Uint::const_iterator itPairDiff = distDiff.begin(); itPairDiff != distDiff.end(); itPairDiff++ )
	output <<"\t"<< itPairDiff->second;
    }
  }
}

void printDistributions_Format2(std::ostream& output , const ExplorationStat& explorationStat ){
  for( uint val = 0; val < Types::nbOfValuations(); val++ ){
    output<<"\n\n\n" <<valuationToString( val ) <<"\t#rec\tNMC\tS.Dif";

    const Map_Double_Uint& distNb   = explorationStat.distribNb()[val];
    const Map_Double_Uint& distDiff = explorationStat.distribDiffMapping()[val];
    const Map_Double_Uint& distNMC  = explorationStat.distribNMC()[val];

    for( Map_Double_Uint::const_iterator itPair = distNb.begin(); itPair != distNb.end(); itPair++ ){
      const Map_Double_Uint::const_iterator itDiff = distDiff.find( itPair->first );
      const Map_Double_Uint::const_iterator itNMC =  distNMC.find( itPair->first );
      Exception( itDiff  == distDiff.end() , "ExplorationStat::printDistributions_Format2 ---> anormal case A" );
      Exception( itNMC   == distNMC.end()   , "ExplorationStat::printDistributions_Format2 ---> anormal case B" );
      
      // Print cost , #rec , NMC , Diff
      output<<endl << dtos( itPair->first , -1 );
      output<<"\t" << itPair->second;
      output<<"\t" << itNMC->second;
      output<<"\t" << itDiff->second;
    }
  }
}

void ExplorationStat::printNMCDist_Distrib( std::ostream& output ) const{
  for( uint i = 0; i < Types::nbOfRecName(); i++ ){
    output<<"\n\n" << recNameToString( i );
    output<<"\nDist\t#Rec";
    for( uint d = 0; d <= _treePair.diameter(); d++ ){
      output<<"\n" << d <<"\t";
      output<< _nmcDist_NbRec[ i ] [ d ];
    }
  }
}

void ExplorationStat::print_NMCDist_Var_Distrib( std::ostream& output ) const{
  output<<"\n\nDeviation = | diam/2 - d_nmc(MinRec , rec) |";
//   output<<"\nDev.\t#Rec";
//   for( uint d = 0; d <= _treePair.diameter()/2; d++ ){
//     output<<"\n" << d <<"\t";
//     output<< _nmcDistDeviation_NbRec[ d ];
//   }
  output <<"\nSum of deviation: " << _sumDistDeviation;
}

void ExplorationStat::printProbDistrib( std::ostream& output ) const{
//   // NMC dist -----> Prob. sum
//   for( uint recName = 0; recName < Types::nbOfRecName(); recName++ ){
//     output<<"\n\n" << recNameToString( recName );
//     output<<"\nDist\t#Prob Sum";
//     for( uint dist = 0; dist <= _treePair.diameter(); dist++ ){
//       output<<"\n" << dist <<"\t";
//       output<< _nmcDist_Prob[ recName ] [ dist ];
//     }
//   }

//   // NMC dist -----> Mass
//   for( uint recName = 0; recName < Types::nbOfRecName(); recName++ ){
//     output<<"\n\n" << recNameToString( recName );
//     output<<"\nDist\t#Mass";
//     for( uint dist = 0; dist <= _treePair.diameter(); dist++ ){
//       double mass = 0;
//       for( uint d = 0; d <= dist; d++ )
// 	mass += _nmcDist_Prob[ recName ] [ d ];
	
// 	output<<"\n" << dist <<"\t";
// 	output<< mass;
//     }
//   }
}

void ExplorationStat::printNMCDist_RecName ( std::ostream& output ) const{
  output<<"\n\nDistances\t\tNMC \t# Diff. Mapping";
  for( uint i = 0; i < Types::nbOfRecName() - 1; i++ )
    for( uint j = i + 1; j < Types::nbOfRecName(); j++ ){
      
      if( !reconciliationNameIsDefined( i ) || !reconciliationNameIsDefined( j ) )
	continue;

      output<<"\n" << recNameToString( i ) << " , " << recNameToString( j ) << " : ";
      output<<"\t" << distNMC         ( *( reconciliationName( i ) ) , *( reconciliationName( j ) ) );
      output<<"\t" << nbOfDiffMapping ( *( reconciliationName( i ) ) , *( reconciliationName( j ) ) );
    }
  
}

void ExplorationStat::printNMCDist_Minima( std::ostream& output ) const{
  output<<"\n\nNMC dist between minimas";
  for( uint val = 0; val < Types::nbOfValuations(); val++ ){
    if( _minimaListVec[ val ].size() > 1 ){
      output<<"\n" <<valuationToString( val );
      for( list<Reconciliation*>::const_iterator rec1 = _minimaListVec[ val ].begin(); rec1 != --( _minimaListVec[ val ].end() ); rec1++ )
	for( list<Reconciliation*>::const_iterator rec2 = (++rec1)--; rec2 != _minimaListVec[ val ].end(); rec2++ )
	  output<<endl << distNMC( **rec1 , **rec2 );
    }
  }
}

void ExplorationStat::printNMCDist_Real( std::ostream& output ) const{
  if( !reconciliationNameIsDefined( RealRec ) )
    return;

  output<<"\n\nNMC dist between minima and Real";
  for( uint val = 0; val < Types::nbOfValuations(); val++ ){
    output<<"\n" <<valuationToString( val );
    for( list<Reconciliation*>::const_iterator rec = _minimaListVec[ val ].begin(); rec != _minimaListVec[ val ].end(); rec++ )
      output<<endl << distNMC( **rec , *( reconciliationName( RealRec ) ) );
  }
}

std::ostream& operator << (std::ostream& output , const ExplorationStat& explorationStat ){
  output << "\nExploration statisics\n";

  if( explorationStat.explorationParam().lik() )
    output << explorationStat.likelihoodStat();
 
//   for( uint val = 0; val < Types::nbOfValuations(); val++ )
//     output<<"\t"<< valuationToString( val );

  // Global counts
//   output <<"\nGlobal";
//   for( uint val = 0; val < Types::nbOfValuations(); val++ )
//     output<<"\t"<< explorationStat.globalCounts( val );

  // Local counts
//   output <<"\nLocal";
//   for( uint val = 0; val < Types::nbOfValuations(); val++ )
//     output<<"\t"<< explorationStat.localCounts( val );

  explorationStat.printNMCDist_Distrib( output );

  //  output<<"\n\nDistribution: Cost -----> # of rec, mean differences, dist nmc";

  /////////////////////////////////
  // TEMPORARY
  /////////////////////////////////
  if( explorationStat.reconciliationNameIsDefined( RealRec ) )
    printDistributions_Format2( output , explorationStat );

  if( explorationStat.explorationParam().computeProb() )
    explorationStat.printProbDistrib( output );

  //  explorationStat.printNMCDist_Minima( output );
  explorationStat.printNMCDist_Real( output );

  explorationStat.printNMCDist_RecName( output );

  output <<"\n\nTotal number of rec: " << explorationStat.nbOfRec();
  output <<"\n";

  explorationStat.print_NMCDist_Var_Distrib( output );

  return output;
}



void ExplorationStat::outputMinimas( string fileName ) const{
  ofstream output ( fileName.c_str() , std::ios::out );

  for( uint val = 0; val < Types::nbOfValuations(); val++ )
    if( _minimaListVec[ val ].size() > 1 ){
      output<<"\n"<< valuationToString( val );
      for( list<Reconciliation*>::const_iterator rec = _minimaListVec[ val ].begin(); rec != _minimaListVec[ val ].end(); rec++ )
	output<<endl << **rec;
    }

  output.close();
}
