/***************************************************************************
			ExplorationStat.h
		-------------------
	begin	 : Tue May 13 11:06:03 EDT 2008
	copyright: (C) 2005 by Universite De Montreal
	email	 :  doyonjea@iro.umontreal.ca
***************************************************************************/
#ifndef EXPLORATIONSTAT_H
#define EXPLORATIONSTAT_H

//! This class defines the base structure for a ExplorationStat object.
/**
*@author Jean-Philippe Doyon
*/

#include "StdClass.h"

class Reconciliation;
class ExplorationParam;

#include "ValuationValues.h"
#include "Types.h"
#include "TreePair.h"
#include "LikelihoodStat.h"


/////////////////////// Remarks 
// Global minimum and minima are synonymus

typedef map< double, uint > Map_Double_Uint;


class ExplorationStat{

  friend std::ostream& operator << (std::ostream& output, const ExplorationStat& explorationStat );

 private:
  // Count the number of local/global min for each (4) cost valuation
  vector<uint> _localCounts;

  const ExplorationParam& _explorationParam;

  const TreePair& _treePair;

  ValuationValues _bestValues;

  LikelihoodStat _likelihoodStat;

  // Vector of size NB_OF_REC_NAME
  vector< Reconciliation* > _recNameVector;

  ///////////////////////////////////
  // Distribution: key ---> value
  // For each criteria: some distributions

  // Distrib: cost ----> nb of reconciliations 
  vector< Map_Double_Uint > _distribNb;

  // R : any rec
  // R*: minima rec
  // Distrib: cost ----> sum of distNMC( R , R* )
  vector< Map_Double_Uint > _distribNMC;

  // Distrib: cost ----> sum of # of diff. mapping (with recmin)
  vector< Map_Double_Uint > _distribDiffMapping;

  // For each criteria: the list of the global optimal
  // First complete exploration: compute the list 
  // Second one: used in the case there is more than one minima
  vector< list<Reconciliation*> > _minimaListVec;


  // Total number of reconciliations visited
  uint _nbOfRec;

  // Distrib: 
  // 0 <= k <= diam/2 -----> # of rec s.t. | diam/2 - d_nmc(MinRec , rec) | = k
  vector< uint > _nmcDistDeviation_NbRec;

  // Sum_{rec} | diam/2 - d_nmc(MinRec , rec) |
  double _sumDistDeviation;

  ///////////////////////////////////
  // Arrays of size 3 x (Diam + 1)
  // For each rec' \in (MinRec , RealRec , LikRec)
  //    0 <= k <= diam -----> # of rec s.t. d_nmc(rec' , rec) = k
  vector< vector<uint> > _nmcDist_NbRec;


 public:

  // Constructor
  ExplorationStat( const TreePair& treePair , const ExplorationParam& explorationParam );
    
  // Destructor
  ~ExplorationStat();

  // Access methods

  const ExplorationParam& explorationParam() const { return _explorationParam; }
  const LikelihoodStat&  likelihoodStat() const { return _likelihoodStat; }

  const vector< Map_Double_Uint >& distribNb() const { return _distribNb; }
  const vector< Map_Double_Uint >& distribDiffMapping() const { return _distribDiffMapping; }
  const vector< Map_Double_Uint >& distribNMC() const { return _distribNMC; }

  const Reconciliation* reconciliationName( uint recName ) const { 
    Exception( !reconciliationNameIsDefined( recName ) , "ExplorationStat::reconciliationName ---> " + recNameToString( recName ) + " is not defined" );
    return _recNameVector[ recName ]; 
  }

  const TreePair& treePair() const { return _treePair; }

  bool reconciliationNameIsDefined( uint recName ) const{
    return _recNameVector[ recName ] != 0;
  }

  uint   nbOfRec() const { return _nbOfRec; }

  uint localCounts ( uint val ) const { return _localCounts[ val ]; }
  uint globalCounts( uint val ) const { return _minimaListVec[ val ].size(); }

  const vector< list<Reconciliation*> >& minimaListVec() const { return _minimaListVec; }

  //////////////////////////////
  // Update methods

  // Update likelihood summation
  void updateLikSum( const Reconciliation& reconciliation ){
    _likelihoodStat.updateLikSum( reconciliation );    
  }

  // Update the reconciliation with the highest lik
  void updateLikRec( const Reconciliation& reconciliation );
  
  // Update statistic distributions
  void update( const Reconciliation& reconciliation , bool withMinimaList );



  void clearDistribNMC();

  // If #minima > 1 for at least one criteria
  bool moreThanOneMinima() const;

  void outputMinimas( string fileName ) const;

  
    

 private:
  void updateDistrib( vector< Map_Double_Uint >& distrib , uint val , uint value , const Reconciliation& reconciliation );
  
  uint computeNMCDist_WithNearestMinima( const Reconciliation& reconciliation , bool withMinimaList , uint val ) const;

  void printNMCDist_Distrib( std::ostream& output ) const;
  void print_NMCDist_Var_Distrib( std::ostream& output ) const;

  void printProbDistrib    ( std::ostream& output ) const;

  // The nmc dist between minRec, realRec, likRec
  void printNMCDist_RecName ( std::ostream& output ) const;

  void printNMCDist_Minima ( std::ostream& output ) const;
  void printNMCDist_Real   ( std::ostream& output ) const;

};

#endif
