#include "LikelihoodStat.h"
#include "Types.h"
#include "ExplorationStat.h"
#include "ExplorationParam.h"
#include "Cell.h"

void LikelihoodStat::initialize( const ExplorationParam& explorationParam , bool active , ExplorationStat* explorationStat ){
  _active = active;
  _likSum = 0;
  _explorationStat = explorationStat;

  if( _active ){
    _distribDeviation.resize( 101 , 0 );

    _postProbDist.resize( 101 , 0 );

    _minRecLik = 0;
    _higherCount = 0;

    _nmcDist_Prob.resize     ( Types::nbOfRecName() );
    _nmcDist_ProbMass.resize ( Types::nbOfRecName() );

    for( int i = 0; i < Types::nbOfRecName(); i++ ){
      _nmcDist_Prob[i].resize ( _treePair.diameter() + 1 , 0 );
      _nmcDist_ProbMass[i].resize ( _treePair.diameter() + 1 , 0 );
    }
  }

  if( reconciliationNameIsDefined( RealRec ) ){
    // True Positive / False Positive vector
    _truePositiveDist.resize( geneTree().nbOfClades() , 0 );
    _truePositiveProb.resize( geneTree().nbOfClades() , 0 );

    _falsePositiveDist.resize( geneTree().nbOfClades() , 0 );
    _falsePositiveProb.resize( geneTree().nbOfClades() , 0 );
  }

  // Define MappingProb matrix
  if( explorationParam.computeMappingProb() )
    _mappingProb.resize( _treePair.geneTree().nbOfNodes() , vector< double >( _treePair.speciesTree().nbOfNodes() + _treePair.speciesTree().nbOfEdges() , 0 ) );
}

bool LikelihoodStat::reconciliationNameIsDefined( uint recName ) const{
  return _explorationStat->reconciliationNameIsDefined( recName );
}

const Reconciliation* LikelihoodStat::reconciliationName( uint recName ) const{
  return _explorationStat->reconciliationName( recName );
}

const GeneTree& LikelihoodStat::geneTree() const{
  return _explorationStat->treePair().geneTree();
}

void LikelihoodStat::updateMappingProb( const Reconciliation& reconciliation ){
  double postProb = reconciliation.costValuation( PostProb );

  const vector<Cell*>& mappingVector = reconciliation.mappingVector();

  for( uint i = 0; i < mappingVector.size(); i++ ){
    uint j = mappingVector[i]->id();
    _mappingProb[ i ][ j ] += postProb;
  }
}

void LikelihoodStat::update( const Reconciliation& reconciliation ){
  if( !_active )
    return;

  double lik = reconciliation.costValuation( Lik );

  if( _likSum != 0 ){
    double postProb = reconciliation.costValuation( PostProb );
    _postProbDist[ (int) (postProb * 100) ]++;

    for( int i = 0; i < Types::nbOfRecName(); i++ ){
      uint nmcDist = reconciliation.valuationValues().nmcDist( (REC_NAME) (i) );    

      // Update nmcDist ---> Prob
      _nmcDist_Prob[ i ] [ nmcDist ] += postProb;

      // Update nmcDist ---> Prob mass
      for( int k = nmcDist;  k < _nmcDist_ProbMass[ i ].size(); k++ )
	_nmcDist_ProbMass[ i ] [ k ] += postProb;
    }
    
    if( !_mappingProb.empty() )
      updateMappingProb( reconciliation );
  }

  if( reconciliationNameIsDefined( RealRec ) ){
    // True Positive / False Positive vector
    uint truePos  = nbOfTruePositive_Dup ( reconciliation , *( reconciliationName( RealRec ) ) );
    uint falsePos = nbOfFalsePositive_Dup( reconciliation , *( reconciliationName( RealRec ) ) );
    _truePositiveDist [ truePos  ]++;
    _falsePositiveDist[ falsePos ]++;

    if( _likSum != 0 ){
      double postProb = reconciliation.costValuation( PostProb );
      _truePositiveProb [ truePos  ] += postProb;
      _falsePositiveProb[ falsePos ] += postProb;
    }
  }


  // Update lik list
  //  _likList.push_back( lik );

  // Define minRec
  if( _minRecLik == 0 )
    _minRecLik = lik;

  // Update stat
  if( lik > _minRecLik ){
    _higherCount++;
  }
  else{
    double deviation = ( _minRecLik - lik ) / _minRecLik * 100;

    for( int i = 1; i < 101; i++ ){
      if( i - 1 < deviation && deviation <= i ){
	_distribDeviation[ i ]++;
	break;
      }
    }
  }
}

void LikelihoodStat::printMappingProb( std::ostream& output ) const{

  output<<"\n\nPosterior Probability mapping";
  output<<"\nGenes\tSpecies\tProb\t\tDist to LCA";

  for( int i = 0; i < _mappingProb.size(); i++ ){
    const Node* gene = _treePair.geneTree().getNode( i );
    if( gene->isLeafNode() )
      continue;

    const Cell* lcaCell = _treePair.getCell( _treePair.mappingFunction( gene ) );

    // Internal node
    for( int j = 0; j < _mappingProb[ i ].size(); j++ ){
      const Cell* cell = _treePair.getCell( j );
      if( *cell <= *lcaCell ){
	double prob = _mappingProb[ i ][ j ];
	if( prob > 0 ){
	  uint dist = ( lcaCell->isARootCell()? 0: distCell( cell , lcaCell ) );
	  output<<"\n" << i << "\t" << cell->toString() << "\t" << prob <<"\t" << dist;
	}
      }
    }
    output<<"\n";
  }
}

std::ostream& operator << (std::ostream& output, const LikelihoodStat& likStat ){

  output<<"\nDeviation from minRec likelihood";
  output<<"\nNb of rec s.t. lik(rec) > lik(minrec): " << likStat.higherCount();
  output<<"\nLikelihood of minRec: " << likStat.minRecLik();
  output<<"\nDev\tCount";
  const vector< uint >& devVector = likStat.distribDeviation();
  for( uint i = 0; i < devVector.size(); i++ ){
    if( devVector[ i ] != 0 ){
      output<<"\n" << i << "\t" << devVector[ i ];
    }
  }
  output<<"\n\n";

  output<<"\n\nAll likelihoods";
  for( list< double >::const_iterator it = likStat.likList().begin(); it != likStat.likList().end(); it++ )
    output<<"\n"<<*it;

  // Posterior Probabilities distrib
  output <<"\n\nPosterior probabilities distribution";
  output <<"\nProb\t#Rec";
  output << distribToString( likStat.postProbDist() , 1 );
  output<<"\n\n";

  // NMC dist ---> sum Post. Prob
  for( int i = 0; i < Types::nbOfRecName(); i++ ){
    output<<"\n\n" << recNameToString( i );
    output <<"\nNMC / Sum Post Prob";
    output << vectorToString( likStat.nmcDist_Prob()[ i ] , 0 , 5 );
  }

  // NMC dist ---> Prob. mass
  for( int i = 0; i < Types::nbOfRecName(); i++ ){
    output<<"\n\n" << recNameToString( i );
    output <<"\nNMC / Prob Mass";
    output << vectorToString( likStat.nmcDist_ProbMass()[ i ] , 0 , 5 );
  }

  // True Positive
  output<< "\n\nTrue/False Positive Duplications";

  if( likStat.reconciliationNameIsDefined( RealRec ) )
    output<< "\nNumber of real dup: "      << likStat.reconciliationName( RealRec )->nbOfDuplications();

  output<< "\nNumber of internal nodes:" << likStat.geneTree().nbOfClades();

  output<<"\n\n#TP\t#Rec";
  output<< distribToString( likStat.truePositiveDist() , 1 );
  output<<"\n\n#TP\tSum. Prob";
  output<< vectorToString( likStat.truePositiveProb() , 0 , 5 );

  output<<"\n\n#FP\t#Rec";
  output<< distribToString( likStat.falsePositiveDist() , 1 );
  output<<"\n\n#FP\tSum. Prob";
  output<< vectorToString( likStat.falsePositiveProb() , 0 , 5 );

  if( !likStat.mappingProb().empty() )
    likStat.printMappingProb( output );

  return output;
}

