/***************************************************************************
			LikelihoodStat.h
		-------------------
	begin	 : Thu May 14 18:27:39 EDT 2009
	copyright: (C) 2005 by Universite De Montreal
	email	 :  doyonjea@iro.umontreal.ca
***************************************************************************/
#ifndef LIKELIHOODSTAT_H
#define LIKELIHOODSTAT_H

//! This class defines the base structure for a LikelihoodStat object.
/**
*@author Jean-Philippe Doyon
*/

#include "Reconciliation.h"
#include "TreePair.h"

class ExplorationStat;
class ExplorationParam;

class LikelihoodStat{

  friend std::ostream& operator << (std::ostream& output, const LikelihoodStat& likStat );

 private:

  const TreePair& _treePair;

  // The likelihood of reconciliation minRec 
  double _minRecLik;

  ExplorationStat* _explorationStat;

  // Distribution:
  //    From 0 <= i <= 100
  //       i ---->  Number of reconciliations rec s.t. i < ( lik(minrec) - lik(rec) ) / lik(minrec) * 100<= i + 1
  // Array of size 101
  vector< uint > _distribDeviation;

  // Number of rec s.t. lik(rec) > lik(minRec)
  uint _higherCount;

  // Active <----> likelihood valuation is computed
  bool _active;

  // List of all likelihoods computed
  list< double > _likList;

  // The sumation of the likelihood of all reconciliations
  // Used to compute posterior probabilities
  double _likSum;

  ///////////////////////////////////
  // Arrays of size 3 x (Diam + 1)
  // For each rec' \in (MinRec , RealRec , LikRec)
  // P(rec) is the post. prob.
  //      0 <= k <= diam -----> sum_{ rec s.t. d_nmc(rec' , rec) = k } P(rec)
  vector< vector<double> > _nmcDist_Prob;
  //      0 <= k <= diam -----> sum_{ rec s.t. d_nmc(rec' , rec) <= k } P(rec)
  vector< vector<double> > _nmcDist_ProbMass;

  // Distribution:
  //       For p in {0, 1, 2, ..., 100}
  //       p ----> nb of rec s.t. int( P(rec) * 100 ) = p
  vector< uint > _postProbDist;

  // True Positive / False Positive Duplications
  // Let D (S) be the number of dup (spec) in recReal
  //      For 0 <= k <= D
  //         k ---> nb of rec s.t. TP( rec) = k
  //         k ---> sum_{ rec s.t. TP(rec) = k } P(rec)
  // True Positive
  vector< uint > _truePositiveDist;
  vector< double > _truePositiveProb;

  // False Positive
  vector< uint > _falsePositiveDist;
  vector< double > _falsePositiveProb;

  // The Post Prob for each mapping
  // k = |V(G)|, m = |V(S)|, n = |E(S)|
  // M: matrix of size = k (m + n)
  // Given a vertex u of G and a cell x of S
  // i: nodeId( u )
  // j: cellId( x ) (see TreePair.h)
  // M[i,j] = Sum_{rec in RecSpace s.t. rec(u) = x} P(rec,G)
  vector< vector< double > > _mappingProb;


 public:

  LikelihoodStat( const TreePair& treePair ):_treePair( treePair ){}

  // Access methods

  double minRecLik() const { return _minRecLik; }

  const vector< uint >& distribDeviation() const { return _distribDeviation; }

  const vector< vector<double> >& nmcDist_Prob() const { return _nmcDist_Prob; }
  const vector< vector<double> >& nmcDist_ProbMass() const { return _nmcDist_ProbMass; }

  // True/False positive
  const vector< uint >&   truePositiveDist()  const { return _truePositiveDist; }
  const vector< double >& truePositiveProb()  const { return _truePositiveProb; }
  const vector< uint >  & falsePositiveDist() const { return _falsePositiveDist; }
  const vector< double >& falsePositiveProb() const { return _falsePositiveProb; }

  const vector< vector< double > >& mappingProb() const { return _mappingProb; }

  uint higherCount() const { return _higherCount; }

  // Shortcut methods
  bool reconciliationNameIsDefined( uint recName ) const;
  const Reconciliation* reconciliationName( uint recName ) const;
  const GeneTree& geneTree() const;


  double likSum()  const { return _likSum; }

  const list< double >& likList() const { return _likList; }

  const vector< uint >& postProbDist() const { return _postProbDist; }

  void updateLikSum( const Reconciliation& reconciliation ){
    _likSum += reconciliation.costValuation( Lik );
  }

  // Others methods
  void initialize( const ExplorationParam& explorationParam , bool active , ExplorationStat* _explorationStat );
  void update( const Reconciliation& reconciliation );

  ~LikelihoodStat(){}

 private:

  void updateMappingProb( const Reconciliation& reconciliation );
  void printMappingProb( std::ostream& output ) const;

};

#endif
