#include "TreeStringSet.h"
#include "toolBox.h"
#include "StringLayer.h"
#include "OBJECT_SET_CPUtimes.h"

int TreeStringSet::_classCounter = 0;

TreeStringSet::~TreeStringSet(){
  _classCounter--;
  deleteAll();
  //  clear();
}

void TreeStringSet::initialize( int upperClassId ){
  // The layer size (except for the root one)
  _upperClassId   = upperClassId;
  _layerSize      = _upperClassId == NULL_CLASS_ID? 0: _upperClassId + 5; 

  // The first layer need an extra space for the empty string
  _rootLayer       = new StringLayer( _layerSize + 1 , true , 0 , 0 );

  _layerList.push_back( _rootLayer );

  for( uint id = 0; id <= upperClassId; id++ )
    _availableStringId.push_back( id );
}

void TreeStringSet::clear(){
  _upperClassId = NULL_CLASS_ID;
  _layerSize = 0;
  objectSetClear();
}

const StringInfo* TreeStringSet::addTreeString( const string& treeString ){
  try{
    OBJECT_SET_CPUtimes::cpuTimes( ADD_TREE_STRING ).startClock();
    Exception( !stringIsValid( treeString ) , "TreeStringSet::addTreeString ---> invalid string: " + treeString );

    vector<uint> posVector;
    stringToPosVector( treeString , posVector );
    uint pos = posVector.back();
    StringLayer* layer = const_cast<StringLayer*> ( (const StringLayer*) ( retrieveLeafLayer( posVector ) ) );

    // String notin Set: create StringInfo with count = 0
    if( !layer->objExists( pos ) ){
      Exception( _availableStringId.empty() , "TreeStringSet::addTreeString ----> no more available string id" );
      const StringInfo* stringInfo = layer->createStringInfo( pos , treeString );
      stringInfoList().push_back( stringInfo );
    }

    StringInfo* stringInfo = const_cast<StringInfo*> ( layer->getObjInfo( pos ) );

    // If count = 0, define StringId
    if( stringInfo->count() == 0 ){
      stringInfo->setStringId( _availableStringId.front() );
      _availableStringId.pop_front();
    }

    // Increment counter
    stringInfo->incCounter();

    OBJECT_SET_CPUtimes::cpuTimes( ADD_TREE_STRING ).stopClock();
    return stringInfo;
  }

  catch( SubException subExc ){
    subExc.addFunDesc( "TreeStringSet::addTreeString" , treeString );
    throw subExc;
  }
}

void TreeStringSet::removeStringInfo( StringInfo* stringInfo ){

  stringInfo->decCounter();

  if( stringInfo->count() == 0 ){
    _availableStringId.push_front( stringInfo->stringId() );

    Layer* layer = const_cast<Layer*> ( stringInfo->father() );
    layer->deleteAbstractObj( stringInfo );

    while( !layer->isRoot() && layer->count() == 0 ){
      Layer* father = const_cast<Layer*> ( layer->father() );
      father->deleteAbstractObj( layer );
      layer = father;
    }
  }
}

uint TreeStringSet::removeTreeString( const string& treeString ){
  Exception( !stringIsValid( treeString )                  , "TreeStringSet::removeTreeString ---> invalid string: " + treeString );
  Exception( !objExists( stringToPosVector( treeString ) ) , "TreeStringSet::removeTreeString ---> treeString doesn't exist: " + treeString );

  vector<uint> posVector;
  stringToPosVector( treeString , posVector );
  uint pos = posVector.back();
  StringLayer* layer = const_cast<StringLayer*> ( (const StringLayer*) ( retrieveLeafLayer( posVector ) ) );

  StringInfo* stringInfo = const_cast<StringInfo*> ( layer->getObjInfo( pos ) );
  uint stringId = stringInfo->stringId();

  stringInfo->decCounter();
     
  if( stringInfo->count() == 0 ){
    layer->deleteStringInfo( pos );
    _availableStringId.push_front( stringId );
  }

  return stringId;
}

void TreeStringSet::decStringInfoCounter( StringInfo* stringInfo ){
  // Decrement counter
  stringInfo->decCounter();

  // If count = 0, insert StringId in the available ids list
  if( stringInfo->count() == 0 )
    _availableStringId.push_front( stringInfo->stringId() );
}

const StringInfo& TreeStringSet::getStringInfo( const string& treeString ) const{
  Exception( !stringIsValid( treeString )                  , "TreeStringSet::getStringInfo ---> invalid string: " + treeString );
  Exception( !objExists( stringToPosVector( treeString ) ) , "TreeStringSet::getStringInfo ---> treeString doesn't exist: " + treeString );  

  vector<uint> posVector;
  stringToPosVector( treeString , posVector );
  return ( const StringInfo& ) ( getObjInfo( posVector ) );
}

StringInfo* TreeStringSet::getStringInfo( const string& treeString ){
  Exception( !stringIsValid( treeString )                  , "TreeStringSet::getStringInfo ---> invalid string: " + treeString );
  Exception( !objExists( stringToPosVector( treeString ) ) , "TreeStringSet::getStringInfo ---> treeString doesn't exist: " + treeString );  

  vector<uint> posVector;
  stringToPosVector( treeString , posVector );
  return &( (StringInfo&)( getObjInfo( posVector ) ) );
}

uint TreeStringSet::getStringId( const string& treeString ) const{
  Exception( !stringIsValid( treeString )                , "TreeStringSet::getStringId ---> invalid string: " + treeString );
  Exception( !objExists( stringToPosVector(treeString) ) , "TreeStringSet::getStringId ---> treeString doesn't exist: " + treeString );

  return getStringInfo( treeString ).stringId();
}

// Let S be the size of the layer
// Char    Root     Other
//  (      S - 4    S - 3
//  )      S - 3    S - 2
//  ,      S - 2    S - 1
//  $      S - 1    
uint TreeStringSet::charPos( char c , bool root ) const{
  Exception( isdigit(c) , "TreeStringSet::charPos ---> digit are not accepted" );

  uint size = _layerSize;
  if( root ) 
    size += 1;

  if( c == '(' )
    return size - (root? 5: 4);

  if( c == ')' )
    return size - (root? 4: 3);

  if( c == ',' )
    return size - (root? 3: 2);

  if( c == ';' )
    return size - (root? 2: 1);

  if( c == EMPTY_STRING_CHAR ){
    Exception( !root , "TreeStringSet::charPos ---> layer is not the root Layer" );
    return size - 1;
  }
  Exception( true , "TreeStringSet::charPos ---> invalid character " );
}  


void TreeStringSet::stringToPosVector( const string& treeString , vector<uint>& posVector ) const{
  if( treeString.empty() ){
    posVector.resize( 1 );
    posVector[0] = charPos( EMPTY_STRING_CHAR , true );
  }
  else{
    list<uint> posList;
    int i = 0;
    while( i < treeString.size() ){
      if( isdigit( treeString[i] ) )
	posList.push_back( stoi( readTaxaId( treeString , i ) ) );
      else
	posList.push_back( charPos( treeString[i] , i == 0 ) );
      i++;
    }
    posVector = vector<uint> ( posList.begin() , posList.end() );
  }
}

bool TreeStringSet::stringIsValid( const string& treeString ) const{
  int i = 0;
  while( i < treeString.size() ){
    if( isdigit( treeString[i] ) ){
      int d = stoi( readTaxaId( treeString , i ) );
      if( !( 0 <= d && d <= _upperClassId ) )
	return false;
    }
    i++;
  }
  return true;
}


std::ostream& operator << (std::ostream& outputStream, const TreeStringSet& set ){
  outputStream<< "\t\tSize: " << set.size();
  outputStream<< "\n\t\t{String Count Id}";

  for( list< const StringInfo* >::const_iterator iter = set.stringInfoList().begin(); iter != set.stringInfoList().end(); iter++ ){
    const StringInfo& stringInfo = set.getStringInfo( (**iter).String() );
    outputStream <<"\n\t\t\t" << stringInfo;
  }

  return outputStream;
}

const StringInfo* TreeStringSet::addTreeString_BASIC( const string& treeString ){
  list< const StringInfo* >::iterator iter;

  for( iter = stringInfoList().begin(); iter != stringInfoList().end(); iter++ ){
    if( (**iter).String() == treeString ){
      ( const_cast<StringInfo*> (*iter) )->incCounter();
      return *iter;
      break;
    }
  }

  if( _availableStringId.empty() ){
    string funName = "TreeStringSet::addTreeString_BASIC ----> no more available string id" ;
    throw SubException( funName , "" );
    return 0;
  }

  if( iter == stringInfoList().end() ){
    StringInfo* stringInfo = new StringInfo( treeString , _availableStringId.front() , 0 , 0 );
    _availableStringId.pop_front();
    stringInfo->incCounter();
    stringInfoList().push_back( stringInfo );
    return stringInfo;
  }
}


void TreeStringSet::removeStringInfo_BASIC( StringInfo* stringInfo ){
  list< const StringInfo* >::iterator iter;

  for( iter = stringInfoList().begin(); iter != stringInfoList().end(); iter++ ){
    if( (**iter).String() == stringInfo->String() ){
      stringInfo->decCounter();

      if( stringInfo->count() == 0 ){
	_availableStringId.push_front( stringInfo->stringId() );
	delete stringInfo;
	stringInfoList().erase( iter );
      }
      break;
    }
  }

  Exception( iter == stringInfoList().end() , "TreeStringSet::removeStringInfo_BASIC ----> stringInfo not found" );
}
