#include "CladeNode.h"
#include "Edge.h"
#include "Tree.h"
#include "TreePair.h"
#include "Cell.h"
#include "Reconciliation.h"

int CladeNode::_classCounter = 0;

CladeNode::CladeNode( uint depth , const Tree& tree , const CladeNode* father ):
  Node( depth , tree , father ){
  _classCounter++;
  _leftEdge = _rightEdge = 0;

  // See addChild method
  _subtreeSize = 1;
}

void CladeNode::addChild( Node* child ){
  Exception( _leftEdge == 0 & _rightEdge != 0 , "CladeNode::addChild ---> _leftEdge == 0 & _rightEdge != 0" );
  Exception( _leftEdge != 0 & _rightEdge != 0 , "CladeNode::addChild ---> _leftEdge != 0  & _rightEdge != 0" );

  Edge* edge = new Edge( this , child );

  if( _leftEdge == 0 )
    _leftEdge  = edge;
  else
    _rightEdge = edge;

  _subtreeSize += child->subtreeSize();
}

const Node* CladeNode::child( Side s) const{
  Exception( _leftEdge == 0 || _rightEdge == 0, "CladeNode::child ---> _leftChild == 0 || _rightEdge == 0" );
  return edge(s)->childNode();
}


const Edge* CladeNode::edge( Side s ) const{
  Exception( _leftEdge == 0 || _rightEdge == 0 , "CladeNode::edge ---> _leftEdge == 0 || _rightEdge == 0" );\
  return ( s == LEFT? _leftEdge: _rightEdge );
}

Node* CladeNode::nextDFSChild(CHILD_ORDER order) const{
  Exception( _leftEdge == 0 ,  "CladeNode::nextDFSChild() ---> _leftEdge == 0 " );
  Exception( _rightEdge == 0 , "CladeNode::nextDFSChild() ---> _rightEdge == 0 " );

  if( order == LR ){
    if ( !child(LEFT)->dfsVisited() )
      return const_cast<Node*> ( child(LEFT) );
  
    if ( !child(RIGHT)->dfsVisited() )
      return const_cast<Node*> ( child(RIGHT) );
  }
  else{
    if ( !child(RIGHT)->dfsVisited() )
      return const_cast<Node*> ( child(RIGHT) );
    
    if ( !child(LEFT)->dfsVisited() )
      return const_cast<Node*> ( child(LEFT) );
  }

  return 0;
}

bool CladeNode::isAForcedDup() const{
  const TreePair& treePair = _tree.treePair();
  return treePair.mappingFunction( this ) == treePair.mappingFunction( child( LEFT ) ) ||
    treePair.mappingFunction( this ) == treePair.mappingFunction( child( RIGHT ) );
}

///////////////////////////////////////
// Cell set method
void CladeNode::defineCellSet(){
  const TreePair& treePair = _tree.treePair();
  _cellSet.resize( treePair.cellSet().size() , 0 );

  if( isAForcedDup() )
    _lowestCell = treePair.getCell( treePair.mappingFunction( this )->fatherEdge() );
  else
    _lowestCell = treePair.getCell( treePair.mappingFunction( this ) );

  Cell* C = _lowestCell;
  while( !C->isARootCell() ){
    Cell* tmp = C->fatherCell();
    _cellSet[ C->id() ] = tmp;
    C = tmp;
  }
}

Cell* CladeNode::highestCell() const{
  // If u = r(G) or u is a forced duplication on a root edge
  if( _lowestCell->isARootCell() )
    return _lowestCell;

  Cell* cell1 = treePair().getCell( treePair().speciesTree().rootEdge( LEFT  ) );
  Cell* cell2 = treePair().getCell( treePair().speciesTree().rootEdge( RIGHT ) );

  if( *cell1 <= *_lowestCell )
    return cell1;

  if( *cell2 <= *_lowestCell )
    return cell2;

  Exception( true , "Node::highestCell() ---> anormal case" );  
}

const Cell* CladeNode::getChildEdgeCell( const Cell* nodeCell ) const{
  Exception( !nodeCell->isANode() , "CladeNode::getChildEdgeCell ---> nodeCell is an edge" );

  const Cell* leftCell  = nodeCell->Cell::getChildEdgeCell( LEFT  );
  const Cell* rightCell = nodeCell->Cell::getChildEdgeCell( RIGHT );

  if( (*leftCell) <= *_lowestCell )
    return leftCell;

  if( (*rightCell) <= *_lowestCell )
    return rightCell;      

  Exception( true , "CladeNode::getChildEdgeCell ----> anormal case" );

}

void CladeNode::computeNormalFormString( const Reconciliation* reconciliation ){
  if( reconciliation != 0 )
    _subtreeNormForm = reconciliation->getMapping( this )->normFormString();
  
  _subtreeNormForm += "(" + child( LEFT  )->subtreeNormForm() + "," + child( RIGHT )->subtreeNormForm()  + ")";
}


bool CladeNode::isFirstInNormForm( const Node* rightNode ) const{
  if( rightNode->isLeafNode() )
    return true;
  else{
    const CladeNode* cladeNode = (const CladeNode*) (rightNode);

    if( subtreeSize() > cladeNode->subtreeSize() )
      return true;

    if( subtreeSize() < cladeNode->subtreeSize() )
      return false;

    // Here, |T1| == |T2|
    // Compare left and then right children
    for( uint s = 0; s < 2; s++ ){
      const Node* node1 = cladeNode->child( (Side) (s) );
      if( child( (Side) (s) )->isFirstInNormForm( node1 ) )
	return true;

      if( node1->isFirstInNormForm( child( (Side) (s) ) ) )
	return false;
    }
    return true;
  }
}


//////////////////
// Output methods
string CladeNode::primeString( const Reconciliation& reconciliation ) const{
  Exception( isASpeciesNode() , "CladeNode::primeString ---> only for gene node" );
  return primeGeneString( reconciliation );
}
