#include "CladeSpecies.h"
#include "State.h"
#include "MathClass.h"
#include "OrthoParam.h"
#include "TREE_PAIR_CPUtimes.h"

int CladeSpecies::_classCounter = 0;

CladeSpecies::CladeSpecies( const State& state , const CladeNode* cladeNode ):
  Species( state , cladeNode ){
  _classCounter++;
  _slicedNodeCount = 0;
}


const SpeciesEdge* CladeSpecies::edge(Side s)  const{ 
  return _state.getSpeciesEdge( _node->edge(s) ); 
}

const Species* CladeSpecies::child(Side s) const{
  return edge(s)->childNode();
}

uint CladeSpecies::maxNbOfClassId() const{ 
  return _state.treePair().extremeRec( Highest ).computeNbOfCovering( _node );
}

void CladeSpecies::compute_Dy(){
  _Dy =  edge(LEFT)->P_Fun( 0 );
  _Dy += edge(RIGHT)->P_Fun( 0 );

  // Compute the B part of the father edge, which require Dy (this one) of the child edge
  if( !_node->isRootNode() )
    const_cast<SpeciesEdge*> ( fatherEdge() )->compute_B_part();
}

void CladeSpecies::addTreeData( const Gene* gene ){
  try{
    TREE_PAIR_CPUtimes::cpuTimes(ADD_TREE_DATA).startClock();
    Exception( !gene->isCoveredByASpecies() , "CladeSpecies::addTreeData ----> gene is not covered by a species" );

    // Compute left/right strings
    string leftString, rightString;
    if( !gene->getSlicedSubtree( LEFT  ).emptyTree() )
      leftString  = gene->getSlicedSubtree( LEFT  ).slicedRootChild()->treeData().String();

    if( !gene->getSlicedSubtree( RIGHT  ).emptyTree() )
      rightString = gene->getSlicedSubtree( RIGHT  ).slicedRootChild()->treeData().String();

    // TreeString: (leftString;rightString)
    string S = "(" + leftString + ";" + rightString + ")";

    const StringInfo* stringInfo = _treeStringSet.addTreeString_BASIC( S );

    ( const_cast<Gene*> (gene) )->setStringInfo( const_cast<StringInfo*> ( stringInfo ) );
    ( const_cast<Gene*> (gene) )->setTreeData( stringInfo->stringId() );

    TREE_PAIR_CPUtimes::cpuTimes(ADD_TREE_DATA).stopClock();


  }

  catch( SubException subExc ){
    string values;
    values += itos( nodeId() );
    values += "(" + itos( isRootNode() ) + " , ";
    values += setToString( cladeSet() ) + ")";

    values += " : " + gene->geneString();
    
    subExc.addFunDesc( "CladeSpecies::addTreeData" ,  values );
    throw subExc;
  }
}

// The left/right tree strings of the gene are not removed 
// from the left/right TreeStringSet.
// Because it would be a waist of time since they will probably reappear
// later in the MCMC process.
void CladeSpecies::removeTreeData( const Gene* gene ){
  TREE_PAIR_CPUtimes::cpuTimes(REMOVE_TREE_DATA).startClock();
  Exception( !gene->isCoveredByASpecies() , "CladeSpecies::removeTreeData ----> gene is not covered by a species" );

  StringInfo* stringInfo = const_cast<StringInfo*> ( gene->stringInfo() );

  // Either A or B
  _treeStringSet.removeStringInfo_BASIC( stringInfo ); // A
  //  _treeStringSet.decStringInfoCounter( stringInfo ); // B

  ( const_cast<Gene*> (gene) )->setStringInfo( 0 );

  TREE_PAIR_CPUtimes::cpuTimes(REMOVE_TREE_DATA).stopClock();
}

void CladeSpecies::constructAllTreeData( int upperClassId ){
  try{
    // Initialize left/right Tree String Set with the corresponding upperClassId value
    //    _treeStringSet.initialize( child( LEFT )->maxNbOfClassId() + child( RIGHT )->maxNbOfClassId() );
    _treeStringSet.initialize( upperClassId );

    // Compute all tree data for left/right Sliced Subtrees
    for( list< Gene* >::const_iterator iterGene = _geneList.begin(); iterGene != _geneList.end(); iterGene++ ){
      if( !(**iterGene).getSlicedSubtree( LEFT  ).emptyTree() )
	const_cast<SlicedSubtree&> ( (**iterGene).getSlicedSubtree( LEFT  ) ).computeAllTreeData();
      if( !(**iterGene).getSlicedSubtree( RIGHT ).emptyTree() )
	const_cast<SlicedSubtree&> ( (**iterGene).getSlicedSubtree( RIGHT ) ).computeAllTreeData();
    }

    // For all covered genes
    //      - Compute the Class Id for the left/right String pair
    //      - Create the TreeData for the gene
    for( list< Gene* >::const_iterator iterGene = _geneList.begin(); iterGene != _geneList.end(); iterGene++ )
      addTreeData( *iterGene );
  }

  catch( SubException subExc ){
    subExc.addFunDesc( "CladeSpecies::constructAllTreeData" );
    cerr<<endl << subExc.message() <<endl;
    abort();
  }
}

void CladeSpecies::clear_H_Phi_Variables(){
  _treeStringSet.clear();
}

const uint caseA = 0;
const uint caseB = 1;
const uint caseC = 2;

double CladeSpecies::probOfCase( const uint c ) const{
  const SpeciesEdge* leftEdge  = const_cast<SpeciesEdge*> ( edge(LEFT) );
  const SpeciesEdge* rightEdge = const_cast<SpeciesEdge*> ( edge(RIGHT) );

  if( c == caseA )
    return leftEdge->extinctProb()         * ( 1 - rightEdge->extinctProb() );

  if( c == caseB )
    return ( 1 - leftEdge->extinctProb() ) * rightEdge->extinctProb();

  if( c == caseC )
    return ( 1 - leftEdge->extinctProb() ) * ( 1 - rightEdge->extinctProb() );

  Exception( true , "CladeSpecies::probOfCase ---> this case doesn't exist" );
}

void CladeSpecies::generateLeafGermes(){
  // Compute the three cases dist.
  double total = probOfCase( caseA ) + probOfCase( caseB ) + probOfCase( caseC );
  vector< double > casesDist( 3 );
  casesDist[ caseA ] = probOfCase( caseA ) / total;
  casesDist[ caseB ] = casesDist[ caseA ] + probOfCase( caseB ) / total;
  casesDist[ caseC ] = casesDist[ caseB ] + probOfCase( caseC ) / total;

  // Check if the last value of dist is close to one, if yes set it to one.
  Exception( fabs( casesDist[ caseC ] - 1 ) > MIN_PRECISION , "CladeSpecies::generateLeafGermes ---> anormal case" );  
  casesDist[ caseC ] = 1;

  // Genereate the leaf germes for all germes
  const SpeciesEdge* leftEdge  = const_cast<SpeciesEdge*> ( edge(LEFT) );
  const SpeciesEdge* rightEdge = const_cast<SpeciesEdge*> ( edge(RIGHT) );

  for( list<Germe*>::iterator germe = _germesList.begin(); germe != _germesList.end(); germe++ ){
    uint c = MathClass::randomNbFromDist( casesDist );

    if( c == caseA ){
      _nbOfCaseA++;
      (**germe).addLeafGermes( 0 , rightEdge->random_P_Fun_Dist() );
    }
    
    if( c == caseB ){
      _nbOfCaseB++;
      (**germe).addLeafGermes( leftEdge->random_P_Fun_Dist() , 0 );
    }

    if( c == caseC )
      (**germe).addLeafGermes( leftEdge->random_P_Fun_Dist() , rightEdge->random_P_Fun_Dist() );
  }
}

void CladeSpecies::generateLeafGermes2(){
  // Genereate the leaf germes for all germes
  const SpeciesEdge* leftEdge  = const_cast<SpeciesEdge*> ( edge(LEFT) );
  const SpeciesEdge* rightEdge = const_cast<SpeciesEdge*> ( edge(RIGHT) );

  for( list<Germe*>::iterator germe = _germesList.begin(); germe != _germesList.end(); germe++ ){
    uint leftGermes  = leftEdge->randomNbOfLeafGermes();
    uint rightGermes = rightEdge->randomNbOfLeafGermes();

    if( leftGermes != 0 || rightGermes != 0 )
      (**germe).addLeafGermes( leftGermes , rightGermes );
  }
}

void CladeSpecies::deleteLostGermes(){
  list<Germe*> tmpList;

  for( list<Germe*>::iterator germe = _germesList.begin(); germe != _germesList.end(); germe++ ){
    (**germe).updateChildGermes( LEFT );
    (**germe).updateChildGermes( RIGHT );
    if( !(**germe).isLost() )
      tmpList.push_back( *germe );
  }

  _germesList.clear();
  _germesList = tmpList;
}

void printCoveredGenes( std::ostream& output, const CladeSpecies& cladeSpecies ){
  output << "\n\n\tCovered genes: ";

  output <<"\n\tGene\tL-Id\tR-Id\tId\t\tLeft\t\tRight";

  for(list< Gene* >::const_iterator iterGene = cladeSpecies.geneList().begin(); iterGene != cladeSpecies.geneList().end(); iterGene++ ){
    string s1;
    string s2;

    if( !(**iterGene).getSlicedSubtree( LEFT  ).emptyTree() )
      s1 = (**iterGene).getSlicedSubtree( LEFT  ).slicedRootChild()->treeData().String();

    if( !(**iterGene).getSlicedSubtree( RIGHT ).emptyTree() )
      s2 = (**iterGene).getSlicedSubtree( RIGHT ).slicedRootChild()->treeData().String();

    output << "\n\t";
    output << **iterGene;
    output << "\t" << (**iterGene).treeData().String(); 
    output << "\t\t" << s1;
    output << "\t\t" << s2;
  }
}

std::ostream& operator << (std::ostream& output, const CladeSpecies& cladeSpecies ){
  output <<"\nPrime Id " << cladeSpecies.nodeId();

  printCoveredGenes( output , cladeSpecies );

  return output;
}

string CladeSpecies::toString() const{
  return itos( nodeId() );
}

void CladeSpecies::printNbOfCases(std::ostream& output) const{
  output<<"\n\tNb of Cases A: " << _nbOfCaseA;
  output<<"\n\tNb of Cases B: " << _nbOfCaseB;
}

uint CladeSpecies::nbOfRec() const{ 
  return _node->treePair().nbOfRec() + 1; 
}
