#include "Node.h"
#include "CladeNode.h"
#include "Tree.h"
#include "SlicedSubtree.h"
#include "Cell.h"
#include "Reconciliation.h"

uint Node::_currentNodeId  = 0;
int  Node::_classCounter   = 0;

std::ostream& operator << ( std::ostream& output , const Node& node ){
  output << node.nodeId();
  return output;
}

Node::Node( uint depth , const Tree& tree , const CladeNode* father ): 
  _depth( depth ) , _tree( tree ) , _father( father ) , _dfsVisited(false) {
  _classCounter++;
  _fatherEdge = 0;
}

bool Node::isRootNode() const{
  return _father == 0;
}

const Node* Node::sibling() const{
  Exception( isRootNode() , "Node::sibling() ----> is Root" );
  if( isLeftChild() )
    return father()->child( RIGHT );
  else
    return father()->child( LEFT );
}


const TreePair& Node::treePair() const{ 
  return _tree.treePair(); 
}

const CladeNode* Node::father() const{
  Exception( isRootNode() , "Node::father() ---> root node" );
  return _father;
}

bool Node::isLeftChild() const{
  Exception( isRootNode() , "Node::isLeftChild() ---> root node" );
  return _father->child(LEFT) == this;
}

bool Node::isASpeciesNode() const{ return _tree.isASpeciesTree(); }

bool Node::operator == ( const Node& node ) const{
  return _postfixId == node.nodeId();
}

bool Node::operator != ( const Node& node ) const{
  return !( *this == node );
}

bool Node::operator <  ( const Node& node ) const{
  return tree().ancestorMatrixValue( this , &node ); 

}

bool Node::operator <= ( const Node& node ) const{
  return  *this < node || *this == node;
}

bool Node::operator >  ( const Node& node ) const{
  return node < *this;
}

bool Node::operator >= ( const Node& node ) const{
  return *this > node || *this == node;
}


//////////////////////////////////////////////////
// Cell set method

uint Node::T2( const Cell* C ) const{
  return _T2[ C->id() ];
}

void Node::setT2( const Cell* C , uint value ){
  _T2[ C->id() ] = value;
}

// 6 cases are possible
Cell* Node::nextCell( const Cell* P , const Cell* C ) const{
  Cell* N = _cellSet[ P->id() ];

  if( C == 0 )
    return N;

  // C in V(S)
  if( C->isANode() ){
    // #1: P in E(S) & P=(s,s1) in E(S) & s == C
    if( !P->isANode() && P->edge()->fatherNode() == C->node() )
      return 0;

    // #2: P in V(S) 
    // #3: C < s
    return N;
  }

  // C=(t,t1) in E(S)
  else{
    // #4: P=(s,s1) in E(S) & (s,s1) = (t,t1)
    if( P == C )
      return 0;

    // #5: P in V(S) 
    // #6: (s,s1) > (t,t1)
    return N;
  }
}

void Node::defineRandomTables(){
  try{
    // Define T1 table size
    uint size = 0;
    for( Cell* C = lowestCell(); C != 0; C = nextCell( C ) )
      size += treePair().T( this , C );

    // Define T1 and T2 table
    _T1.resize( size , 0 );
    _T2.resize( _cellSet.size() , 0 );
    uint i = 0;

    for( Cell* C = lowestCell(); C != 0; C = nextCell( C ) ){
      for( uint j = 0; j < treePair().T( this , C ); j++ )
	_T1[ i++ ] = C;

      setT2( C , (--i)++ );
    }
  }

  catch( std::length_error exc ){
    Exception( true , "Batch to handle too large vector size (Fri Dec  7)" );
  }

  catch( std::bad_alloc exc ){
    Exception( true , "Node::defineRandomTables ----> bad alloc" );
  }

}

Cell* Node::randomCell( const Cell* fatherCell ) const{
  Cell* maxCell = const_cast<Cell*> ( fatherCell );

  if( fatherCell->isANode() ){
    const Node* lca = treePair().mappingFunction( this );
    const CladeNode* node = (const CladeNode*) ( fatherCell->node() );
    if( *( node->child( LEFT ) ) <= *lca )
      maxCell = treePair().getCell( node->edge( LEFT ) );
    else
      maxCell = treePair().getCell( node->edge( RIGHT ) );
  }

  return _T1[ MathClass::random( 0 , T2( maxCell ) ) ];
}

const Cell* Node::getCellDownward( const Cell* cell ) const{
  Exception( isLowestCell( cell ) , "Node::getCellDownward ---> is first cell " );
  // Cell is not a node of S

  const TreePair& treePair = cell->treePair();

  if( isLowestCell( cell->getChildNodeCell() ) )
    return cell->getChildNodeCell();

  else{
    const Cell* cell1 = cell->getChildEdgeCell( LEFT );
    const Cell* cell2 = cell->getChildEdgeCell( RIGHT );
    if( *cell1 <= *( lowestCell() ) )
      return cell1;
    else 
      return cell2;
  }
}

const Cell* Node::getDirCell( DIRECTION dir , const Cell* cell ) const{
  if( dir == UP )
    return cell->fatherCell();
  else
    return getCellDownward( cell );
}

uint Node::distNMC( const Cell* cell1 , const Cell* cell2 )const{
  Exception( !( *cell1 <= *_lowestCell ) , "Node::distNMC ----> cell1 is not valid" );
  Exception( !( *cell2 <= *_lowestCell ) , "Node::distNMC ----> cell2 is not valid" );
  
  Cell* tmp1 = const_cast<Cell*> (cell1);
  Cell* tmp2 = const_cast<Cell*> (cell2);

  if( *tmp2 < *tmp1 )
    interchange(tmp1,tmp2);

  // Here, tmp1 <= tmp2

  if( tmp1 == tmp2 )
    return 0;

  if( !tmp1->isANode() && tmp2->isANode() )
    return distCell( tmp1 , tmp2 ) + 1;

  if( !tmp1->isANode() && !tmp2->isANode() )
    return distCell( tmp1 , tmp2 );

  Exception( true , "Node::distNMC ---> anormal case" );
}

string Node::primeGeneString( const Reconciliation& reconciliation ) const{
  string primeString;
  const ParamSpeciesTree paramSpeciesTree = reconciliation.speciesTree().paramTree();

  //  cerr<<"\nPrime gene string: " << *this;

  if( isLeafNode() )
    primeString += paramSpeciesTree.id2SpeciesName( itos( ( (const LeafNode*) (this) )->specieId() ) );

  primeString += "[&&PRIME";  
  primeString += " ID=" + itos( nodeId() );

  // D flag only for non leaf node
  // D= 1 iff non speciation, otherwise D=0
  if( isLeafNode() )
    primeString += " S=" + paramSpeciesTree.id2SpeciesName( itos( ( (const LeafNode*) (this) )->specieId() ) );
  else
    primeString += " D=" + itos( !reconciliation.isCoveredByASpecies(this) );

  // Define the AC string
  string ac;

  // Step 1: define species1 used in the bottom-up traversal of the species tree
  const Node* species1;
  if( reconciliation.isCoveredByASpecies(this) ){
    const Node* coveringSpecies = reconciliation.coveringSpecies(this);
    ac += itos( coveringSpecies->nodeId() );
    species1 = coveringSpecies->isRootNode()? 0: coveringSpecies->father();
  }
  else
    species1 = reconciliation.coveringSpeciesEdge(this)->fatherNode();

  if( !isRootNode() ){
    // Step 2: define species2
    const Node* species2 = 0;

    if( reconciliation.isCoveredByASpecies( _father ) )
      species2 = reconciliation.coveringSpecies(_father);
    else
      species2 = reconciliation.coveringSpeciesEdge(_father)->fatherNode();
  
    // Step 3: bottom-up traversal to print AC species id
    while( *species1 > *species2 ){
      ac += " " + itos( species1->nodeId() );
      if( species1->isRootNode() )
	break;
      species1 = species1->father();
    }
  }

  if( !ac.empty() )
    primeString += " AC=(" + ac + ")";
      
  primeString += "]";

  return primeString;
}
