/***************************************************************************
			Node.h
		-------------------
	begin	 : Wed Jun  7 17:27:25 EDT 2006
	copyright: (C) 2005 by Universite De Montreal
	email	 :  doyonjea@iro.umontreal.ca
***************************************************************************/
#ifndef NODE_H
#define NODE_H

//! This class defines the base structure for a Node object.
/**
*@author Jean-Philippe Doyon
*/

// This is an ABSTRACT CLASS

#include "StdClass.h"
#include "toolBox.h"

class Node;
class CladeNode;
class Edge;
class Tree;
class TreePair;
class Cell;
class Reconciliation;


////////////////////////////////////////////////////
class Node{

  friend std::ostream& operator << ( std::ostream& output , const Node& node );

 private:
  static int _classCounter;
  static uint _currentNodeId;

  uint _depth;
  bool _dfsVisited;

  const CladeNode* _father;

 protected:
  // The postfix node id is used by Tree::_nodeVector
  uint _postfixId;
  // The prefix node id is used to perform the exhaustive exploration algorithm of reconciliations 
  uint _prefixId;

  const Tree& _tree;
  Edge* _fatherEdge;

  //////////////////////////////
  // Only for the node u in V(G)

  // Implementation of the set M(u)
  // M(u) is implemented as a chain s.t.
  //   lowestCell: first c in M(u)
  //   M[i]: i+1 th cell in M(u)
  //   i is the cell id of the i th c in M(u)
  Cell* _lowestCell;
  vector<Cell*> _cellSet;

  // For a in M(u)
  // T1[i] = a
  vector<Cell*> _T1;

  // For a in M(u)
  // T2[u] is a natural
  vector<uint> _T2;

  uint _subtreeSize;
  string _subtreeNormForm;

 public:

  static int classCounter(){ return _classCounter; };
  static uint currentNodeId() { return _currentNodeId; }
  static void resetNodeId()   { _currentNodeId = 0; }

  // Constructor and Destructor
  Node( uint depth , const Tree& tree , const CladeNode* father = 0 ); 

  virtual ~Node(){ _classCounter--; }

  virtual const Edge* edge( Side s ) const = 0;
  const Edge* fatherEdge() const { return _fatherEdge; }

  // Access methods
  uint nodeId()            const { return _postfixId; }
  uint prefixId()          const { return _prefixId; }

  uint depth()             const { return _depth; }
  const Tree& tree()       const { return _tree; }
  bool dfsVisited()        const { return _dfsVisited; }

  const vector<Cell*>& T1() const { return _T1; }

  const Node* sibling() const;


  const TreePair& treePair() const;

  const CladeNode* father() const;

  void setFatherEdge( Edge* fatherEdge ) { _fatherEdge = fatherEdge; }
  void setDFS_Visited( bool b ) { _dfsVisited = b; }

  void setPostfixNodeId() { _postfixId = _currentNodeId++; }
  void setPrefixId( uint id ) { _prefixId = id; }

  // Boolean methods
  bool nullFatherEdge() const { return _fatherEdge == 0; }
  bool isRootNode() const;
  bool isLeftChild() const;
  bool isASpeciesNode() const;

  // Cell set method
  virtual void defineCellSet() = 0;
  Cell* lowestCell() const { return _lowestCell; }

  virtual Cell* highestCell() const = 0;
  Cell* nextCell( const Cell* P , const Cell* C = 0 ) const;
  void defineRandomTables();
  bool isLowestCell( const Cell* cell ) const{ return _lowestCell == cell; }

  uint T2( const Cell* C ) const;
  void setT2( const Cell* C , uint value );

  Cell* randomCell( const Cell* fatherCell ) const;

  const Cell* getCellDownward( const Cell* cell ) const;
  const Cell* getDirCell( DIRECTION dir , const Cell* cell ) const;

  // Number of NMC operators from cell1 to cell2
  uint distNMC( const Cell* cell1 , const Cell* cell2 ) const;

  // Virtual methods
  virtual bool isLeafNode()         const = 0;
  virtual bool withoutEdge()        const = 0;
  virtual bool fictiveLeaf()        const = 0;

  virtual const UintSet& cladeSet() const = 0;
  virtual string primeString( const Reconciliation& reconciliation ) const = 0;
  string primeGeneString( const Reconciliation& reconciliation ) const;


  // Normal form methods
  uint subtreeSize() const { return _subtreeSize; }
  string subtreeNormForm() const { return _subtreeNormForm; }
  virtual bool isFirstInNormForm( const Node* rightNode ) const = 0;


  // Binary comparaison operators
  // x < y iff x is an ancestor of y
  bool operator == ( const Node& node ) const;
  bool operator != ( const Node& node ) const;
  bool operator <  ( const Node& node ) const;
  bool operator <= ( const Node& node ) const;
  bool operator >  ( const Node& node ) const;
  bool operator >= ( const Node& node ) const;
  bool disjointTo  ( const Node& node ) const;

};

#endif
