#include "RealGene.h"
#include "State.h"
#include "SpeciesEdge.h"
#include "Species.h"
#include "GeneEdge.h"
#include "ArtifGene.h"

int RealGene::_classCounter = 0;

RealGene::RealGene( const State& state , const Node* node ):
  Gene( state ) , _node( node ){
  _classCounter++;
  _speciesEdge   = 0;
  _slicedSubtree = 0;
}

void RealGene::setSlicedSubtree( SlicedSubtree* tree ){
  Exception( isCoveredByASpecies() , "RealGene::setSlicedSubtree ---> Real Gene is covered by a species " );
  _slicedSubtree = tree;
}

const SlicedSubtree* RealGene::slicedSubtree() const{
  Exception( isCoveredByASpecies() , "RealGene::slicedSubtree ---> Real Gene is covered by a species " );
  Exception( _slicedSubtree == 0   , "RealGene::slicedSubtree ---> sliced subtree undefined ");
  return _slicedSubtree;
}

const RealGene* RealGene::retrieveChild( const SpeciesEdge* speciesEdge )const{
  Exception( !isCoveredByASpecies()                , "RealGene::retrieveChild ---> Real Gene is NOT covered by a species " );
  Exception( _species != speciesEdge->fatherNode() , "RealGene::retrieveChild ---> speciesEdge (1,2) is not s.t. rec(u) = 1 " );
  
  for( uint side = 0; side < 2; side++ ){
    const RealGene* gene = child( (Side) (side) );
    if( gene->isCoveredByASpecies() && *( speciesEdge->childNode() ) <= * ( gene->coveringSpecies() ) )
      return gene;

    if( !gene->isCoveredByASpecies() && *speciesEdge <= *( gene->coveringSpeciesEdge() ) )
      return gene;
  }

  Exception( true , "RealGene::retrieveChild --> child not found" );  
}

const RealGene* RealGene::child( Side s ) const{
  return (const RealGene*) ( edge( s )->childNode() );
}

const Gene* RealGene::father() const{
  Exception( isRootNode() , "RealGene::father() ---> root node" );
  return fatherEdge()->fatherNode();
}

const GeneEdge* RealGene::edge( Side s ) const{ 
  return _state.getGeneEdge( _node->edge(s) ); 
}

const GeneEdge* RealGene::fatherEdge() const{ 
  return _state.getGeneEdge( _node->fatherEdge() ); 
}

void RealGene::setCovering( const SpeciesEdge* speciesEdge ){
  Exception( speciesEdge == 0 , "RealGene::setCovering ---> speciesEdge == 0" );
  _species = 0;
  _speciesEdge = speciesEdge;
  const_cast<SpeciesEdge*> (_speciesEdge)->addCoveredGene( this );
}

void RealGene::changeCovering( const SpeciesEdge* newEdge ){
  Exception( newEdge == 0                                      , "RealGene::changeCovering ---> newEdge == 0" );
  Exception( !isCoveredByASpecies() && _speciesEdge == newEdge , "RealGene::changeCovering ---> _speciesEdge == newEdge" );

  if( isCoveredByASpecies() )
    const_cast<Species*> (_species)->removeCoveredGene( this );
  else
    const_cast<SpeciesEdge*> (_speciesEdge)->removeCoveredGene( this );

  _speciesEdge = newEdge;
  _species = 0;
  const_cast<SpeciesEdge*> (_speciesEdge)->addCoveredGene( this );
}

void RealGene::changeCovering( const Species* species ){
  Exception( species == 0          , "RealGene::changeCovering ---> species == 0" );
  Exception( isCoveredByASpecies() , "RealGene::changeCovering ---> already covered by a species" );

  const_cast<SpeciesEdge*> (_speciesEdge)->removeCoveredGene( this );

  _speciesEdge = 0;
  _species = species;

  const_cast<Species*> (_species)->addCoveredGene( this );
}

bool RealGene::isCoveredBy( const SpeciesEdge* speciesEdge ) const{
  return !isCoveredByASpecies() && _speciesEdge == speciesEdge;
}

const SpeciesEdge* RealGene::coveringSpeciesEdge() const{
  Exception( _speciesEdge == 0 , "RealGene::coveringSpeciesEdge ---> _speciesEdge == 0" );
  return _speciesEdge;
}

const SpeciesEdge* RealGene::coveringChildSpeciesEdge() const{
  Exception( _speciesEdge == 0          , "RealGene::coveringChildSpeciesEdge ---> _speciesEdge == 0" );
  Exception( _speciesEdge->isLeafEdge() , "RealGene::coveringChildSpeciesEdge ---> covering species edge is a leaf" );

  Exception( !slicedChild(LEFT)->isArtifNode() ,  "RealGene::coveringChildSpeciesEdge  ---> left sliced child is not artif gene" );
  Exception( !slicedChild(RIGHT)->isArtifNode() , "RealGene::coveringChildSpeciesEdge ---> right sliced child is not artif gene" );

  const ArtifGene* A1 = (const ArtifGene*) ( slicedChild(LEFT) );
  const ArtifGene* A2 = (const ArtifGene*) ( slicedChild(RIGHT) );

  Exception( A1->speciesEdgeSide() != A2->speciesEdgeSide() , "RealGene::coveringChildSpeciesEdge ---> covering species edge are different" );

  return coveringSpeciesEdge()->childEdge( A1->speciesEdgeSide() );
}

bool RealGene::isBlockedBy( Side side , DIRECTION dir ) const{
  Exception( isCoveredByASpecies() , "RealGene::isBlockedBy ---> RealGene is covered by a species" );
  Exception( dir == UP             , "RealGene::isBlockedBy ---> this case is not considered yet" );

  const RealGene* Child = child( side );
  if( Child->isCoveredByASpecies() )
    return Child->coveringSpecies() == _speciesEdge->childNode();
  else
    return Child->coveringSpeciesEdge() == _speciesEdge;
}

uint RealGene::coveringLayerId() const{
  if( isCoveredByASpecies() )
    return coveringSpecies()->nodeId();
  else
    return coveringSpeciesEdge()->layerId();
}

const Gene* RealGene::slicedFather() const{
  if( fatherEdge()->lastArtifGene() != 0  )
    return (const Gene*) ( fatherEdge()->lastArtifGene() );
  else
    return fatherEdge()->fatherNode();
}

const Gene* RealGene::slicedChild( Side s , const SlicedSubtree& slicedSubtree ) const{
  Exception( slicedSubtree.isTheSlicedRoot( this ) , "RealGene::slicedChild ----> sliced root" );

  if( edge(s)->firstArtifGene() != 0 )
    return (const Gene*) ( edge(s)->firstArtifGene() );
  else
    return edge(s)->childNode();
}

const Gene* RealGene::slicedChild ( Side s ) const{
  Exception( isCoveredByASpecies() , "RealGene::slicedChild(Side) ---> Real Gene is covered by a species " );
  Exception( _slicedSubtree == 0   , "RealGene::slicedChild(Side) ---> sliced subtree undefined ");
  return slicedChild( s , *_slicedSubtree );
}

const Gene* RealGene::nextDFS_SlicedChild( const SlicedSubtree& slicedSubtree ) const{
    
  // Case 1: sliced root
  if( slicedSubtree.isTheSlicedRoot( this ) ){
    const Gene* rootChild = slicedSubtree.slicedRootChild();
    if( !rootChild->dfsVisited() )
      return rootChild;
    else
      return 0;
  }
  // Case 2: not sliced root
  else{
    if ( !slicedChild( LEFT , slicedSubtree )->dfsVisited() )
      return slicedChild( LEFT , slicedSubtree );
  
    if ( !slicedChild( RIGHT , slicedSubtree )->dfsVisited() )
      return slicedChild( RIGHT , slicedSubtree );

    return 0;
  }
}

bool RealGene::isFirstSlicedChild( const SlicedSubtree& slicedSubtree  , const Gene* childGene ) const{
  if( slicedSubtree.isTheSlicedRoot( this ) )
    return childGene == slicedSubtree.slicedRootChild();
  else
    return childGene == slicedChild( LEFT , slicedSubtree );
}

void RealGene::updateTreeData(){
  Exception( isCoveredByASpecies() , "RealGene::updateTreeData ---> is covered by a species" );
  Exception( _treeData == 0        , "Gene::updateTreeData     ---> treeData is not defined" );
  delete _treeData;
  _treeData = new TreeData( slicedChild( LEFT )->treeData() , slicedChild( RIGHT )->treeData() );
}

std::ostream& operator << (std::ostream& outputStream, const RealGene& realGene ){
  outputStream<< "{";
  outputStream << realGene.geneString();

//   if( !realGene.isRootNode() )
//     outputStream<< "," << realGene.slicedFather()->geneString();

  outputStream<< "}";

  return outputStream;
}

void RealGene::checkNode( std::ostream& output ) const{
  output<< "\nCheck node " << nodeId() <<" ... ";

  if( !isCoveredByASpecies() ){
    slicedChild(LEFT);
    slicedChild(RIGHT);
  }
  output <<" done.\n";
}

