/***************************************************************************
			RealGene.h
		-------------------
	begin	 : Tue Jan 16 10:09:59 EST 2007
	copyright: (C) 2005 by Universite De Montreal
	email	 :  doyonjea@iro.umontreal.ca
***************************************************************************/
#ifndef REALGENE_H
#define REALGENE_H

//! This class defines the base structure for a RealGene object.
/**
*@author Jean-Philippe Doyon
*/

#include "Node.h"
#include "Gene.h"

class RealGene:public Gene{

  friend std::ostream& operator << (std::ostream& outputStream, const RealGene& realGene );

 private:
  static int _classCounter;

  // The corresponding node in the gene tree 
  const Node* _node; 

  // Reconciliation attributes: 
  // A RealGene is either reconcilied to a SpeciesNode or a SpeciesEdge
  // The SpeciesNode attribute is in Gene class
  const SpeciesEdge* _speciesEdge;

  // This is neither the left/right/father SlicedSubtree
  // It is the Unique Tree that contains this node
  // For RealGene covered by Species Edge only
  SlicedSubtree* _slicedSubtree;

 public:

  static int classCounter(){ return _classCounter; };

  // Constructor and Destructor
  RealGene( const State& state , const Node* node );

  ~RealGene(){ _classCounter--; }

  void clearForRedefine(){
    Gene::clearForRedefine();
    _speciesEdge = 0;
    _slicedSubtree = 0;
  }

  // Access methods (Class attributes)
  const Node* node() const { return _node; }

  // For RealGene covered by an edge only
  void setSlicedSubtree( SlicedSubtree* tree );
  const SlicedSubtree* slicedSubtree() const;

  // For RealGene u covered by a node only
  // Antecedent: speciesEdge (1,2) s.t. rec(u) = 1
  // Consequent: Return the gene child u' s.t. rec(u') <= (1,2)
  const RealGene* retrieveChild( const SpeciesEdge* speciesEdge ) const;

  // DFS visited status
  bool dfsVisited() const     { return _node->dfsVisited(); }
  void setDFS_Visited(bool b) { ( const_cast<Node*> (_node) )->setDFS_Visited(b); }

  // Node type methods
  bool isLeafNode()  const { return _node->isLeafNode(); }
  bool isRootNode()  const { return _node->isRootNode(); }
  bool isArtifNode() const { return false; }
  bool isLeftChild() const { return _node->isLeftChild(); }
  bool fictiveLeaf() const { return _node->fictiveLeaf(); }
  
  // Return the RealGene child
  const RealGene* child( Side s ) const;

  // Shortcut methods
  const Gene* father() const;
  const GeneEdge* edge( Side s ) const;
  const GeneEdge* fatherEdge() const;

  const Tree& tree() const { return _node->tree(); }

  uint nodeId()     const { return _node->nodeId(); }

  // Reconciliation methods
  void setCovering   ( const SpeciesEdge* speciesEdge );
  void changeCovering( const SpeciesEdge* newEdge );
  void changeCovering( const Species* species );

  bool isCoveredBy( const SpeciesEdge* speciesEdge ) const;
  const SpeciesEdge* coveringSpeciesEdge() const;

  const SpeciesEdge* coveringChildSpeciesEdge() const; // The child species edge that can covers this

  bool isBlockedBy( Side side , DIRECTION dir ) const;

  uint coveringLayerId() const; // The layerId of either the speciesNode or speciesEdge 

  // Sliced subtree functions
  const Gene* slicedFather() const;
  const Gene* slicedChild ( Side s , const SlicedSubtree& slicedSubtree ) const;
  const Gene* slicedChild ( Side s ) const;

  const Gene* nextDFS_SlicedChild( const SlicedSubtree& slicedSubtree ) const;
  bool isFirstSlicedChild        ( const SlicedSubtree& slicedSubtree  , const Gene* childGene ) const;

  // Update the tree data according to its children
  // Only for RealGene covered by species edge
  // Used in the case of Dup2DownDup
  void updateTreeData();

  // Debug methods
  void checkNode( std::ostream& output ) const; // Verify that all node attributes are OK

  // Output methods
  string geneString() const { return itos( nodeId() ); }
};

#endif
