#include "Reconciliation.h"
#include "Cell.h"
#include "Types.h"
#include "State.h"

Reconciliation::Reconciliation( const TreePair& treePair ):
  _treePair( treePair ) , _mappingVector( treePair.geneTree().nbOfNodes() , 0 ){
}

Reconciliation::Reconciliation( const TreePair& treePair , const vector<uint>& inputVector ):
  _treePair( treePair ) , _mappingVector( treePair.geneTree().nbOfNodes() , 0 ){
  redefine( inputVector );
}

Reconciliation::Reconciliation( const Reconciliation& origin ):
  _treePair       ( origin.treePair() ) ,
  _mappingVector  ( origin.mappingVector() ) , 
  _localMin       ( origin.localMin() ) ,
  _valuationValues( origin.valuationValues() ){
}

Reconciliation::Reconciliation( const Reconciliation& origin , const Slide& slide ):
  _treePair       ( origin.treePair() ) ,
  _mappingVector  ( origin.mappingVector() ) , 
  _localMin       ( origin.localMin() ) ,
  _valuationValues( origin.valuationValues() ){

  applyNMC( slide , 0 , 0 );
}

Reconciliation::Reconciliation( const TreePair& treePair , const string& geneTreeString ):
  _treePair( treePair ) , _mappingVector( treePair.geneTree().nbOfNodes() , 0 ){
  defineFromGeneTreeString( geneTreeString );
}

void Reconciliation::defineFromGeneTreeString( const string& geneTreeString ){

  // Set the mapping for all genes: Read geneTreeString
  list<Node*>::const_iterator node = _treePair.geneTree().nodeList().begin();
  for( int i = 0; i < geneTreeString.size(); i++ ){
    // Clade gene node
    if( geneTreeString[i] == '[' ){
      // Move at the begining of cell id
      i++;

      string cellStringId = readTaxaId( geneTreeString , i );

      // Move at char \]
      i++;

      _mappingVector[ (**node).nodeId() ] = const_cast<Cell*> ( _treePair.retrieveCell( *node , cellStringId ) );

      node++;
      continue;
    }
    if( isalnum( geneTreeString[i] ) ){
      // Skip taxa id
      while( isalnum( geneTreeString[i] ) ) i++;

      // A leaf g \in V(G) is mapped onto M(g)
      _mappingVector[ (**node).nodeId() ] = (**node).lowestCell();

      // Skip Leaf node
      node++;
    }
  }

  // Compute dup/lost/mut costs
  for( uint val = 0; val < Types::nbOfValuations(); val++ ){
    Valuation valuation = (Valuation) (val);
    if( valuation != Lik && valuation != PostProb )
      _valuationValues.setValue( val , computeCost( val ) );
  }

  Exception( !isValid() , " Reconciliation::Reconciliation ----> reconciliation is not valid" );
}

void Reconciliation::redefine( const vector<uint>& inputVector ){
  // Set the mapping for all genes
  for( uint i = 0; i < inputVector.size(); i++ )
    _mappingVector[ i ] = _treePair.getCell( inputVector[i] );

  // Compute dup/lost/mut costs
  for( uint val = 0; val < Types::nbOfValuations(); val++ ){
    Valuation valuation = (Valuation) (val);
    if( valuation != Lik && valuation != PostProb )
      _valuationValues.setValue( val , computeCost( val ) );
  }

  Exception( !isValid()                 , "Reconciliation::Reconciliation ----> reconciliation is not valid" );
  Exception( recVector() != inputVector , "Reconciliation::Reconciliation ----> error" );
}

////////////////////////
// Set Mapping functions
////////////////////////

void Reconciliation::updateLostCost( const Node* node , OperatorDir opDir ){
  // # losses for u = r(G) & alpha(u) != r(S)
  if( node->isRootNode() && getMapping( node ) != getCell( speciesTreeRoot() ) )
    _valuationValues.addSub( Lost , opDir , distCell( getCell( speciesTreeRoot() ) , getMapping( node ) ) + 1 );

  // # losses for u != r(G)
  if( !node->isRootNode() )
    _valuationValues.addSub( Lost , opDir , distCell( getMapping( node->father() ) , getMapping( node ) ) );
}

void Reconciliation::setExtremeMapping( ExtremeType extremeType , const Node* node ){
  if( extremeType == Lowest )
    _mappingVector[ node->nodeId() ] = node->lowestCell();
  else
    _mappingVector[ node->nodeId() ] = node->highestCell();

  // # dup
  if( !getMapping( node )->isANode() )
    _valuationValues.add( Dup );

  // # losses
  updateLostCost( node , Add );
}

void Reconciliation::moveMappingUpward( const Node* node ){
  Exception( !isNMC( Slide( node->nodeId() , UP ) ) , "Reconciliation::moveMappingUpward ---> is highest mapping" );

  // # dup: Speciation to duplication
  if( _mappingVector[ node->nodeId() ]->isANode() )
    _valuationValues.add( Dup );

  // # losses
  updateLostCost( node , Sub );

  _mappingVector[ node->nodeId() ] =  _mappingVector[ node->nodeId() ]->fatherCell();

  updateLostCost( node , Add );
}

void Reconciliation::clearMapping( const Node* node ){
  // # dup
  if( !getMapping( node )->isANode() )
    _valuationValues.sub( Dup );

  // # losses
  updateLostCost( node , Sub );

  _mappingVector[ node->nodeId() ] = 0;
}

///////////////////////////
// Access mapping functions
///////////////////////////

const Cell* Reconciliation::getMapping( const Node* node ) const{
  Exception( _mappingVector[ node->nodeId() ] == 0 , "Reconciliation::getMapping ----> is not defined for " + itos( node->nodeId() ) );
  return _mappingVector[ node->nodeId() ];
}

///////////////////////////
// Covering functions
///////////////////////////

bool Reconciliation::isCoveredByASpecies( const Node* node ) const{
  return getMapping( node )->isANode();
}

const Edge* Reconciliation::coveringSpeciesEdge( const Node* node ) const{
  return getMapping( node )->edge();
}

const Node* Reconciliation::coveringSpecies( const Node* node ) const{
  return getMapping( node )->node();
}

bool Reconciliation::geneEdgePassThroughSpecies( const Edge* geneEdge , const Node* targetSpecies ) const{
  const Cell* fatherCell = getMapping( geneEdge->fatherNode() );
  const Cell* childCell  = getMapping( geneEdge->childNode() );
  const Cell* targetCell = treePair().getCell( targetSpecies );

  return *fatherCell < *targetCell && *targetCell < *childCell;
}


uint Reconciliation::computeNbOfCovering( const Node* targetSpecies ) const{
  Exception( !targetSpecies->isASpeciesNode() , "Reconciliation::computeNbOfCovering ---> only for species node" );

  uint nbOfCovering = 0;

  // x is targetSpecies

  // Count the number of u \in V(G) s.t. rec(u) == x
  for( vector<Node*>::const_iterator gene = geneTree().nodeVector().begin(); gene != geneTree().nodeVector().end(); gene++ )
    if( getMapping( *gene ) == treePair().getCell( targetSpecies ) )
      nbOfCovering++;

  // Count the number of (u,v) \in E(G) s.t. rec(u) < x < rec(v)
  for( vector<Edge*>::const_iterator edge = geneTree().edgeVector().begin(); edge != geneTree().edgeVector().end(); edge++ )
    if( geneEdgePassThroughSpecies( *edge , targetSpecies ) )
      nbOfCovering++;

  return nbOfCovering;
}

uint Reconciliation::computeMaximalNbOfCovering() const{
  uint max = 0;
  for( vector< Node* >::const_iterator species = speciesTree().nodeVector().begin(); species != speciesTree().nodeVector().end(); species++ ){
    uint current = computeNbOfCovering( *species );
    if( current > max )
      max = current;
  }
  return max;
}
      

///////////////////////////
// Boolean functions
///////////////////////////

bool Reconciliation::ancestorConstraints( const CladeNode* cladeNode , Side side ) const{
  const Cell* cell      = getMapping( cladeNode );
  const Cell* childCell = getMapping( cladeNode->child( side ) );

  if( !cell->isANode() && !childCell->isANode() )
    return *cell <= *childCell;
  else
    return *cell < *childCell;
}

bool Reconciliation::isValid() const{
  for( uint i = 0; i < size(); i++ ){
    const Node* node = geneNode(i);
    const Cell* cell = getMapping( i );

    // u = r(G)
    if( node->isRootNode() ){
      // L(G) = L(S)
      if( _treePair.coversAllSpecies() ){

	// r(G) is a forced up
	if( ( (const CladeNode*) (node) )->isAForcedDup() )
	  return false; // # of rec = 0

	if( !cell->isANode() )
	  return false;
      }
    }

    // u \in L(G) : Base Case Constraint
    if( node->isLeafNode() )
      return ( cell->isANode() && mappingFunction( node ) == cell->node() );

    // u \notin L(G)
    else{
      const CladeNode* cladeNode   = (const CladeNode*) (node);
      const Cell*      mappingCell = _treePair.getCell( mappingFunction( node ) );

      // Tree Mapping Constraints
      if( cell->isANode() )
	if( cladeNode->isAForcedDup() || cell != mappingCell )
	  return false;
      else
	if( !( (*cell) <= (*mappingCell) ) )
	  return false;

      // Ancestor Relationship Constraints
      if( !ancestorConstraints( cladeNode , LEFT ) || !ancestorConstraints( cladeNode , RIGHT ) )
	return false;
    }
  }
  
  // The reconciliation is Valid
  return true;
}

///////////////////////////
// Cost functions
///////////////////////////

uint Reconciliation::computeCost( uint val ) const{
  Valuation valuation = (Valuation) (val);
  Exception( valuation == Lik      , "Reconciliation::computeCost ----> not for likelihood" );
  Exception( valuation == PostProb , "Reconciliation::computeCost ----> not for Post. Prob" );

  uint cost = 0;

  // Lost / Mutation costs ---> compute nb of losses
  if( valuation != Dup ){

    // If alpha(r(G)) != r(S) ---> D( r(S) , alpha(r(G)) ) + 1
    if( getMapping( geneTreeRoot() ) != getCell( speciesTreeRoot() ) )
      cost += distCell( getCell( speciesTreeRoot() ) , getMapping( geneTreeRoot() ) ) + 1;

    for( uint i = 0; i < _mappingVector.size(); i++ ){
      if( !geneNode(i)->isLeafNode() ){
	const Cell* cell  = getMapping( i );
	const Cell* cell1 = getMapping( ( (const CladeNode*) ( geneNode(i) ) )->child( LEFT ) );
	const Cell* cell2 = getMapping( ( (const CladeNode*) ( geneNode(i) ) )->child( RIGHT ) );
	cost += distCell( cell , cell1 ) + distCell( cell , cell2 );
      }
    }
  }

  // Duplication / Mutation costs ---> compute nb of duplications
  if( valuation != Lost ){
    for( uint i = 0; i < _mappingVector.size(); i++ )
      if( !_mappingVector[i]->isANode() )
	cost++;
  }

  return cost;
}

void Reconciliation::computeLocalMin( double likSum , State* state ){
  _localMin.resize( Types::nbOfValuations() , true );

  // Set likelihood value
  if( state != 0 ){
    _valuationValues.setValue( Lik , exp( state->logL() ) );
    if( likSum != 0 )
      _valuationValues.setValue( PostProb , _valuationValues.value( Lik ) / likSum );
  }

  // Save current values
  ValuationValues originValues( _valuationValues );

  // Visits all reconciliation neighbor:
  //    traverses node and direction
  for( uint i = 0; i < _mappingVector.size(); i++ ){
    if( geneNode(i)->isLeafNode() )
      continue;

    for( uint d = 0; d < 2; d++ ){
      Slide slide( i , DIRECTION(d) );

      if( isNMC( slide ) ){
	applyNMC( slide , likSum , state );

	// Consider neighbor valuation costs
	for( uint val = 0; val < Types::nbOfValuations(); val++ )
	  if( _valuationValues.hasBetterValue( val , originValues.value( val ) ) )
	    _localMin[val] = false;

	applyNMC( slide.inverse() , likSum , state );
      }
    }
  }
}

///////////////////////////
// NMC operators methods
///////////////////////////

bool Reconciliation::isNMC( const Slide& slide ) const{

  const Node* node = geneNode( slide.nodeId() );

  // If u \in L(G)
  if( node->isLeafNode() )
    return false;

  const Cell* cell = getMapping( node );
  Exception( cell == 0 , "Reconciliation::isNMC --> node mapping undefined" );

  // Upward NMC & u = r(G)
  if( slide.dir() == UP && node->isRootNode() )
    return !cell->isARootCell();

  // Upward NMC & u != r(G)
  if( slide.dir() == UP && !node->isRootNode() ){
    const Cell* fatherCell = getMapping( node->father() );
    Exception( fatherCell == 0 , "Reconciliation::isNMC --> father mapping undefined" );
    return !cell->isARootCell() && *fatherCell <= *( cell->fatherCell() );
  }

  // Downward NMC
  else{
    if( node->isLowestCell( cell ) )
      return false;

    const Cell* cell1 = getMapping( ( (const CladeNode*)(node) )->child( LEFT ) );
    const Cell* cell2 = getMapping( ( (const CladeNode*)(node) )->child( RIGHT ) );
    const Cell* nextCell = node->getCellDownward( cell );

    return ( *nextCell <= *cell1 && *nextCell <= *cell2 );
  }
}

void Reconciliation::applyNMC( const Slide& slide , double likSum , State* state ){

  // Define likelihood value
  if( state != 0 ){
    state->performSlide( slide , true );
    _valuationValues.setValue( Lik      , exp( state->logL() ) );
    if( likSum > 0 )
      _valuationValues.setValue( PostProb , costValuation( Lik ) / likSum );
  }

  const Node* node = geneNode( slide.nodeId() );
  DIRECTION dir = slide.dir();

  const Cell* cell1 = getMapping( ( (const CladeNode*)(node) )->child( LEFT ) );
  const Cell* cell2 = getMapping( ( (const CladeNode*)(node) )->child( RIGHT ) );

  // Substract cost
  _valuationValues.sub( Lost , distCell( getMapping( node ) , cell1 ) + distCell( getMapping( node ) , cell2 ) );
  if( !node->isRootNode() )
    _valuationValues.sub( Lost , distCell( getMapping( node->father() )  , getMapping( node ) ) );

  // Move Upward
  if( dir == UP )
    if( getMapping( node )->isANode() )
      _valuationValues.add( Dup );

  _mappingVector[ node->nodeId() ] =  const_cast<Cell*> ( node->getDirCell( dir , getMapping( node ) ) );

  // Move Downward
  if( dir == DOWN )
    if( getMapping( node )->isANode() )
      _valuationValues.sub( Dup );

  // Add cost
  _valuationValues.add( Lost , distCell( getMapping( node ) , cell1 ) + distCell( getMapping( node ) , cell2 ) );
  if( !node->isRootNode() )
    _valuationValues.add( Lost , distCell( getMapping( node->father() )  , getMapping( node ) ) );
}

void Reconciliation::compute_uNMC( list<const CladeNode*>& slideList ) const{
  const list<Node*>& prefixList = _treePair.geneTree().prefixList();
  for( list<Node*>::const_iterator node = prefixList.begin(); node != prefixList.end(); node++ ){
    if( isNMC( Slide( (**node).nodeId() , UP ) ) )
      slideList.push_back( (const CladeNode*) *node );
  }
}

///////////////////////////
  // Reconciliation Vector
///////////////////////////

vector<uint> Reconciliation::recVector() const{
  vector<uint> outputVector( size() , 0 );
  for( uint i = 0; i < size(); i++ )
    outputVector[i] = getMapping(i)->layerId();
  return outputVector;
}

bool Reconciliation::operator == ( const StateInfo& stateInfo ) const{
  return recVector() == stateInfo.recVector();
}

///////////////////////////
// Shortcut methods to const Attributes
///////////////////////////

const CladeNode* Reconciliation::geneTreeRoot() const{ 
  return _treePair.geneTree().root(); 
}

Cell*  Reconciliation::getCell( const Node* node ) const{ 
  return _treePair.getCell( node ); 
}

const Node*  Reconciliation::mappingFunction( const Node* node ) const{ 
  return _treePair.mappingFunction( node ); 
}

///////////////////////////
// Output methods
///////////////////////////

void Reconciliation::printMappingFct( std::ostream& output ) const{
  output<<"\nGene\tSpecies";

  //  for( uint i = 0; i < size(); i++ )
  const list<Node*>& prefixList = _treePair.geneTree().prefixList();
  for( list<Node*>::const_iterator node = prefixList.begin(); node != prefixList.end(); node++ ){
    uint i = (**node).nodeId();
    if( getMapping(i) != 0 && !geneNode(i)->isLeafNode() ){
      output<<"\n" << geneNode(i)->prefixId();
      output<<"\t" << getMapping(i)->toString();
      if( !geneNode(i)->isRootNode() )
	output<<"\t" << (geneNode(i)->isLeftChild()? "L": "R" );
    }
  }
}

string valuation2String( uint val ){
  if( val == 0 )
    return "Dup";

  if( val == 1 )
    return "Lost";

  if( val == 2 )
    return "Mut";

  if( val == 3 )
    return "Lik";

  if( val == 4 )
    return "PostProb";

  Exception( true , "Types::valuationToString ----> invalid valuation" );
}


std::ostream& operator << ( std::ostream& output , const Reconciliation& reconciliation ){

  reconciliation.printMappingFct( output );

//   for( uint val = 0; val < Types::nbOfValuations(); val++ ){
//     output<<"\n\t";
//     output<< valuation2String( val );
//     output<<" cost : ";
//     output<< reconciliation.costValuation( val );
//   }

//   if( !reconciliation.localMin().empty() ){
//     output<<"\nLocal min: ";
//     for( uint val = 0; val < Types::nbOfValuations(); val++ )
//       if( reconciliation.isLocalMin( val ) )
// 	output<< valuationToString( val ) << " ";
//   }

  output<<endl;
  return output;
}

const string SPECIES_FILE_NAME = ".speciesTreePrime";
void Reconciliation::writeToPrimeFile( const string& geneFileName  ) const{
  Exception( !speciesTree().paramTree().isPrimeFormat() , "Reconciliation::writeToPrimeFile ---> gene and species trees have to be on prime format" );

  ofstream geneStream   ( geneFileName.c_str()      , std::ios::out );
  ofstream speciesStream( SPECIES_FILE_NAME.c_str() , std::ios::out );

  ( const_cast<GeneTree&> ( geneTree() ) ).writeToPrimeFile ( geneStream    , *this );

  speciesStream<<speciesTree().paramTree().treeString1();

  geneStream.close();
  speciesStream.close();

  // primetv needs to be installed on the MacBook
  string command = "primetv -l r -v -i -t -f ps -o rec.ps " + geneFileName + " " + SPECIES_FILE_NAME;
  commandSystem ( command );
  removeFile( geneFileName );
  removeFile( SPECIES_FILE_NAME );
}


///////////////////////////
// Distance methods

uint distNMC( const Reconciliation& rec1 , const Reconciliation& rec2 ){
  Exception( !rec1.isValid() , "distNMC( rec1 , rec2) ---> rec1 is not valid" );
  Exception( !rec2.isValid() , "distNMC( rec1 , rec2) ---> rec2 is not valid" );

  uint total = 0;
  for( uint i = 0; i < rec1.size(); i++ )
    total += rec1.geneNode(i)->distNMC( rec1.getMapping(i) , rec2.getMapping(i) );

  return total;
}

uint nbOfDiffMapping( const Reconciliation& rec1 , const Reconciliation& rec2 ){
  Exception( !rec1.isValid() , "nbOfDiffMapping( rec1 , rec2) ---> rec1 is not valid" );
  Exception( !rec2.isValid() , "nbOfDiffMapping( rec1 , rec2) ---> rec2 is not valid" );

  uint total = 0;
  for( uint i = 0; i < rec1.size(); i++ )
    if( rec1.getMapping( i ) != rec2.getMapping( i ) )
      total++;

  return total;
}

uint nbOfTruePositive_Dup( const Reconciliation& rec , const Reconciliation& realRec ){
  uint count = 0;

  for( uint i = 0; i < realRec.size(); i++ ){
    const Cell* recCell  = rec.getMapping( i );
    const Cell* realCell = realRec.getMapping( i );
    if( !realCell->isANode() && recCell == realCell )
      count++;
  }

  return count;
}

uint nbOfFalsePositive_Dup( const Reconciliation& rec , const Reconciliation& realRec ){
  uint count = 0;

  for( uint i = 0; i < rec.size(); i++ ){
    const Cell* recCell  = rec.getMapping( i );
    const Cell* realCell = realRec.getMapping( i );
    if( !recCell->isANode() && recCell != realCell )
      count++;
  }

  return count;
}
