/***************************************************************************
			Reconciliation.h
		-------------------
	begin	 : Thu Apr 24 13:58:59 EDT 2008
	copyright: (C) 2005 by Universite De Montreal
	email	 :  doyonjea@iro.umontreal.ca
***************************************************************************/
#ifndef RECONCILIATION_H
#define RECONCILIATION_H

//! This class defines the base structure for a Reconciliation object.
/**
*@author Jean-Philippe Doyon
*/

#include "Node.h"
#include "TreePair.h"
#include "SlideClasses.h"
#include "ValuationValues.h"
#include "StateInfo.h"

class State;

class Reconciliation{

  friend std::ostream& operator << ( std::ostream& output , const Reconciliation& reconciliation );

 private:
  const TreePair& _treePair;

  // A reconciliation is defined by M:V(G) ---> V(S) \union E(S)
  // M[i] = cell, where i is the nodeId of u \in V(G)
  // m = |V(S)| , n = |E(S)|
  // Species:      0 <= id < m:
  // Species edge: m <= id < m + n
  vector<Cell*> _mappingVector;

  // Indicates whether or not it is a local min for each (4) cost valuation
  vector<bool> _localMin;

  ValuationValues _valuationValues;

 public:

  // Constructor
  Reconciliation( const TreePair& treePair );
  Reconciliation( const TreePair& treePair , const vector<uint>& inputVector );

  Reconciliation( const Reconciliation& origin );

  Reconciliation( const Reconciliation& origin , const Slide& slide );

  // The Gene tree string contains the mapping for all genes covered by species
  Reconciliation( const TreePair& treePair , const string& geneTreeString );

  // Used in the constructor and after State::randomGeneTree
  void defineFromGeneTreeString( const string& geneTreeString );

  // Used when a swap, between two chains (Metropolis Coupling), is made
  void redefine( const vector<uint>& inputVector );

  // Destructor
  ~Reconciliation(){}

  // Information methods

  uint size() const { return _mappingVector.size(); }

  const TreePair& treePair() const { return _treePair; }

  const vector<Cell*>& mappingVector() const { return _mappingVector; }

  const ValuationValues& valuationValues() const { return _valuationValues; }

  ValuationValues& valuationValuesUpdate() { return _valuationValues; }

  const vector<bool>& localMin() const { return _localMin; }

  uint nbOfDuplications() const { return (int) ( _valuationValues.value( Dup ) ); }
  uint nbOfSpeciations()  const { return _treePair.geneTree().nbOfClades() - nbOfDuplications(); }

  
  ////////////////////////
  // Set Mapping functions

  // Have to be used according to a prefix traversal of G (needed to compute nb of losses)
  void setExtremeMapping( ExtremeType extremeType , const Node* node );

  void moveMappingUpward( const Node* node );
  void clearMapping( const Node* node );

  // Access mapping functions
  const Cell* getMapping( const Node* node ) const;
  const Cell* getMapping( uint i ) const { return _mappingVector[i]; }

  // Covering functions
  bool isCoveredByASpecies( const Node* node ) const;
  const Edge* coveringSpeciesEdge( const Node* node ) const;
  const Node* coveringSpecies( const Node* node ) const;

  bool geneEdgePassThroughSpecies( const Edge* geneEdge , const Node* targetSpecies ) const;
					 

  // Return the number of genes and geneEdges covered by the species node
  uint computeNbOfCovering( const Node* targetSpecies ) const;

  // Maximal value for all x \in V(S)
  uint computeMaximalNbOfCovering() const;

  // Boolean functions
  bool isValid() const;

  // Cost functions
  double costValuation( uint val ) const { return _valuationValues.value( val ); }
  bool isLocalMin   ( uint val ) const { return _localMin[ val ]; }
  uint computeCost( uint val ) const;
  void computeLocalMin( double likSum , State* state = 0 );
  bool localMinComputed() const { return !_localMin.empty(); }


  // NMC operators methods
  bool isNMC( const Slide& slide ) const;
  void applyNMC( const Slide& slide , double likSum , State* state );

  // Compute the list of allowed uNMC
  void compute_uNMC( list<const CladeNode*>& slideList ) const;

  // Reconciliation Vector
  vector<uint> recVector() const;

  ////////////////////////////
  // Shortcut methods to const Attributes
  const CladeNode* geneTreeRoot() const;

  const CladeNode* speciesTreeRoot() const{
    return _treePair.speciesTree().root(); 
  }

  Cell* getCell( const Node* node ) const;
  const Node* mappingFunction( const Node* node ) const;

  const Node* geneNode( uint i ) const { return _treePair.geneTree().getNode(i); }

  const SpeciesTree& speciesTree() const { return _treePair.speciesTree(); }
  const GeneTree& geneTree()       const { return _treePair.geneTree(); }

  bool operator == ( const StateInfo& stateInfo ) const;

  // Output methods
  void writeToPrimeFile( const string& geneFileName ) const;
  void printMappingFct( std::ostream& output ) const;



 private:
  bool ancestorConstraints( const CladeNode* cladeNode , Side side ) const;
  void updateLostCost( const Node* node , OperatorDir opDir );

};

// Distance methods
uint distNMC( const Reconciliation& rec1 , const Reconciliation& rec2 );

// Number of nodes u \in V(G) s.t rec1(u) != rec2(u)
uint nbOfDiffMapping( const Reconciliation& rec1 , const Reconciliation& rec2 );

// Count the number of duplications in RealRec that are in Rec
uint nbOfTruePositive_Dup( const Reconciliation& rec , const Reconciliation& realRec );

// Count the number of duplications in Rec that are not in RealRec
uint nbOfFalsePositive_Dup( const Reconciliation& rec , const Reconciliation& realRec );

#endif
