#include "SlicedSubtree.h"
#include "ArtifGene.h"
#include "RealGene.h"
#include "Tree.h"
#include "Species.h"
#include "MathClass.h"
#include "State.h"

int SlicedSubtree::_classCounter = 0;

std::ostream& operator << (std::ostream& outputStream, const SlicedSubtree& slicedSubtree ){

  outputStream<< "\nSpecies edge" << *( slicedSubtree.speciesEdge()->edge() );
  outputStream<< "\n\tSliced Subtree: ";
  outputStream<< "\n\t\tNb Of Leaves: " << slicedSubtree.nbOfLeaves();

  outputStream<< "\n\t\t" << *( slicedSubtree.rootNode() );

  if( slicedSubtree.emptyTree() )
    return outputStream;

  // Print the sliced subtree by applying a DFS
  const_cast<SlicedSubtree&> (slicedSubtree).restartAll_DFS_Visited();

  list<const Gene*> stackNode;
  stackNode.push_back( slicedSubtree.rootNode() );
  const_cast<Gene*> ( slicedSubtree.rootNode() )->setDFS_Visited( true );

  while( !stackNode.empty() ){
    const Gene* currentNode = stackNode.back();

    if( slicedSubtree.isASlicedLeaf( currentNode ) )
      stackNode.pop_back();
    else{
      const Gene* nextNode = currentNode->nextDFS_SlicedChild( slicedSubtree );

      if( nextNode != 0 ){
	if( currentNode->isFirstSlicedChild( slicedSubtree , nextNode ) )
	  outputStream<<" ( ";
	else
	  outputStream<<" , ";
	
	outputStream<< *nextNode;

	const_cast<Gene*> ( nextNode )->setDFS_Visited( true );
	stackNode.push_back( nextNode );
      }
      else{
	outputStream<<" )";
	stackNode.pop_back();
      }
    }
  }
  return outputStream;
}

SlicedSubtree::SlicedSubtree( const SpeciesEdge* speciesEdge , const Gene* rootNode , bool empty ):
  _rootNode(rootNode) , _speciesEdge(speciesEdge){
  _classCounter++;

  Exception( _rootNode->isLeafNode() , "SlicedSubtree::SlicedSubtree ---> root node not supposed to be a leaf node" );

  // Invalid values
  _old_H_fun   = -1;
  _old_Phi_fun = -1;;

  if( empty ){
    if( rootNode->isArtifNode() )
      _prefixOrderList.push_back( _rootNode );
    return;
  }

  // Build the sliced subtree by applying a DFS 
  restartAll_DFS_Visited();

  list<Gene*> stackNode;
  stackNode.push_back( const_cast<Gene*> (_rootNode) );
  ( const_cast<Gene*> (_rootNode) )->setDFS_Visited( true );

  while( !stackNode.empty() ){
    Gene* currentNode = stackNode.back();

    // Case 1 - Current node is the sliced root or is a clade node covered by the species edge 
    if( isTheSlicedRoot(currentNode) || 
	( !currentNode->isArtifNode() && ( (RealGene*) (currentNode) )->isCoveredBy( _speciesEdge ) ) ){

      Gene* nextNode = const_cast<Gene*> ( currentNode->nextDFS_SlicedChild( *this ) );
      if( nextNode != 0 && !nextNode->isArtifNode() ){
	stackNode.push_back( nextNode );
	const_cast<Gene*> ( nextNode )->setDFS_Visited( true );
      }
      else
	stackNode.pop_back();

      continue;
    }

    // Case 2 - Current node is a leaf node of the sliced subtree
    if( currentNode->isCoveredBy( _speciesEdge->childNode() ) ){
      addLeafGene( currentNode );
      stackNode.pop_back();
      continue;
    }

    // Case 3 - Current node is not covered by the species edge neither by the child species node
    else{
      ArtifGene* artifNode = new ArtifGene( rootNode->state() , ( (RealGene*) (currentNode) )->fatherEdge() ,  (const CladeSpecies*) ( _speciesEdge->childNode() ) );
      artifNode->setDFS_Visited( true ); // So that the nextDFS child will not be this artif gene
      addLeafGene( artifNode );
      stackNode.pop_back();
      continue;
    }
  }
  
  // Post computation to Compute The Prefix Order List
  computeOrderedLists();

  // Set SlicedSubtree / Real Gene relationship (for interal RealGene only)
  for( list<const Gene*>::iterator iterGene = _prefixOrderList.begin(); iterGene != _prefixOrderList.end(); iterGene++ )
    if( !(**iterGene).isArtifNode() && !isASlicedLeaf(*iterGene) && !isTheSlicedRoot(*iterGene) )
      ( (RealGene&) (**iterGene) ).setSlicedSubtree( this );
}

void SlicedSubtree::addLeafGene( Gene* gene ){
  _leafNodeList.push_back( gene );
  gene->setFatherSlicedSubtree( this );
}

void SlicedSubtree::restartAll_DFS_Visited(){
  ( const_cast<GeneTree&> (geneTree()) ).restartAll_DFS_Visited();

  const_cast<Gene&> (*_rootNode).setDFS_Visited( false ); // In case where the root node is an artificial node

  // For all leaf nodes (in case where there is ArtifGene)
  list<Gene*>::const_iterator iterNode;
  for( iterNode = _leafNodeList.begin(); iterNode != _leafNodeList.end(); iterNode++ )
    ( const_cast<Gene&> (**iterNode) ).setDFS_Visited( false );
}

// A sliced subtree is empty only when the root node is artificial and the corresponding edge is not covered by the speciesEdge
bool SlicedSubtree::emptyTree() const{
  if( _rootNode->isArtifNode() && !( (const ArtifGene*) ( _rootNode ) )->geneEdge()->isCoveredBy( _speciesEdge ) )
    return true;
  return false;
}

bool SlicedSubtree::isAncestor( const Gene* gene1 , const Gene* gene2 ) const{
  Exception( !gene1->isArtifNode()     , "SlicedSubtree::isAncestor---> untreated case" );
  Exception( !isTheSlicedRoot( gene1 ) , "SlicedSubtree::isAncestor---> untreated case" );
  Exception( !isASlicedLeaf( gene2 )   , "SlicedSubtree::isAncestor---> untreated case" );

  const ArtifGene* artifGene1 = (const ArtifGene*) (gene1);
  const Node* node1  = artifGene1->geneEdge()->edge()->childNode();
  const Tree& tree   = rootNode()->state().geneTree();

  if( gene2->isArtifNode() ){
    const ArtifGene* artifGene2 = (const ArtifGene*) (gene2);
    if( artifGene1->geneEdge() == artifGene2->geneEdge() )
      return true;
    else{
      const Node* node2  = artifGene2->geneEdge()->edge()->fatherNode();
      return tree.ancestorMatrixValue( node1 , node2 ) || node1 == node2;
    }
  }
  else{
    const Node* node2 = ( (const RealGene*) (gene2) )->node();
    return tree.ancestorMatrixValue( node1 , node2 ) || node1 == node2;
  }
}

const Gene* SlicedSubtree::slicedRootChild() const{
  if( _rootNode->isArtifNode() )
    return ( (const ArtifGene*) (_rootNode) )->slicedChild();
  else{
    const RealGene* rootNode  = (const RealGene*) (_rootNode);
    const GeneEdge* childEdge = rootNode->edge(LEFT)->isCoveredBy( _speciesEdge )? rootNode->edge(LEFT): rootNode->edge(RIGHT);

    if( childEdge->firstArtifGene() != 0 )
      return (const Gene*) ( childEdge->firstArtifGene() );
    else
      return childEdge->childNode();
  }
}

double SlicedSubtree::H_fun( bool recompute ){
  Exception( !recompute && _old_H_fun == -1 , "SlicedSubtree::H_fun ---> anormal case" );

  if( recompute ){
    // COMPUTE ALPHA PART
    // w(T_u) <--- compute the number of automorphique non leaf nodes
    uint w = slicedRootChild()->treeData().nbAuto();

    _old_H_fun = pow( 2 , nbOfLeaves() - 1 - w );

    _old_H_fun *= slicedRootChild()->treeData().betaFct();
    _old_H_fun /= MathClass::factorial( nbOfLeaves() - 1 ); 
  }
  return log( _old_H_fun );
}


double SlicedSubtree::Phi_fun( bool recompute ){
  Exception( !recompute && _old_Phi_fun == -1 , "SlicedSubtree::Phi_fun ---> anormal case" );

  // Pi(T_u) <--- compute the number of automorphique non leaf nodes with distinct leaf label
  if( recompute )
    _old_Phi_fun = (int) ( pow(2, slicedRootChild()->treeData().nbLabel() ) );

  return log( _old_Phi_fun );
}


// Compute all tree string by a posfix order traversal of the sliced subtree
void SlicedSubtree::computeAllTreeData() const{
  for( list<const Gene*>::const_reverse_iterator iterGene = _prefixOrderList.rbegin(); iterGene != _prefixOrderList.rend(); iterGene++ )
    if( !isASlicedLeaf( *iterGene ) && !isTheSlicedRoot( *iterGene ) ){
      RealGene* realGene = (RealGene*) ( const_cast<Gene*> (*iterGene) );
      const TreeData& leftData  = realGene->slicedChild( LEFT  )->treeData();
      const TreeData& rightData = realGene->slicedChild( RIGHT )->treeData();
      realGene->setTreeData( leftData , rightData );
    }
}

void SlicedSubtree::recomputeTreeData( const Gene* firstGene ) const{
  Gene* gene = const_cast<Gene*> (firstGene);
  while( !isTheSlicedRoot( gene ) ){
    RealGene* realGene = (RealGene*) (gene);
//    realGene->unsetTreeData();
    realGene->setTreeData( realGene->slicedChild( LEFT )->treeData() , realGene->slicedChild( RIGHT )->treeData() );
    gene = const_cast<Gene*> ( realGene->slicedFather() );
  }
}


void SlicedSubtree::removeGenes( const RealGene* B , const ArtifGene* A ){
  Exception( A != 0 && _prefixOrderList.front() != A , "SlicedSubtree::removeGene --> the first gene have to be the ArtifGene (Root node)" );
  Exception( A != 0 && _rootNode != A                , "SlicedSubtree::removeGene --> ArtifGene have to be the Root" );

  Exception( A == 0 && _prefixOrderList.front() != B , "SlicedSubtree::removeGene --> the first gene have to be the RealGene (Root node)" );
  Exception( A == 0 && _rootNode != B                , "SlicedSubtree::removeGene --> RealGene have to be the Root" );

  _rootNode = 0;

  // Remove Artif Gene
  if( A != 0 )
    _prefixOrderList.pop_front();

  // Remove Real Gene
  Exception( _prefixOrderList.front() != B , "SlicedSubtree::removeGene --> the second gene have to be the RealGene (Sliced Child of the root node)" );  
  _prefixOrderList.pop_front();

  const_cast<RealGene*> (B)->setSlicedSubtree(0);
}

void SlicedSubtree::replaceGene( RealGene* B , ArtifGene* & A1 , ArtifGene* & A2 ,  ArtifGene* A ){
  if( A == 0 )
    _leafNodeList.remove( B );
  else
    _leafNodeList.remove( A );

  //Create new ArtifGene A1/A2 on both child edge of B
  A1 = new ArtifGene( B->state() , B->edge(LEFT)  ,  (const CladeSpecies*) ( _speciesEdge->childNode() ) );
  A2 = new ArtifGene( B->state() , B->edge(RIGHT) ,  (const CladeSpecies*) ( _speciesEdge->childNode() ) );

  // Add A1/A2 to LeafNode List and set fatherSlicedSubtree relationship
  addLeafGene( A1 );
  addLeafGene( A2 );

  if( A != 0 ){
    // Insert B, A1 and A2 to the PrefixOrderList and remove A
    list< const Gene* >::iterator iterArtif = find( _prefixOrderList.begin() , _prefixOrderList.end() , A );
    Exception( iterArtif == _prefixOrderList.end() , "\n Replace Gene" );
    _prefixOrderList.insert( iterArtif , B );
    _prefixOrderList.insert( iterArtif , A1 );
    _prefixOrderList.insert( iterArtif , A2 );
    _prefixOrderList.erase( iterArtif );
  }
  else{
    // Insert A1 and A2 to the PrefixOrderList
    list< const Gene* >::iterator iterReal = find( _prefixOrderList.begin() , _prefixOrderList.end() , B );
    Exception( iterReal == _prefixOrderList.end() , "\n Replace Gene" );
    iterReal++;
    _prefixOrderList.insert( iterReal , A1 );
    _prefixOrderList.insert( iterReal , A2 );
  }

  B->setSlicedSubtree(this);
}

void SlicedSubtree::replaceGene( ArtifGene* A1 , ArtifGene* A2 , RealGene* B , ArtifGene* & A ){
  Exception( _leafNodeList.empty() , "SlicedSubtree::replaceGene ---> leaf node list is empty" );
  Exception( find( _leafNodeList.begin() , _leafNodeList.end() , A1 ) == _leafNodeList.end() , "SlicedSubtree::replaceGene ---> A1 is not found in leafNodeList" );
  Exception( find( _leafNodeList.begin() , _leafNodeList.end() , A2 ) == _leafNodeList.end() , "SlicedSubtree::replaceGene ---> A2 is not found in leafNodeList" );

  _leafNodeList.remove( A1 );
  _leafNodeList.remove( A2 );

  if( A != 0 ){
    //Create new ArtifGene A on father edge of B
    A = new ArtifGene( B->state() , B->fatherEdge()  ,  (const CladeSpecies*) ( _speciesEdge->childNode() ) );
    // Add A to LeafNode List and set fatherSlicedSubtree relationship
    addLeafGene( A );
  }
  else
    addLeafGene( B );
 

  if( A != 0 ){
    // Insert A to the PrefixOrderList and remove A1/A2 and B
    list< const Gene* >::iterator iterArtif = find( _prefixOrderList.begin() , _prefixOrderList.end() , A1 );
    Exception( iterArtif == _prefixOrderList.end() , "\n Replace Gene" );
    _prefixOrderList.insert( iterArtif , A );
    _prefixOrderList.remove( B );
  }

  _prefixOrderList.remove( A1 );
  _prefixOrderList.remove( A2 );

//  B->setSlicedSubtree(0);
}

// IMPORTANT NOTE: rootNode of both tree had already been added to the "Prefix Order List"
// in the Constructor for Empty tree
void SlicedSubtree::splitIntoTwoSlicedSubtree( SlicedSubtree& tree1 , SlicedSubtree& tree2 ){
  // Set child of Artif Nodes A1 and A2: B1 and B2
  const Gene* gene1 = ( (const ArtifGene*) (tree1.rootNode()) )->slicedChild();
  const Gene* gene2 = ( (const ArtifGene*) (tree2.rootNode()) )->slicedChild();

  list< const Gene* >& list1 = const_cast< list< const Gene* >& > ( tree1.prefixOrderList() );
  list< const Gene* >& list2 = const_cast< list< const Gene* >& > ( tree2.prefixOrderList() );


  Exception( _prefixOrderList.front() != gene1 && _prefixOrderList.front() != gene2 ,
	     "SlicedSubtree::splitIntoTwoSlicedSubtree ---> Anormal situation A: " + gene1->geneString() + " " + gene2->geneString() );

  // Split list L into L1 and L2
  if( _prefixOrderList.front() == gene1 ){
    list< const Gene* >::iterator it2 = find( _prefixOrderList.begin() , _prefixOrderList.end() , gene2 );
    list1.splice( list1.end() , _prefixOrderList , _prefixOrderList.begin() , it2 );
  }
  else{
    list< const Gene* >::iterator it1 = find( _prefixOrderList.begin() , _prefixOrderList.end() , gene1 );
    list1.splice( list1.end() , _prefixOrderList , it1 ,  _prefixOrderList.end() );
  }

  list2.splice( list2.end() , _prefixOrderList );

  // Insert each LeafNode of T to the corresponding SlicedSubtree (T1 or T2)
  for( list<Gene*>::iterator node = _leafNodeList.begin(); node != _leafNodeList.end(); node++ ){
    if( tree1.isAncestor( tree1.rootNode() , *node ) )
      tree1.addLeafGene( *node );
    else{
      Exception( !tree2.isAncestor( tree2.rootNode() , *node ) , "SlicedSubtree::splitIntoTwoSlicedSubtree ---> Anormal situation B" );
      tree2.addLeafGene( *node );
    }
  }

  tree1.updateGenesRelationships();
  tree2.updateGenesRelationships();
}

void SlicedSubtree::unSplitIntoOneSlicedSubtree ( SlicedSubtree& tree1 , SlicedSubtree& tree2 ){
  Exception( !rootNode()->isArtifNode()                                        , "SlicedSubtree::unSplitIntoOneSlicedSubtree ---> Anormal Case A" );
  Exception( ( (const ArtifGene*) (rootNode()) )->slicedChild()->isArtifNode() , "SlicedSubtree::unSplitIntoOneSlicedSubtree ---> Anormal Case B" );

  // Set child of Artif Nodes A1 and A2: B1 and B2
  const RealGene* rootChild = (const RealGene*) ( ( (const ArtifGene*) (rootNode()) )->slicedChild() );
  const Gene* gene1 = rootChild->slicedChild( LEFT );
  const Gene* gene2 = rootChild->slicedChild( RIGHT );

  list< const Gene* >& list1 = const_cast< list< const Gene* >& > ( tree1.prefixOrderList() );
  list< const Gene* >& list2 = const_cast< list< const Gene* >& > ( tree2.prefixOrderList() );

  Exception( list1.empty() , "SlicedSubtree::unSplitIntoOneSlicedSubtree ---> Anormal Case C" );
  Exception( list2.empty() , "SlicedSubtree::unSplitIntoOneSlicedSubtree ---> Anormal Case D" );

  Exception( list1.front() != gene1 , "SlicedSubtree::unSplitIntoOneSlicedSubtree ---> Anormal situation E" );
  Exception( list2.front() != gene2 , "SlicedSubtree::unSplitIntoOneSlicedSubtree ---> Anormal situation F" );

  // Split lists L1 and L2 into L
  _prefixOrderList.splice( _prefixOrderList.end() , list1 );
  _prefixOrderList.splice( _prefixOrderList.end() , list2 );

  // Insert each LeafNode of T1 and T2 to T
  _leafNodeList.splice( _leafNodeList.end() , const_cast< list<Gene*>& > ( tree1.leafNodeList() ) );
  _leafNodeList.splice( _leafNodeList.end() , const_cast< list<Gene*>& > ( tree2.leafNodeList() ) );

  updateGenesRelationships();
}

// In case where root is defined, it have to be added at the front of the prefix orderlist
void SlicedSubtree::splitIntoOneSlicedSubtree( SlicedSubtree& tree , const RealGene* root ){
  if( root != 0 )
    _prefixOrderList.push_front( root );

  _prefixOrderList.splice( _prefixOrderList.end() , const_cast< list< const Gene* >& > ( tree.prefixOrderList() ) );
  _leafNodeList.splice( _leafNodeList.end() , const_cast< list<Gene*>& > ( tree.leafNodeList() ) );

  updateGenesRelationships();
}

void SlicedSubtree::updateGenesRelationships(){
  for( list<const Gene*>::iterator gene = _prefixOrderList.begin(); gene != _prefixOrderList.end(); gene++ ){
    if( isASlicedLeaf( *gene ) )
      const_cast<Gene&> ( (**gene) ).setFatherSlicedSubtree( this );
    else
      if( !isTheSlicedRoot( *gene ) ){
	RealGene* realGene = const_cast<RealGene*> ( (const RealGene*) ( *gene ) );
	realGene->setSlicedSubtree( this );
      }
  }
}



void SlicedSubtree::checkConsistancy( ostream& output ) const{
  string S  = "\n\tRoot node: ";
  S += _rootNode->geneString();
  S += "\n\tSpecies edge: ";
  S += _speciesEdge->toString();
  
  Exception( _leafNodeList.empty()    , "SlicedSubtree::checkConsistancy ---> leaf node list is empty" + S);
  Exception( _prefixOrderList.empty() , "SlicedSubtree::checkConsistancy ---> prefix order list is empty"  + S);

  Exception( _prefixOrderList.front() != _rootNode , "SlicedSubtree::checkConsistancy ---> first node have to be the root"  + S );

  for( list<const Gene*>::const_iterator iterGene = _prefixOrderList.begin(); iterGene != _prefixOrderList.end(); iterGene++ )
    if( isASlicedLeaf( *iterGene ) )
      Exception( find( _leafNodeList.begin() , _leafNodeList.end() , *iterGene ) == _leafNodeList.end() , "SlicedSubtree::checkConsistancy ---> Gene " + (**iterGene).geneString() + " is not in leaf node list");
}

// The computation of the Prefix Order List is done with
// a Depth First Search traversal
void SlicedSubtree::computeOrderedLists(){
  restartAll_DFS_Visited();

  list<const Gene*> stackNode;
  stackNode.push_back( _rootNode );
  _prefixOrderList.push_back( _rootNode );

  while( !stackNode.empty() ){
    const Gene* currentNode = stackNode.back();
    const Gene* nextNode    = currentNode->nextDFS_SlicedChild( *this );

    if( nextNode != 0 ){
      ( const_cast<Gene*> (nextNode) )->setDFS_Visited( true );
      _prefixOrderList.push_back( nextNode );

      if( !isASlicedLeaf(nextNode) )
	stackNode.push_back( nextNode );
    }
    else
      stackNode.pop_back();
  }
}

void SlicedSubtree::printLeaves( std::ostream& output ) const{
  for( list<Gene*>::const_iterator node = _leafNodeList.begin(); node != _leafNodeList.end(); node++ )
    output<<(**node).geneString() << " ";
}

void SlicedSubtree::printPrefixOrderList( std::ostream& output ) const{
  output<<"\nOrdered list:";
  for( list<const Gene*>::const_iterator gene = _prefixOrderList.begin(); gene != _prefixOrderList.end(); gene++ )
    output<< "\n\t" << **gene;
}
