/***************************************************************************
			SlicedSubtree.h
		-------------------
	begin	 : Thu Oct 26 17:57:09 EDT 2006
	copyright: (C) 2005 by Universite De Montreal
	email	 :  doyonjea@iro.umontreal.ca
***************************************************************************/
#ifndef SLICEDSUBTREE_H
#define SLICEDSUBTREE_H

//! This class defines the base structure for a SlicedSubtree object.
/**
*@author Jean-Philippe Doyon
*/

#include "RealGene.h"
#include "SpeciesEdge.h"
#include "TreePair.h"

class ArtifGene;

class SlicedSubtree{
  
  friend std::ostream& operator << (std::ostream& outputStream, const SlicedSubtree& slicedSubtree );

 private:
  static int _classCounter;

  // Either a RealGene or an ArtifGene
  const Gene* _rootNode; 

  // The species edge on which the sliced subtree is
  const SpeciesEdge* _speciesEdge; 

  // List of Gene (either RealGene or ArtifGene)
  list< Gene* > _leafNodeList; 

  // Ordered lists of node with regard to a
  //    Prefix order : VLR or VRL
  list< const Gene* > _prefixOrderList;

  // Previous values before the modification of the sliced subtree caused by a NNI
  double _old_H_fun;
  int _old_Phi_fun;

 public:

  static int classCounter(){ return _classCounter; };

  // Create all the artificial nodes
  // Fill the leafNodeList with the leaf node of the sliced subtree
  SlicedSubtree( const SpeciesEdge* speciesEdge , const Gene* rootNode , bool empty = false );

  // All ArtifGene created are delete in State Destructor
  ~SlicedSubtree(){ _classCounter--; }

  // Shortcut methods
  const GeneTree& geneTree() { return _speciesEdge->edge()->fatherNode()->tree().treePair().geneTree(); }

  // Access methods
  const SpeciesEdge* speciesEdge()  const { return _speciesEdge; }
  const Gene*    rootNode()         const { return _rootNode; }
  int nbOfLeaves()                  const { return _leafNodeList.size(); }
  const list<Gene*>& leafNodeList() const { return _leafNodeList; }

  const list< const Gene* >& prefixOrderList() const { return _prefixOrderList; }

  void setRootNode( const Gene* gene ){ _rootNode = gene; }

  // Boolean methods
  bool isASlicedLeaf( const Gene* gene )   const { return gene->isCoveredBy( _speciesEdge->childNode() ); }
  bool isTheSlicedRoot( const Gene* gene ) const { return gene->isCoveredBy( _speciesEdge->fatherNode() ); }

  bool emptyTree() const;

  bool isAncestor( const Gene* gene1 , const Gene* gene2 ) const;

  // Sliced methods
  const Gene* slicedRootChild() const;
  void addLeafGene( Gene* gene );
  void removeGenes( const RealGene* realGene , const ArtifGene* artifGene = 0 );

  void removeRoot(){ _prefixOrderList.pop_front(); }
  
  // DFS methods
  void restartAll_DFS_Visited();

  // Likelihood methods
  double H_fun  ( bool recompute );
  double Phi_fun( bool recompute );
  void computeAllTreeData() const;
  void recomputeTreeData( const Gene* firstGene ) const;

  ///////////////////////////////
  // Methods used by NNI operators

  // For dup2UpDup and spec2Dup operators
  void replaceGene( RealGene* realGene , ArtifGene* & artifGene1 , ArtifGene* & artifGene2 , ArtifGene* artifGene = 0 );

  // For dup2DownDup and dup2Spec operators
  void replaceGene( ArtifGene* artifGene1 , ArtifGene* artifGene2 , RealGene* realGene , ArtifGene* & artifGene );

  void splitIntoTwoSlicedSubtree   ( SlicedSubtree& tree1 , SlicedSubtree& tree2 );
  void unSplitIntoOneSlicedSubtree ( SlicedSubtree& tree1 , SlicedSubtree& tree2 );
  void splitIntoOneSlicedSubtree   ( SlicedSubtree& tree , const RealGene* root = 0 );

  // Debug methods
  void checkConsistancy( ostream& output ) const;

  // Output leaves
  void printLeaves( std::ostream& output ) const;
  void printPrefixOrderList( std::ostream& output ) const;

 private:
  void computeOrderedLists();

  // Define the relationships:
  // - leaf gene / father sliced subtree
  // - internal gene / slicedSubtree
  void updateGenesRelationships();
};

#endif
