#include "SlideSet.h"
#include "MathClass.h"

void SlideSet::initialize( const State& state ){
  // The maximal number of possible slides is equal to nbOfClades * 2
  _possibleSlides.resize ( state.geneTree().nbOfClades() * 2 );

  // One SlidePos / Node (LeafNode SlidePos has -1 values)
  _slidesPositions.resize( state.geneTree().nbOfNodes() );

  uint i = 0;

  // Traverse the Vector of RealGene : u in V(G)/L(G)
  for( vector<RealGene*>::const_iterator realGene = state.geneVector().begin(); realGene != state.geneVector().end(); realGene++ ){
    SlidePos& slidePos = _slidesPositions[ (**realGene).nodeId() ];

    if( (**realGene).isRootNode() || (**realGene).isLeafNode() )
      continue;

    // u is Covered by a Species (only Slide Up is possible)
    if( (**realGene).isCoveredByASpecies()  ){
      slidePos.setPosition( UP , i );
      _possibleSlides[ i++ ].setValues( (**realGene).nodeId() , UP );
    }
    // u is Covered by an Edge (both slides could be possible)
    else{
      // Father of u is an ArtifGene
      if( (**realGene).slicedFather()->isArtifNode() ){
	slidePos.setPosition( UP , i );
	_possibleSlides[ i++ ].setValues( (**realGene).nodeId() , UP );
      }
      // u1 and u2 are both ArtifGene
      if( (**realGene).slicedChild( LEFT )->isArtifNode() && (**realGene).slicedChild( RIGHT )->isArtifNode() ){
	slidePos.setPosition( DOWN , i );
	_possibleSlides[ i++ ].setValues( (**realGene).nodeId() , DOWN );
      }
    }
  }
  _size = i;
}

void SlideSet::addSlide( const RealGene* realGene , DIRECTION dir ){
  if( !containsSlide( realGene , dir ) ){
    SlidePos& slidePos = _slidesPositions[ realGene->nodeId() ];
    slidePos.setPosition( dir , _size );
    _possibleSlides[ _size ].setValues( realGene->nodeId() , dir );
    _size++;
  }
}

void SlideSet::removeSlide( const RealGene* realGene , DIRECTION dir ){
  Exception( !containsSlide( realGene , dir ) , "SlideSet::removeSlide --> slide not in set" );
  SlidePos& slidePos = _slidesPositions[ realGene->nodeId() ];
  int pos = slidePos.position( dir );

  slidePos.unsetPosition( dir );
  _possibleSlides[ pos ].unsetValues();
  if( pos != _size - 1 ){
    _possibleSlides[ pos ] = _possibleSlides[ _size - 1 ];
    SlidePos& newSlidePos = _slidesPositions[ _possibleSlides[ pos ].nodeId() ];
    newSlidePos.setPosition( _possibleSlides[ pos ].dir() , pos );
  }
    
  _size--;
}

bool SlideSet::containsSlide( const RealGene* realGene , DIRECTION dir ) const{
  const SlidePos& slidePos = _slidesPositions[ realGene->nodeId() ];
  return slidePos.isDefined( dir );
}


Slide SlideSet::randomSlide( const Slide& last , const State& state ) const{
  Exception( isEmpty() , "SlideSet::randomSlide ---> is empty" );

  if( size() == 1 )
    return _possibleSlides[ 0 ];
  else
    return _possibleSlides[ MathClass::random( 0 , size() - 1 ) ];
}

  /////////////////////////////////////////
  // UPDATE THE SLIDE SET
  // B is the RealGene being slided
  // S is the species such that M(B) = S
void SlideSet::update( const Slide& next , const State& state ){
  const RealGene* B = state.getRealGene( next.nodeId() );
  const Species*  S = state.mappingFunction( B );

  ////////////////////////////////////////////////
  // SPEC TO DUP : B is covered by a Species (UP)
  // DUP  TO SPEC: B is covered by an edge 12 and dir = DOWN and M(b) = 2
  if( B->isCoveredByASpecies() || ( B->coveringSpeciesEdge()->childNode() == S &&  next.dir() == DOWN ) )
    specDup( next , state );
  else
    dupDup( next , state );
}

void SlideSet::specDup( const Slide& next , const State& state ){
  const RealGene* B = state.getRealGene( next.nodeId() );
  const Species*  S = state.mappingFunction( B );
  const RealGene* B1 = B->child( LEFT );
  const RealGene* B2 = B->child( RIGHT );

  // Next dir = UP
  if( next.dir() == UP ){
    if( !B1->isCoveredByASpecies() && B1->coveringSpeciesEdge()->fatherNode() == S )
      addSlide( B1 , UP );

    if( !B2->isCoveredByASpecies() && B2->coveringSpeciesEdge()->fatherNode() == S )
      addSlide( B2 , UP );

    if( !B->slicedFather()->isArtifNode() )
      removeSlide( B , UP );

    addSlide( B , DOWN );
  }
  // Next dir = DOWN
  else{
    if( !B1->isCoveredByASpecies() && B1->coveringSpeciesEdge()->fatherNode() == S )
      removeSlide( B1 , UP );

    if( !B2->isCoveredByASpecies() && B2->coveringSpeciesEdge()->fatherNode() == S )
      removeSlide( B2 , UP );

    if( !B->slicedFather()->isArtifNode() )
      addSlide( B , UP );

    removeSlide( B , DOWN );
  }
}

void SlideSet::dupDup ( const Slide& next , const State& state ){
  const RealGene* B = state.getRealGene( next.nodeId() );
  const RealGene* B1 = B->child( LEFT );
  const RealGene* B2 = B->child( RIGHT );
  const RealGene* F  = (const RealGene*) ( B->father() );
  const SpeciesEdge* edge = B->coveringSpeciesEdge();

  // DUP 2 UP DUP
  // B is covered by an Edge 23 and dir = UP
  if( next.dir() == UP ){

    if( !B1->isCoveredByASpecies() && B1->coveringSpeciesEdge() == edge )
      addSlide( B1 , UP );

    if( !B2->isCoveredByASpecies() && B2->coveringSpeciesEdge() == edge )
      addSlide( B2 , UP );
    
    if( F->isCoveredByASpecies() && F->coveringSpecies() == edge->fatherEdge()->fatherNode() )
      removeSlide( B, UP );

    if( !F->isCoveredByASpecies() && F->coveringSpeciesEdge()->isFatherEdge( edge ) ){
      removeSlide( B, UP );
      if( containsSlide( F, DOWN ) )
	removeSlide( F , DOWN );
    }
    
    if( !containsSlide( B , DOWN ) )
      addSlide( B , DOWN );
  }

  // DUP 2 DOWN DUP
  // B is covered by an Edge 12 and dir = DOWN and M(b) \neq 2
  if( next.dir() == DOWN ){
    if( !B1->isCoveredByASpecies() && edge->isFatherEdge( B1->coveringSpeciesEdge() ) )
      removeSlide( B1 , UP );

    if( !B2->isCoveredByASpecies() && edge->isFatherEdge( B2->coveringSpeciesEdge() ) )
      removeSlide( B2 , UP );

    Side childSide = B->isLeftChild()? RIGHT: LEFT;
    if( !F->isCoveredByASpecies() && F->coveringSpeciesEdge() == edge && !F->isBlockedBy( childSide , DOWN ) )
      addSlide( F , DOWN );

    if( !containsSlide( B , UP ) )
      addSlide( B , UP );

    // Remove B-down slide if needed
    const SpeciesEdge* childEdge = B->coveringChildSpeciesEdge();

    if( ( (const Gene*)(B1) )->isCoveredBy( childEdge->childNode() ) || ( (const Gene*)(B2) )->isCoveredBy( childEdge->childNode() ) ||
	B1->isCoveredBy( childEdge ) || B2->isCoveredBy( childEdge ) )
      removeSlide( B , DOWN );
  }
}


std::ostream& operator << ( std::ostream& output , const SlideSet& slideSet ){
  output<<"\nSet size: " << slideSet.size();
  for( int i = 0; i < slideSet.size(); i++ ){
    const Slide& slide = slideSet.possibleSlides()[i];
    output<<"\n\t" << slide;
  }
  
  return output;
}
