#include "SpeciesEdge.h"
#include "CladeSpecies.h"
#include "State.h"
#include "MathClass.h"
#include "OrthoParam.h"

int SpeciesEdge::_classCounter = 0;


std::ostream& operator << (std::ostream& output, const SpeciesEdge& speciesEdge){
  output<< *( speciesEdge.edge() );

  // Print covered gene
  output<<": ";
  for( list< const RealGene* >::const_iterator iterGene = speciesEdge.geneList().begin(); iterGene != speciesEdge.geneList().end(); iterGene++ ){
    output<<"\n\t" << **iterGene;
    output<<"\"" << (**iterGene).treeData().String() <<"\"";
  }

//   output <<"\n\tlambda\t= " << speciesEdge.lambda();
//   output <<"\n\tmu\t= "     << speciesEdge.mu();
//   output <<"\n\tlength\t= "   << speciesEdge.length();

//   output <<"\n\tpart C\t= "   << speciesEdge.partC();
//   output <<"\n\tmu_t\t= "     << speciesEdge.mu_t();
//   output <<"\n\tpart A\t= "   << speciesEdge.partA();

  
//   output << "\n\nP fun dist:\n" << speciesEdge.P_Fun_Dist();

//   output << "\n\nP fun value:\n";
//   speciesEdge.print_P_Fun_Value( output , 10 );

  return output;
}


SpeciesEdge::SpeciesEdge( const State& state , const Edge* edge, const EdgeValues& edgeValues ):
  _state( state ) , _edge( edge ){
  _classCounter++;

  _lambda = edgeValues.lambda();
  _mu     = edgeValues.mu();
  _length = edgeValues.length();

  // Case 1
  if( _lambda > _mu ){
    double partC = log( 1 - exp( log(_mu) - _length*(_lambda - _mu) - log(_lambda) ) );

    _mu_t = log( 1 - exp( -( _length*(_lambda - _mu) ) ) ) - partC;

    _P_0_t = log(_lambda - _mu) - log(_lambda) - partC;
  }

  // Case 2
  if( _lambda < _mu ){
    _mu_t  = log( _lambda * exp( _length * (_mu - _lambda) ) - _lambda );
    _mu_t -= log( _mu     * exp( _length * (_mu - _lambda) ) - _lambda );

    _P_0_t  = log( _mu - _lambda ); 
    _P_0_t -= log( _mu    * exp( _length * (_mu - _lambda) ) - _lambda );
  }

  // Case 3
  if( _lambda == _mu ){
    _mu_t  = log( _lambda ) + log ( _length ) - log( 1 + ( _lambda * _length ) );

    _P_0_t = - log( 1 +  ( _lambda * _length ) );
  }

  // Compute part A
  _partA = log( 1 - exp(_mu_t) );
}


const Species* SpeciesEdge::childNode() const{ 
  return _state.getSpecies( _edge->childNode() );  
}

const Species* SpeciesEdge::fatherNode() const{ 
  return _state.getSpecies( _edge->fatherNode() ); 
}

const SpeciesEdge* SpeciesEdge::fatherEdge() const{
  Exception( isRootEdge() , "SpeciesEdge::fatherEdge() ---> is Root Edge" );
  return fatherNode()->fatherEdge();
}

const SpeciesEdge* SpeciesEdge::childEdge( Side s ) const{
  Exception( isLeafEdge() , "SpeciesEdge::childEdge --> is leaf edge" );
  return ( (const CladeSpecies*) (childNode()) )->edge(s);
}


bool SpeciesEdge::isLeftEdge() const{
  return ( (const CladeSpecies*) ( fatherNode() ) )->edge(LEFT) == this; 
}

bool SpeciesEdge::isLeafEdge() const{
  return childNode()->isLeafNode();
}


void SpeciesEdge::removeCoveredGene( const RealGene* realGene ){ 
  _geneList.remove( realGene ); 
}

double SpeciesEdge::P_Fun( uint l ) const{
  // y in L(S)
  if( !childNode()->isCladeSpecies() )
    return P_n_t_fun( l );

  // y notin L(S) and l=0
  if( l == 0 ){
    double valueA = P_n_t_fun( 0 );
    double valueB = log( 1 + exp( _P_0_t + _partA + ( (const CladeSpecies*) ( childNode() ) )->Dy() - _partB - valueA ) );
    return valueA + valueB;
  }

  // y notin L(S) and l \neq 0
  double value = _P_0_t + _partA +  _mu_t * (l - 1) - _partB * (l + 1);

  return value;
}

void SpeciesEdge::compute_B_part(){
  _partB = log( 1 - exp( _mu_t + ( (const CladeSpecies*) ( childNode() ) )->Dy() ) );
}

double SpeciesEdge::P_n_t_fun( uint n ) const {
  if( n == 0 && _lambda == _mu )
    return log( _lambda ) + log( _length ) - log( 1 + (_lambda * _length ) );

  if( n == 0 && _lambda != _mu )
    return log( 1 - exp( _P_0_t ) );

  if( n != 0 && _lambda == _mu )
    return (n - 1) * ( log( _lambda ) + log( _length ) ) - (n + 1) * log( 1 + (_lambda * _length ) );

  if( n != 0 && _lambda != _mu )
    return _P_0_t + _partA + _mu_t*(n - 1);
}


void SpeciesEdge::compute_P_Fun_Dist(){
  _P_Fun_Dist.resize( DIST_VECTOR_SIZE );

  double total = 0;
  for( uint i = 0; i < _P_Fun_Dist.size(); i++ )
    total += exp( P_Fun(i + 1) );

  _P_Fun_Dist[0] = exp( P_Fun( 1 ) ) / total;
  for( uint i = 1; i < _P_Fun_Dist.size(); i++ )
    _P_Fun_Dist[i] = _P_Fun_Dist[i-1] + ( exp( P_Fun( i + 1 ) ) / total );

  Exception( fabs( 1 - _P_Fun_Dist.back() ) > MIN_PRECISION , "SpeciesEdge::compute_P_Fun_Dist ---> anormal case B" );

  // Find the first value > 0.9
  uint toErase = 0;
  for( uint i = _P_Fun_Dist.size() - 1; i > 0; i-- ){
    if( _P_Fun_Dist[i-1] > 0.9 )
      toErase++;
    else
      break;
  }

  // Remove all the values > 0.9 except the first one
  _P_Fun_Dist.resize( _P_Fun_Dist.size() - toErase + 1 );
  // The last value of the dist has to be 1
  _P_Fun_Dist[ _P_Fun_Dist.size() - 1 ] = 1;
}

uint SpeciesEdge::randomNbOfLeafGermes() const{
  double v = ( (double) MathClass::random( 0 , 99 ) ) / 100.0;

  int n = 0;
  double total = exp( P_n_t_fun( n ) );

  while( v >= total ){
    n++;
    total += exp( P_n_t_fun( n ) );
  }

  return n;
}

bool SpeciesEdge::operator == ( const SpeciesEdge& speciesEdge ) const{
  return *fatherNode() == *( speciesEdge.fatherNode() ) && *childNode() == *( speciesEdge.childNode() );
}

bool SpeciesEdge::operator != ( const SpeciesEdge& speciesEdge ) const{
  return !( *this == speciesEdge );
}

bool SpeciesEdge::operator <  ( const SpeciesEdge& speciesEdge ) const{
  return *fatherNode() < *( speciesEdge.fatherNode() ) && *childNode() < *( speciesEdge.childNode() );
}

bool SpeciesEdge::operator <= ( const SpeciesEdge& speciesEdge ) const{
  return *this < speciesEdge || *this == speciesEdge;
}

bool SpeciesEdge::operator >  ( const SpeciesEdge& speciesEdge ) const{
  return speciesEdge < *this;
}

bool SpeciesEdge::operator >= ( const SpeciesEdge& speciesEdge ) const{
  return speciesEdge <= *this;
}

string SpeciesEdge::toString() const{
  string S = "(" + itos( fatherNode()->nodeId() ) + "," + itos( childNode()->nodeId() ) + ")";
  return S;
}

void SpeciesEdge::ouptut_P_Fun_Dist( std::ostream& output ) const{
  output<<endl;
  for( uint i = 0; i < _P_Fun_Dist.size(); i++ )
    output<<" " << _P_Fun_Dist[i];
}

void SpeciesEdge::print_P_Fun_Value( std::ostream& output , uint max ) const{
  for( uint l = 0; l <= max; l++ )
    output<< setprecision( 2 ) << exp( P_Fun( l ) ) <<",";
}
