/***************************************************************************
			SpeciesEdge.h
		-------------------
	begin	 : Sun Jan  7 18:52:10 EST 2007
	copyright: (C) 2005 by Universite De Montreal
	email	 :  doyonjea@iro.umontreal.ca
***************************************************************************/
#ifndef EDGEVALUES_H
#define EDGEVALUES_H

//! This class defines the base structure for a SpeciesEdge object.
/**
*@author Jean-Philippe Doyon
*/

// A SpeciesEdge object is associated to an Edge from the species Tree

#include "Edge.h"
#include "RealGene.h"
#include "MathClass.h"
#include "BirthDeathParam.h"

class Species;
class State;

class SpeciesEdge{

  friend std::ostream& operator << ( std::ostream& outputStream, const SpeciesEdge& speciesEdge );

 private:
  static int _classCounter;

  const State& _state;
  const Edge* _edge;

  // Edge values
  double _lambda; // Birth rate
  double _mu;      // Death rate
  uint   _length; // UNSIGNED integer: does not allow (-_length) operation
  double _partA;
  double _partB;
  double _mu_t;
  double _P_0_t;

  // Reconciliation attributes
  // Let u be a Gene and <x,y> be the current SpeciesEdge
  //      u in delta(<x,y>) <----> u in List
  list< const RealGene* > _geneList;

  // Vector of double representing the dist of each value of l > 0
  //     - l is represented by the vector position l-1
  //     - distVector[l-1] ... distVector[l] represents the prob. window
  vector<double> _P_Fun_Dist;

 public:

  static int classCounter(){ return _classCounter; }

  // Constructor and Destructor
  SpeciesEdge( const State& state , const Edge* edge , const EdgeValues& edgeValues );

  ~SpeciesEdge(){ _classCounter--; }

  void clearForRedefine(){ _geneList.clear(); }

  // Access methods
  const Edge* edge() const { return _edge; }
  const Species* childNode() const;
  const Species* fatherNode() const;
  const list< const RealGene* >& geneList() const { return _geneList; }

  const SpeciesEdge* fatherEdge() const;
  const SpeciesEdge* childEdge( Side s ) const;

  uint layerId() const{ return _edge->layerId(); }


  // Values access methods
  double mu() const     { return _mu;     }
  double lambda() const { return _lambda; }
  uint   length() const   { return _length; }
  double partA() const  { return _partA;  }
  double partB() const  { return _partB;  }

  double mu_t() const   { return _mu_t;   }
  double P_0_t() const  { return _P_0_t;  }
  const vector<double>& P_Fun_Dist() const { return _P_Fun_Dist; }
			     
  // Boolean methods
  // Note on September 2008:
  //            isRootEdge and isArtificial edge have to be different
  bool isRootEdge()   const { return _edge->isArtificial(); }
  bool isArtificial() const { return _edge->isArtificial(); }

  bool isLeftEdge() const;
  bool isLeafEdge() const;

  Side side() const { return isLeftEdge()? LEFT: RIGHT; }

  // True <---> childNode == speciesEdge.fatherNode
  bool isFatherEdge( const SpeciesEdge* speciesEdge ) const{
    return childNode() == speciesEdge->fatherNode();
  }

  // Reconciliation methods
  void addCoveredGene   ( const RealGene* realGene ) { _geneList.push_back( realGene ); }
  void removeCoveredGene( const RealGene* realGene );

  // Likelihood methods
  // Variable 'l' represents the number of genes
  double P_Fun( uint l ) const;
  void compute_B_part();

  // Gene tree random generation methods
  void compute_P_Fun_Dist();
  uint random_P_Fun_Dist() const{
    return MathClass::randomNbFromDist( _P_Fun_Dist ) + 1;
  }

  // Return a random number of leaf germes (from a starting germe) according to Pn(t)
  uint randomNbOfLeafGermes() const;

  double extinctProb() const { return exp( P_Fun( 0 ) ); }

  // Binary operator
  bool operator == ( const SpeciesEdge& speciesEdge ) const;
  bool operator != ( const SpeciesEdge& speciesEdge ) const;
  bool operator <  ( const SpeciesEdge& speciesEdge ) const;
  bool operator <= ( const SpeciesEdge& speciesEdge ) const;

  bool operator >  ( const SpeciesEdge& speciesEdge ) const;
  bool operator >= ( const SpeciesEdge& speciesEdge ) const;

  bool operator == ( const NodePair& nodePair ) const{ return *_edge == nodePair; }

  string toString() const;

  // Output methods
  void ouptut_P_Fun_Dist( std::ostream& output ) const;
  void print_P_Fun_Value( std::ostream& output , uint max ) const;

 private:
  double P_n_t_fun( uint n ) const;

};

#endif
