#include "State.h"
#include "SlicedSubtree.h"
#include "RealGene.h"
#include "ArtifGene.h"
#include "CladeSpecies.h"
#include "LeafSpecies.h"
#include "toolBox.h"
#include "OrthoParam.h"

#include "TREE_PAIR_CPUtimes.h"

int State::_classCounter = 0;



void State::initializeVectors( const TreePair& treePair , const BirthDeathParam& birthDeathParam ){
  const SpeciesTree& speciesTree = treePair.speciesTree();

  // Define Species vector
  _speciesVector.resize( speciesTree.nbOfNodes() , 0 );
  for( uint i = 0; i < _speciesVector.size(); i++ ){
    const Node* node = speciesTree.getNode(i);
    if( node->isLeafNode() )
      _speciesVector[i] = new LeafSpecies( *this , (LeafNode*) node );
    else
      _speciesVector[i] = new CladeSpecies( *this , (CladeNode*) node );
  }

  // Define SpeciesEdge vector
  _speciesEdgeVector.resize( speciesTree.nbOfEdges()  , 0 );
  for( uint i = 0; i < _speciesEdgeVector.size(); i++ ){
    const EdgeValues& edgeValues = birthDeathParam.edgeValues( speciesTree.getEdge(i)->childNode()->cladeSet() );
    _speciesEdgeVector[i] = new SpeciesEdge( *this , speciesTree.getEdge(i) , edgeValues );
  }
}

State::State( const TreePair& treePair , const vector<uint>& recVector , const BirthDeathParam& birthDeathParam )
  :Reconciliation( treePair , recVector ) {

  _classCounter++;
  initializeVectors( treePair , birthDeathParam );

  compute_All_Dy_and_B_Values();
  
  const GeneTree& geneTree = treePair.geneTree();

  if( !geneTree.empty() ){
    // Define Gene vector
    _geneVector.resize( geneTree.nbOfNodes()  );
    for( uint i = 0; i < _geneVector.size(); i++ )
      _geneVector[ i ] = new RealGene( *this , geneTree.getNode(i) );

    // Define GeneEdge vector
    _geneEdgeVector.resize( geneTree.nbOfEdges() );
    for( uint i = 0; i < _geneEdgeVector.size(); i++ )
      _geneEdgeVector[ i ] = new GeneEdge( *this , geneTree.getEdge(i) );

    // Compute the reconciliation defined by recVector
    computeReconciliation( recVector );

    computeOrthologousStateVector();

    computeAllSlicedSubtrees();
    computeLogL( false );
  }
}

State::State( const TreePair& treePair , const BirthDeathParam& birthDeathParam )
  :Reconciliation( treePair ){
  _classCounter++;
  initializeVectors( treePair , birthDeathParam );

  compute_All_Dy_and_B_Values();
}

State::~State(){
  _classCounter--;

  // Delete Sliced Subtrees and Artif Genes
  deleteSlicedSubtrees();
  deleteArtifGenes();

  // Delete all Species
  for( vector<Species*>::iterator species = _speciesVector.begin(); species != _speciesVector.end(); species++ )
    delete *species;

  // Delete all SpeciesEdge
  for( vector<SpeciesEdge*>::iterator speciesEdge = _speciesEdgeVector.begin(); speciesEdge != _speciesEdgeVector.end(); speciesEdge++ )
    delete *speciesEdge;

  // Delete all RealGene
  for( vector<RealGene*>::iterator realGene = _geneVector.begin(); realGene != _geneVector.end(); realGene++ )
    delete *realGene;

  // Delete all GeneEdge
  for( vector<GeneEdge*>::iterator geneEdge = _geneEdgeVector.begin(); geneEdge != _geneEdgeVector.end(); geneEdge++ )
    delete *geneEdge;
}

void State::deleteSlicedSubtrees(){
  for( vector<Node*>::const_reverse_iterator node = speciesTree().nodeVector().rbegin(); node != speciesTree().nodeVector().rend(); node++)
    if( !(**node).isLeafNode() ){
      const Species* species = getSpecies( *node );

      list< Gene* >::const_iterator gene;
      for( gene = species->geneList().begin(); gene != species->geneList().end(); gene++ )
	(**gene).deleteSlicedSubtrees();
    }
}

void State::deleteArtifGenes(){
  for( vector<Species*>::iterator species = _speciesVector.begin(); species != _speciesVector.end(); species++ )
    for( list< Gene* >::const_iterator gene = (**species).geneList().begin(); gene != (**species).geneList().end(); gene++ )
      if( (**gene).isArtifNode() )
	delete ( const_cast< ArtifGene* > ( (const ArtifGene*) (*gene) ) );
}

void State::redefine( const vector<uint>& recVector ){
  // Delete Sliced Subtrees and Artif Genes
  deleteSlicedSubtrees();
  deleteArtifGenes();

  // Clear attributes of Species, SpeciesEdge, RealGene, and GeneEdge
  for( vector<Species*>::iterator species = _speciesVector.begin(); species != _speciesVector.end(); species++ )
    (**species).clearForRedefine();

  for( vector<SpeciesEdge*>::iterator speciesEdge = _speciesEdgeVector.begin(); speciesEdge != _speciesEdgeVector.end(); speciesEdge++ )
    (**speciesEdge).clearForRedefine();

  for( vector<RealGene*>::iterator realGene = _geneVector.begin(); realGene != _geneVector.end(); realGene++ )
    (**realGene).clearForRedefine();

  for( vector<GeneEdge*>::iterator geneEdge = _geneEdgeVector.begin(); geneEdge != _geneEdgeVector.end(); geneEdge++ )
    (**geneEdge).clearForRedefine();

  Reconciliation::redefine( recVector );

  // Compute the reconciliation defined by recVector
  computeReconciliation( recVector );

  computeOrthologousStateVector();
  computeAllSlicedSubtrees();
  computeLogL( false );
}

bool State::orthologousGenes( const RealGene* gene1 , const RealGene* gene2 ) const{
  Exception( gene1 == gene2               , "State::orthologousGenes ---> gene1 == gene2" );
  Exception( !gene1->node()->isLeafNode() , "State::orthologousGenes ---> gene1 is not a leaf node" );
  Exception( !gene2->node()->isLeafNode() , "State::orthologousGenes ---> gene2 is not a leaf node" );
  return _orthoStateVector[ ( (const LeafNode*) ( gene1->node() ) )->leafId() ] [ ( (const LeafNode*) ( gene2->node() ) )->leafId() ];
}

// Postfix order Species tree traversal
double State::computeLogL( bool clear , bool debug ){
  if( !debug ){
    if( clear )
      clear_PretreatmentFor_H_Phi();
    pretreatmentFor_H_Phi();
  }

  // The eV computations has to be done with a Postfix Order of the Species Tree
  for( vector<Node*>::const_iterator node = speciesTree().nodeVector().begin(); node != speciesTree().nodeVector().end(); node++){
    const Species* species = getSpecies( *node );
    list< Gene* >::const_iterator gene;
    for( gene = species->geneList().begin(); gene != species->geneList().end(); gene++ )
      ( const_cast<Gene&> (**gene) ).compute_eV( debug );
  }

  _logL = getGene( geneTree().root() )->eV();
  return _logL;
}

void State::updateLogL_SpecDup_1( RealGene* B , SpeciesEdge*& edge23 , SpeciesEdge*& edge24 , bool updateLogL ){
  if( !updateLogL )
    return;

  TREE_PAIR_CPUtimes::cpuTimes(UPDATE_LOGL).startClock();

  // Define the Species 2 and SpeciesEdge 12
  const SpeciesEdge* edge12 = B->fatherSlicedSubtree()->speciesEdge();
  CladeSpecies* S2 = const_cast<CladeSpecies*> ( (const CladeSpecies*) ( edge12->childNode() ) );

  // Define edge23 and edge24
  edge23 = const_cast<SpeciesEdge*> ( S2->edge( LEFT ) );
  edge24 = const_cast<SpeciesEdge*> ( S2->edge( RIGHT ) );

  S2->removeTreeData( B );

  TREE_PAIR_CPUtimes::cpuTimes(UPDATE_LOGL).stopClock();
}

void State::updateLogL_SpecDup_2( RealGene* B , SpeciesEdge* edge23 , SpeciesEdge* edge24 , bool updateLogL ){
  try{
    if( !updateLogL )
      return;

    TREE_PAIR_CPUtimes::cpuTimes(UPDATE_LOGL).startClock();

    const SpeciesEdge* edge12 = B->slicedSubtree()->speciesEdge();
    CladeSpecies* S2 = const_cast<CladeSpecies*> ( (const CladeSpecies*) ( edge12->childNode() ) );

    // Define A1/A2
    ArtifGene* A1 = const_cast<ArtifGene*> ( (const ArtifGene*) ( B->slicedChild( LEFT ) ) );
    ArtifGene* A2 = const_cast<ArtifGene*> ( (const ArtifGene*) ( B->slicedChild( RIGHT ) ) );

    if( A1->speciesEdge() != edge23 )
      interchange( A1 , A2 );

    S2->addTreeData( A1 );
    S2->addTreeData( A2 );

    SlicedSubtree& TA1 = const_cast<SlicedSubtree&> ( A1->uniqueSlicedSubtree() );
    SlicedSubtree& TA2 = const_cast<SlicedSubtree&> ( A2->uniqueSlicedSubtree() );

    TA1.H_fun( true );
    TA1.Phi_fun( true );
    TA2.H_fun( true);
    TA2.Phi_fun( true );

    A1->define_eV( B->eA( LEFT )      , edge24->P_Fun( 0 ) );
    A2->define_eV( edge23->P_Fun( 0 ) , B->eA( RIGHT ) );
    constructTreeData_BottomUp ( A1 );
    compute_eV_BottomUp        ( A2 );

    TREE_PAIR_CPUtimes::cpuTimes(UPDATE_LOGL).stopClock();
  }

  catch( SubException subExc ){
    subExc.addFunDesc( "State::updateLogL_SpecDup_2" );
    throw subExc;
  }
}

void State::updateLogL_DupSpec_1( RealGene* B , ArtifGene*& A1 , ArtifGene*& A2 , bool updateLogL ){
  if( !updateLogL )
    return;

  TREE_PAIR_CPUtimes::cpuTimes(UPDATE_LOGL).startClock();

  const SpeciesEdge* edge12 = B->slicedSubtree()->speciesEdge();
  CladeSpecies* S2 = const_cast<CladeSpecies*> ( (const CladeSpecies*) ( edge12->childNode() ) );

  // Define A1/A2
  A1 = const_cast<ArtifGene*> ( (const ArtifGene*) ( B->slicedChild( LEFT ) ) );
  A2 = const_cast<ArtifGene*> ( (const ArtifGene*) ( B->slicedChild( RIGHT ) ) );

  if( A1->speciesEdgeSide() != LEFT )
    interchange( A1 , A2 );

  S2->removeTreeData( A1 );
  S2->removeTreeData( A2 );

  TREE_PAIR_CPUtimes::cpuTimes(UPDATE_LOGL).stopClock();
}

void State::updateLogL_DupSpec_2( RealGene* B , ArtifGene* A1 , ArtifGene* A2 , bool updateLogL ){
  try{
    if( !updateLogL )
      return;

    TREE_PAIR_CPUtimes::cpuTimes(UPDATE_LOGL).startClock();

    const SpeciesEdge* edge12 = B->fatherSlicedSubtree()->speciesEdge();
    CladeSpecies* S2 = const_cast<CladeSpecies*> ( (const CladeSpecies*) ( edge12->childNode() ) );

    S2->addTreeData( B );

    SlicedSubtree& TB1 = const_cast<SlicedSubtree&> ( B->getSlicedSubtree( LEFT ) );
    SlicedSubtree& TB2 = const_cast<SlicedSubtree&> ( B->getSlicedSubtree( RIGHT ) );

    TB1.H_fun( true );
    TB1.Phi_fun( true );
    TB2.H_fun( true);
    TB2.Phi_fun( true );

    B->define_eV( A1->eA( LEFT ) , A2->eA( RIGHT ) );
    constructTreeData_BottomUp ( B );
    compute_eV_BottomUp        ( B );

    TREE_PAIR_CPUtimes::cpuTimes(UPDATE_LOGL).stopClock();
  }

  catch( SubException subExc ){
    subExc.addFunDesc( "State::updateLogL_DupSpec_2" );
    throw subExc;
  }
}

void State::updateLogL_DupDup ( const int sign , DIRECTION dir , RealGene* B , bool updateLogL ){
  try{
    if( !updateLogL )
      return;

    TREE_PAIR_CPUtimes::cpuTimes(UPDATE_LOGL).startClock();

    bool state2 = ( dir == UP && sign == 1 ) || ( dir == DOWN && sign == -1 );

    // Define the SlicedSubtree T_12
    const SlicedSubtree* T12 = ( state2? B->slicedSubtree(): B->slicedFather()->fatherSlicedSubtree() );

    // Define the 3 SpeciesEdge 
    const SpeciesEdge* edge12 = T12->speciesEdge();
    CladeSpecies* S2 = const_cast<CladeSpecies*> ( (const CladeSpecies*) ( edge12->childNode() ) );

    if( state2 ){
      ArtifGene* A1 = (ArtifGene*) ( B->slicedChild( LEFT ) );
      ArtifGene* A2 = (ArtifGene*) ( B->slicedChild( RIGHT ) );

      if( sign == 1 ){
	S2->addTreeData( A1 );
	S2->addTreeData( A2 );

	SlicedSubtree& TA1 = const_cast<SlicedSubtree&> ( A1->uniqueSlicedSubtree() );
	SlicedSubtree& TA2 = const_cast<SlicedSubtree&> ( A2->uniqueSlicedSubtree() );
	TA1.H_fun( true );
	TA2.H_fun( true );
	TA1.Phi_fun( true );
	TA2.Phi_fun( true );

	A1->compute_eV();
	A2->compute_eV();
	constructTreeData_BottomUp( A1 );
	compute_eV_BottomUp       ( A1 );
      }
      else{
	S2->removeTreeData( A1 );
	S2->removeTreeData( A2 );
      }
    }
    else{
      ArtifGene* A = (ArtifGene*) ( B->slicedFather() );
      if( sign == 1 ){
	B->updateTreeData();
	S2->addTreeData( A );

	SlicedSubtree& TA = const_cast<SlicedSubtree&> ( A->uniqueSlicedSubtree() );
	TA.H_fun( true );
	TA.Phi_fun( true );

	A->compute_eV();
	constructTreeData_BottomUp( A );
	compute_eV_BottomUp       ( A );
      }
      else
	S2->removeTreeData( A );
    }

    TREE_PAIR_CPUtimes::cpuTimes(UPDATE_LOGL).stopClock();
  }

  catch( SubException subExc ){
    subExc.addFunDesc( "State::updateLogL_DupDup" );
    throw subExc;
  }
}


//////////////////////
//MCMC Operators

void State::performSlide( const Slide& slide , bool updateLogL ){
  try{
    TREE_PAIR_CPUtimes::cpuTimes(PERFORM_SLIDE).startClock();

    RealGene* B = const_cast<RealGene*> ( getGene( geneTree().getNode( slide.nodeId() ) ) );
    Exception( B->isLeafNode() , "State::slide ---> leaf gene can not be moved" );
    Exception( B->isRootNode() , "State::slide ---> root gene can not be moved (temporary)" );
  
    SlideName slideName = getSlideName( B , slide );

    if( slideName == Spec2Dup ){
      SpeciesEdge *edge23, *edge24;
      updateLogL_SpecDup_1( B , edge23 , edge24 , updateLogL );
      spec2Dup( B );
      updateLogL_SpecDup_2( B , edge23 , edge24 , updateLogL );
      B->deleteSlicedSubtrees();
    }

    if( slideName == Dup2Spec ){
      ArtifGene *A1, *A2;
      updateLogL_DupSpec_1( B , A1 , A2 , updateLogL );
      dup2Spec( B );
      updateLogL_DupSpec_2( B , A1 , A2 , updateLogL );
      A1->deleteSlicedSubtrees();
      A2->deleteSlicedSubtrees();
      delete A1;
      delete A2;
    }

    if( slideName == Dup2UpDup ){
      ArtifGene *A;
      updateLogL_DupDup( -1 , UP , B , updateLogL );
      dup2UpDup( B , A ); 
      updateLogL_DupDup( 1  , UP , B , updateLogL );
      A->deleteSlicedSubtrees();
      delete A;
    }

    if( slideName == Dup2DownDup ){
      ArtifGene *A1, *A2;
      updateLogL_DupDup( -1 , DOWN , B , updateLogL );
      dup2DownDup( B , A1 , A2 ); 
      updateLogL_DupDup( 1  , DOWN , B , updateLogL );
      A1->deleteSlicedSubtrees();
      A2->deleteSlicedSubtrees();
      delete A1;
      delete A2;
    }

    TREE_PAIR_CPUtimes::cpuTimes(PERFORM_SLIDE).stopClock();
  }

  catch( SubException subExc ){
    subExc.addFunDesc( "State::performSlide" );
    cerr<<endl << subExc.message() <<endl;
    abort();
  }
}

SlideName State::getSlideName( RealGene* B , const Slide& slide ) const{
  if( B->isCoveredByASpecies()  && slide.dir() == UP )
    return Spec2Dup;

  if( !B->isCoveredByASpecies() && slide.dir() == DOWN && 
      B->coveringSpeciesEdge()->childNode() == mappingFunction( B ) )
    return Dup2Spec;

  if( !B->isCoveredByASpecies() && slide.dir() == UP )
    return Dup2UpDup;

  if( !B->isCoveredByASpecies() && slide.dir() == DOWN )
    return Dup2DownDup;

  Exception( true , "State::slideName ---> Anormal Case" );
}


// Antecedent:
//      - RealGene B is covered by a Species S2
//      - S2 is not the Root of the Species Tree
// Consequent:
//      - RealGene's speciation is replaced by a duplication on the father Edge (12) of S2

// Created Artif Genes are noted A1 and A2
void State::spec2Dup( RealGene* B ){
  TREE_PAIR_CPUtimes::cpuTimes(MAP2MAP).startClock();

  Exception( !B->isCoveredByASpecies()          , "State::spec2Dup ---> RealGene is covered by a Species" );
  Exception( B->coveringSpecies()->isRootNode() , "State::spec2Dup ---> Covering species edge is a Root Edge of the Species Tree" );

  // Remove B from Species S2 and ad it to edge 12
  B->changeCovering( B->coveringSpecies()->fatherEdge() );

  // Remove B from the SlicedSubtrees T_23 T_24
  const_cast<SlicedSubtree&> ( B->getSlicedSubtree( LEFT ) ).removeGenes( B );
  const_cast<SlicedSubtree&> ( B->getSlicedSubtree( RIGHT ) ).removeGenes( B );

  // New ArtifGenes create below
  ArtifGene* A1;
  ArtifGene* A2;

  // Create A1 and A2 and add it to the Sliced Subtree T_12
  // Redefine the position of B in T_12
  const_cast<SlicedSubtree*> ( B->fatherSlicedSubtree() )->replaceGene( B , A1 , A2 );

  // Create Empty sliced subtrees for A1 and A2
  A1->buildSlicedSubtrees( true );
  A2->buildSlicedSubtrees( true );    

  // Split Sliced subtree T_BL into T_AL (left of A1) 
  // Split Sliced subtree T_BR into T_AR (right of A2)
  SlicedSubtree& leftB   = const_cast<SlicedSubtree&> ( B->getSlicedSubtree  ( LEFT ) );
  SlicedSubtree& rightB  = const_cast<SlicedSubtree&> ( B->getSlicedSubtree  ( RIGHT ) );

  SlicedSubtree& tree1   = const_cast<SlicedSubtree&> ( A1->uniqueSlicedSubtree() );
  SlicedSubtree& tree2   = const_cast<SlicedSubtree&> ( A2->uniqueSlicedSubtree() );

  if( tree1.speciesEdge() == leftB.speciesEdge() ){
    tree1.splitIntoOneSlicedSubtree ( leftB  );
    tree2.splitIntoOneSlicedSubtree ( rightB );
  }
  else{
    tree2.splitIntoOneSlicedSubtree ( leftB  );
    tree1.splitIntoOneSlicedSubtree ( rightB );
  }

  TREE_PAIR_CPUtimes::cpuTimes(MAP2MAP).stopClock();
}

// Antecedent:
//      - RealGene B is covered by a Species Edge 12
//      - Both children of B are ArtifGenes noted A1 and A2 and are covered by species S2
//      - A1 (A2) evolves on edge 23 (24)
//      - The mapping M(B) = S2
// Consequent:
//      - RealGene's duplication is replaced by a speciation on the Species node 2

void State::dup2Spec( RealGene* B ){
  TREE_PAIR_CPUtimes::cpuTimes(MAP2MAP).startClock();
  Exception( B->isCoveredByASpecies() , "State::dup2Spec ---> RealGene is covered by a Species" );
  Exception( !B->slicedChild(LEFT)->isArtifNode() || !B->slicedChild(RIGHT)->isArtifNode() , 
	     "State::dup2Spec ---> RealGene's children are not Artificial Genes" );

  // Define ArtifGenes A1/A2 and Species S2
  ArtifGene* A1 = const_cast<ArtifGene*> ( (const ArtifGene*) ( B->slicedChild(LEFT) ) );
  ArtifGene* A2 = const_cast<ArtifGene*> ( (const ArtifGene*) ( B->slicedChild(RIGHT) ) );
  Species*   S2 = const_cast<Species*>   ( getSpecies( speciesTree().treePair().mappingFunction( B->node() ) ) );

  SlicedSubtree& tree1  = const_cast<SlicedSubtree&> ( A1->uniqueSlicedSubtree() );
  SlicedSubtree& tree2  = const_cast<SlicedSubtree&> ( A2->uniqueSlicedSubtree() );

  Exception( A1->speciesEdgeSide() == A2->speciesEdgeSide() , "A1 and A2 have to evolve on differents species edges" );
  Exception( S2 != B->coveringSpeciesEdge()->childNode() , "M(B) is not S2" );

  // Remove B from Species Edge 12 and ad it to Species S2
  B->changeCovering( S2 );

  // Remove A1/A2 from Species Gene List and set A1/A2 Species to null
  A1->unSetCovering();
  A2->unSetCovering();

  // Remove A1/A2 from the Gene Edge
  // Important: it does not set it GeneEdge variable to null (needed in splitSlicedSubtrees method)
  A1->removeFromGeneEdge();
  A2->removeFromGeneEdge();

  // Remove A1/A2 from the Sliced Subtree T_12, and put B in its right place
  ArtifGene* A = 0; // One value indicate that it HAS NOT to be created in replaceGene methods
  const_cast<SlicedSubtree*> ( A1->fatherSlicedSubtree() )->replaceGene( A1 , A2 , B , A );

  if( !B->childrenSlicedSubtree_AreDefined() )
    B->buildSlicedSubtrees( true );   

  SlicedSubtree& leftB   = const_cast<SlicedSubtree&> ( B->getSlicedSubtree  ( LEFT ) );
  SlicedSubtree& rightB  = const_cast<SlicedSubtree&> ( B->getSlicedSubtree  ( RIGHT ) );

  // Remove A1 and A2 from their respective sliced subtree
  tree1.removeRoot();
  tree2.removeRoot();

  // Split Sliced subtree T_AL (left of A1) into T_BL
  // Split Sliced subtree T_AR (right of A2) into T_BR
  if( tree1.speciesEdge() == leftB.speciesEdge() ){
    leftB.splitIntoOneSlicedSubtree ( tree1 , B );
    rightB.splitIntoOneSlicedSubtree( tree2 , B );
  }
  else{
    leftB.splitIntoOneSlicedSubtree ( tree2 , B );
    rightB.splitIntoOneSlicedSubtree( tree1 , B );
  }

  // Define B as the root node of T_BL and T_BR
  leftB.setRootNode( B );
  rightB.setRootNode( B );

  TREE_PAIR_CPUtimes::cpuTimes(MAP2MAP).stopClock();
}

// Antecedent:
//      - RealGene  covered by an Edge
//      - The Father Node Edge is not the Root of the Species Tree
//      - The Father is an Artificial Gene
// Consequent:
//      - RealGene's duplication is moved on the father Edge of the Edge that currently
//        covers it.
// Notations:
//      - (1,2): Species father edge
//      - (2,3): Species Edge
//      - A: artif gene
//      - B: real gene
void State::dup2UpDup( RealGene* B , ArtifGene*& A ){
  TREE_PAIR_CPUtimes::cpuTimes(MAP2MAP).startClock();
  Exception( B->isCoveredByASpecies()               , "State::dup2UpDup ---> RealGene is covered by a Species" );
  Exception( B->coveringSpeciesEdge()->isRootEdge() , "State::dup2UpDup ---> Covering species edge is a Root Edge of the Species Tree" );
  Exception( !B->slicedFather()->isArtifNode()      , "State::dup2UpDup ---> Father node is not an Artificial Gene" );

  SpeciesEdge* speciesEdge = const_cast<SpeciesEdge*> ( B->coveringSpeciesEdge() );
  A = const_cast<ArtifGene*> ( (const ArtifGene*) ( B->slicedFather() ) );

  SlicedSubtree& tree  = const_cast<SlicedSubtree&> ( A->uniqueSlicedSubtree()  );

  // Remove A from Species Gene List and set A Species to null
  A->unSetCovering();

  // Remove B from Edge (2,3) and add it to (1,2)
  B->changeCovering( speciesEdge->fatherEdge() );

  // Remove A and B from the SlicedSubtree on edge (2,3)
  const_cast<SlicedSubtree&> (tree).removeGenes ( B , A );

  // Remove A from the Gene Edge
  // Important: it does not set it GeneEdge variable to null (needed in splitSlicedSubtrees method)
  A->removeFromGeneEdge();

  // New ArtifGenes create below
  ArtifGene* A1;
  ArtifGene* A2;

  // Remove A, add B and create A1 and A2 from/to Sliced Subtree T_12
  const_cast<SlicedSubtree*> ( A->fatherSlicedSubtree() )->replaceGene( B , A1 , A2 , A );

  // Create Empty sliced subtrees for A1 and A2
  A1->buildSlicedSubtrees( true );
  A2->buildSlicedSubtrees( true );    

  // Split Sliced subtree T_A into T_A1 and T_A2
  SlicedSubtree& tree1 = const_cast<SlicedSubtree&> ( A1->uniqueSlicedSubtree() );
  SlicedSubtree& tree2 = const_cast<SlicedSubtree&> ( A2->uniqueSlicedSubtree() );

  tree.splitIntoTwoSlicedSubtree( tree1 , tree2 );

  TREE_PAIR_CPUtimes::cpuTimes(MAP2MAP).stopClock();
}

// Antecedent:
//      - B is covered by an Edge
//      - Both children of B are Artif
// Consequent:
//      - RealGene's duplication is moved on the child Edge of the Edge that currently
//        covers it.
// Notations:
//      - (1,2): Species Edge
//      - (2,3): Species Child Edge
//      - A1,A2: existant artif genes
//      - B: real gene
void State::dup2DownDup( RealGene* B , ArtifGene*& A1 , ArtifGene*& A2 ){
  TREE_PAIR_CPUtimes::cpuTimes(MAP2MAP).startClock();
  Exception( B->isCoveredByASpecies() , "State::dup2DownDup ---> RealGene is covered by a Species" );
  Exception( !B->slicedChild(LEFT)->isArtifNode() || !B->slicedChild(RIGHT)->isArtifNode() , 
	     "State::dup2DownDup ---> RealGene's children are not Artificial Genes" );

  // Define the new covering species edge (2,3)
  const SpeciesEdge* speciesEdge = B->coveringChildSpeciesEdge();
  A1 = const_cast<ArtifGene*> ( (const ArtifGene*) ( B->slicedChild(LEFT) ) );
  A2 = const_cast<ArtifGene*> ( (const ArtifGene*) ( B->slicedChild(RIGHT) ) );
  SlicedSubtree& tree1 = const_cast<SlicedSubtree&> ( A1->uniqueSlicedSubtree() );
  SlicedSubtree& tree2 = const_cast<SlicedSubtree&> ( A2->uniqueSlicedSubtree() );

  // Remove A1/A2 from Species Gene List and set A1/A2 Species to null
  A1->unSetCovering();
  A2->unSetCovering();

  // Remove B from Edge (1,2) and ad it to (2,3)
  B->changeCovering( speciesEdge );

  // Remove A1/A2 from the Gene Edge
  // Important: it does not set it GeneEdge variable to null (needed in splitSlicedSubtrees method)
  A1->removeFromGeneEdge();
  A2->removeFromGeneEdge();

  // Remove A1/A2 and B and create A from/to Sliced Subtree T_12
  ArtifGene* A = (ArtifGene*) (1); // One value indicate that it HAS to be created in replaceGene methods

  const_cast<SlicedSubtree*> ( A1->fatherSlicedSubtree() )->replaceGene( A1 , A2 , B , A );

  // Create Empty sliced subtrees for A
  A->buildSlicedSubtrees( true );

  // Split Sliced subtree T_A into T_A1 and T_A2
  SlicedSubtree& tree  = const_cast<SlicedSubtree&> ( A->uniqueSlicedSubtree()  );

  B->setSlicedSubtree( &tree );
  ( const_cast< list< const Gene* >& > ( tree.prefixOrderList() ) ).push_back( B );

  // Remove A1 and A2 from their respective sliced subtree on edge (2,3)
  tree1.removeRoot();
  tree2.removeRoot();

  tree.unSplitIntoOneSlicedSubtree( tree1 , tree2 );

  TREE_PAIR_CPUtimes::cpuTimes(MAP2MAP).stopClock();
}


//////////////////////////////////////////
// Private methods

void State::computeOrthologousStateVector(){
  _orthoStateVector.resize( geneTree().nbOfLeaves() );
  for( uint i = 0; i < _orthoStateVector.size(); i++ )
    _orthoStateVector[i].resize( geneTree().nbOfLeaves() , false );

  const list<LeafNode*>& leafList = geneTree().leafList();

  list<LeafNode*>::const_iterator iterLeaf1, iterLeaf2;
  for( iterLeaf1 = leafList.begin(); iterLeaf1 != leafList.end(); iterLeaf1++ )
    for( iterLeaf2 = leafList.begin(); iterLeaf2 != leafList.end(); iterLeaf2++ )
      if( iterLeaf1 != iterLeaf2 ){
	const RealGene* lcaGene = getGene( geneTree().lcaFunction( *iterLeaf1 , *iterLeaf2 ) );
	_orthoStateVector[ (**iterLeaf1).leafId() ] [ (**iterLeaf2).leafId() ] = lcaGene->isCoveredByASpecies();
      }
}


// Compute for all clade nodes the Dy variable
// Compute for all edges the B part
// The computation is done for the whole tree
void State::compute_All_Dy_and_B_Values(){
  // The Dy computations has to be done with a Postfix Order of the Species Tree
  for( vector<Node*>::const_iterator node = speciesTree().nodeVector().begin(); node != speciesTree().nodeVector().end(); node++)
    if( !(**node).isLeafNode() )
      ( (CladeSpecies*) ( getSpecies( *node ) ) )->compute_Dy();
}


void State::computeReconciliation( const vector<uint>& recVector ){
  for( vector<RealGene*>::iterator gene = _geneVector.begin(); gene != _geneVector.end(); gene++ ){
    uint sId = recVector[ (**gene).nodeId() ];
    if( sId < _speciesVector.size() )
      ( (Gene*) (*gene) )->setCovering( _speciesVector[ sId ] );
    else
      (**gene).setCovering( _speciesEdgeVector[ sId - _speciesVector.size() ] );
  }
}

// Compute all the Sliced Subtree of  Genes covered by a species node
// This is done by a Prefix Order traversal of the Species tree
// Each gene covered by a species is treated seperately
void State::computeAllSlicedSubtrees(){
  for( vector<Node*>::const_reverse_iterator node = speciesTree().nodeVector().rbegin(); node != speciesTree().nodeVector().rend(); node++){
    if( !(**node).isLeafNode() ){
      const Species* species = getSpecies( *node );

      list< Gene* >::const_iterator gene;
      for( gene = species->geneList().begin(); gene != species->geneList().end(); gene++ )
	(**gene).buildSlicedSubtrees( false );
    }
  }
}

bool State::allLeafSpeciesAreCovered() const{
  for( vector<Species*>::const_iterator species = _speciesVector.begin(); species != _speciesVector.end(); species++ )
    if( (**species).isLeafNode() && (**species).geneList().empty() )
      return false;

  return true;
}

// Perform a postfix traversal order of the Species Tree to compute
// each TreeData for genes reconcilied to species
void State::pretreatmentFor_H_Phi(){
  for( vector<Node*>::const_iterator node = speciesTree().nodeVector().begin(); node != speciesTree().nodeVector().end(); node++){
    Species* species = const_cast<Species*> ( getSpecies( *node ) );
    if( species->isLeafNode() )
      ( (LeafSpecies*) (species) )->constructAllTreeData();
    else{
      int upperClassId = ( const_cast<GeneTree&> ( geneTree() ) ).nbOfNodes();
       upperClassId   += ( const_cast<GeneTree&> ( geneTree() ) ).nbOfEdges();
      ( (CladeSpecies*) (species) )->constructAllTreeData( upperClassId );
    }

  }
}

void State::clear_PretreatmentFor_H_Phi(){
  // Unset tree data for all Artif Genes
  for( vector<Species*>::iterator species = _speciesVector.begin(); species != _speciesVector.end(); species++ )
    for( list< Gene* >::const_iterator gene = (**species).geneList().begin(); gene != (**species).geneList().end(); gene++ )
      if( (**gene).isArtifNode() )
	( const_cast< ArtifGene* > ( (const ArtifGene*) (*gene) ) )->unsetTreeData();

  // Unset tree data for all Real Genes
  for( vector<RealGene*>::iterator gene = _geneVector.begin(); gene != _geneVector.end(); gene++ )
    (**gene).unsetTreeData();

  for( vector<Species*>::iterator species = _speciesVector.begin(); species != _speciesVector.end(); species++ )
    if( (**species).isCladeSpecies() )
      ( (CladeSpecies&) (**species) ).clear_H_Phi_Variables();
}

void State::constructTreeData_BottomUp( const Gene* firstGene ){
  try{
    TREE_PAIR_CPUtimes::cpuTimes(TREEDATA_BOTTOMUP).startClock();

    Gene* gene = const_cast<Gene*> (firstGene);

    // While gene \neq r(G)
    while( gene->isArtifNode() || !( (RealGene*) (gene) )->isRootNode() ){
      CladeSpecies* species = const_cast<CladeSpecies*> ( (const CladeSpecies*) ( gene->coveringSpecies()->father() ) );

      SlicedSubtree* T = const_cast<SlicedSubtree*> ( gene->fatherSlicedSubtree() );

      // Remove tree data
      species->removeTreeData( T->rootNode() );

      // Update H and Phi functions
      T->recomputeTreeData( gene->slicedFather() );
      T->H_fun( true );
      T->Phi_fun( true );

      // Add tree data
      species->addTreeData( T->rootNode() );

      gene = const_cast<Gene*> ( T->rootNode() );

      // Halt iff 
      //    Old count move from one  to zero
      //    New count move from zero to one.
      //    In that case, the root id still the same
      //    and there is no need to move upward the gene tree
      //     if(  oldIdInfo.count() == 0 && newIdInfo.count() == 1 ){
      //       oldIdInfo = IdInfo();
      //       break;
      //     }
    }

    TREE_PAIR_CPUtimes::cpuTimes(TREEDATA_BOTTOMUP).stopClock();
  }

  catch( SubException subExc ){
    subExc.addFunDesc( "State::constructTreeData_BottomUp" );
    throw subExc;
  }
}

void State::compute_eV_BottomUp( const Gene* firstGene ){
  TREE_PAIR_CPUtimes::cpuTimes(EV_BOTTOMUP).startClock();

  Gene* gene = const_cast<Gene*> (firstGene);
  const Species* S2 = firstGene->coveringSpecies();

  // While gene \neq r(G)
  while( gene->isArtifNode() || !( (RealGene*) (gene) )->isRootNode() ){
    gene = const_cast<Gene*> ( gene->fatherSlicedSubtree()->rootNode() );

    // Define side
    const Species* Sa = ( (const CladeSpecies*) ( gene->coveringSpecies() ) )->edge( LEFT )->childNode();
    Side side = (*Sa <= *S2? LEFT : RIGHT );
    gene->compute_eA( side );
    gene->define_eV( gene->eA( LEFT ) , gene->eA( RIGHT ) );
  }

  _logL = getGene( geneTree().root() )->eV();
//   computeLogL( true );
//   Exception( newLogL != _logL , "State::compute_eV_BottomUp ---> log differ" );

  TREE_PAIR_CPUtimes::cpuTimes(EV_BOTTOMUP).stopClock();
}

/////////////////////////////////////
// Debug methods

void State::checkAllNodes( std::ostream& output ) const{
  // Check attributes for all nodes
  for( vector<RealGene*>::const_iterator gene = _geneVector.begin(); gene != _geneVector.end(); gene++ )
    (**gene).checkNode( output );
}

void State::checkConsistancy( std::ostream& output ) const{

  output<<"\n\tState checking consistancy ...";

  // Check consistancy for all Sliced Subtrees
  vector<Species*>::const_iterator species;
  for( species = speciesVector().begin(); species != speciesVector().end(); species++ ){
    if( (**species).isCladeSpecies() ){
      list<Gene*>::const_iterator gene;
      for( gene = (**species).geneList().begin(); gene != (**species).geneList().end(); gene++ ){
	if( !(**gene).isArtifNode() ){
	  (**gene).getSlicedSubtree( LEFT ).checkConsistancy( output );
	  (**gene).getSlicedSubtree( RIGHT ).checkConsistancy( output );
	}
	else
	  ( (const ArtifGene&) (**gene) ).uniqueSlicedSubtree().checkConsistancy( output );
      }
    }
  }

  output<<" done!\n";
}

void State::randomGeneTree( string& geneTreeString ){
  // For all the species edge, compute the distribution of the P_Fun function
  for( vector<SpeciesEdge*>::const_iterator speciesEdge = _speciesEdgeVector.begin(); speciesEdge != _speciesEdgeVector.end(); speciesEdge++ )
    if( !(**speciesEdge).isArtificial() )
      (**speciesEdge).compute_P_Fun_Dist();

  // Add the Germe of the whole Birth-and-death process at the root species
  CladeSpecies* rootSpecies = (CladeSpecies*) ( getSpecies( speciesTree().root() ) );
  rootSpecies->addGerme( new Germe( rootSpecies ) );

  // By a prefix order of the species tree:
  // do a random generation of the germe leaves for both left and right species children
  for( vector<Node*>::const_reverse_iterator node = speciesTree().nodeVector().rbegin(); node != speciesTree().nodeVector().rend(); node++)
    if( !(**node).isLeafNode() ){
      CladeSpecies* species = const_cast<CladeSpecies*> ( (CladeSpecies*) ( getSpecies( *node ) ) );
      if( !species->germesList().empty() ) 
	species->generateLeafGermes();
    }

  // By a postfix order of the species tree:
  // Compute the string for all germes of the considered species
  // This is done by a random generation of the induced sliced subtree
  for( vector<Node*>::const_iterator node = speciesTree().nodeVector().begin(); node != speciesTree().nodeVector().end(); node++){
    Species* species = const_cast<Species*> ( getSpecies( *node ) );
    species->computeAllGermeStrings();
  }

  geneTreeString = getSpecies( speciesTree().root() )->germesList().front()->treeString();

  geneTreeString += ';';
}

void State::randomGeneTree2( string& geneTreeString ){

  // Add the Germe of the whole Birth-and-death process at the root species
  CladeSpecies* rootSpecies = (CladeSpecies*) ( getSpecies( speciesTree().root() ) );
  Germe* rootGerme = new Germe( rootSpecies );
  rootSpecies->addGerme( rootGerme );

  // By a prefix order of the species tree:
  // do a random generation of the germe leaves for both left and right species children
  for( vector<Node*>::const_reverse_iterator node = speciesTree().nodeVector().rbegin(); node != speciesTree().nodeVector().rend(); node++)
    if( !(**node).isLeafNode() ){
      CladeSpecies* species = const_cast<CladeSpecies*> ( (CladeSpecies*) ( getSpecies( *node ) ) );
      species->generateLeafGermes2();
    }

  // By a postfix order of the species tree:
  //         Delete germes that have been lost
  for( vector<Node*>::const_iterator node = speciesTree().nodeVector().begin(); node != speciesTree().nodeVector().end(); node++){
    if( !(**node).isLeafNode() ){
      CladeSpecies* species = const_cast<CladeSpecies*> ( (CladeSpecies*) ( getSpecies( *node ) ) );
      species->deleteLostGermes();
    }
  }

  // By a postfix order of the species tree:
  // Compute the string for all germes of the considered species
  // This is done by a random generation of the induced sliced subtree
  for( vector<Node*>::const_iterator node = speciesTree().nodeVector().begin(); node != speciesTree().nodeVector().end(); node++){
    Species* species = const_cast<Species*> ( getSpecies( *node ) );
    species->computeAllGermeStrings();
  }

  // In this case, the rootGerme has not been deleted in the destructor of the (root) Species
  if( getSpecies( speciesTree().root() )->germesList().empty () )
    delete rootGerme;
  else
    geneTreeString = getSpecies( speciesTree().root() )->germesList().front()->treeString();

  geneTreeString += ';';
}

uint State::computeNbOfLost() const{
  uint nbOfLost = 0;
  for( vector<Species*>::const_iterator species = _speciesVector.begin(); species != _speciesVector.end(); species++ )
    if( (**species).isCladeSpecies() )
      nbOfLost += ( (const CladeSpecies&) (**species) ).nbOfLost();
  return nbOfLost;
}

///////////////////////////////
// OUTPUT METHODS

void printRecNodeFct( std::ostream& outputStream, const State& state ){
//  outputStream<< "\nReconciliation node function:";

  outputStream << "\nGene\tSpecies";
  outputStream << "\t\tSize\tAuto\tDistinct\tUpper\tBeta\tString";

  vector<RealGene*>::const_iterator iterNode;
  for( iterNode = state.geneVector().begin(); iterNode != state.geneVector().end(); iterNode++ ){
    if( (**iterNode).isLeafNode() )
      continue;

    outputStream <<"\n" << (**iterNode).geneString();

    if( (**iterNode).isCoveredByASpecies() )
      outputStream <<"\t" << (**iterNode).coveringSpecies()->nodeId();
    else{
      outputStream <<"\t" << *( (**iterNode).coveringSpeciesEdge()->edge() );
      outputStream <<"\t\t" << (**iterNode).treeData();
    }
  }
}

void printOrthoRelationship( std::ostream& outputStream, const State& state ){
  outputStream<< "\n\nOrthologous relationship";
  outputStream<< "\n\t";

  const list<LeafNode*>& leafList = state.geneTree().leafList();

  for( list<LeafNode*>::const_iterator iterLeaf = leafList.begin(); iterLeaf != leafList.end(); iterLeaf++ )
    outputStream << (**iterLeaf).nodeId() <<"\t";

  for( list<LeafNode*>::const_iterator iterLeaf1 = leafList.begin(); iterLeaf1 != leafList.end(); iterLeaf1++ ){
    outputStream<< "\n" << (**iterLeaf1).nodeId();
    for( list<LeafNode*>::const_iterator iterLeaf2 = leafList.begin(); iterLeaf2 != leafList.end(); iterLeaf2++ ){
      outputStream<< "\t";
      if( iterLeaf1 != iterLeaf2 && state.orthologousGenes( state.getGene(*iterLeaf1) , state.getGene( *iterLeaf2 ) ) )
        outputStream<< "Y";
    }
  }
}

void printSlicedSubtrees( std::ostream& outputStream, const State& state ){
  outputStream<<"\n\n";
  vector<Species*>::const_iterator species;
  for( species = state.speciesVector().begin(); species != state.speciesVector().end(); species++ ){
    if( (**species).isCladeSpecies() ){
      list<Gene*>::const_iterator gene;
      for( gene = (**species).geneList().begin(); gene != (**species).geneList().end(); gene++ ){
	outputStream<< (**gene).getSlicedSubtree(LEFT)  <<endl;
	outputStream<< (**gene).getSlicedSubtree(RIGHT)  <<endl;
      }
    }
  }
}

void printCoveringRelationship(  std::ostream& outputStream, const State& state ){
  outputStream<<"\n\n";
  outputStream<<"Species id\tGenelist";

  vector<Species*>::const_iterator species;
  for( species = state.speciesVector().begin(); species != state.speciesVector().end(); species++ ){
    outputStream<<"\n"<< (**species).nodeId() <<"\t\t";
    list<Gene*>::const_iterator gene;
    for( gene = (**species).geneList().begin(); gene != (**species).geneList().end(); gene++ )
      outputStream<< (**gene) <<" ";
  }
}

void printCoveringGeneEdgeRelationship( std::ostream& outputStream, const State& state ){
  outputStream<<"\n\n";
  outputStream<<"Gene Edge vs Species Edge covering relationship";
  outputStream<<"\n\n";

  vector<GeneEdge*>::const_iterator geneEdge;
  for( geneEdge = state.geneEdgeVector().begin(); geneEdge != state.geneEdgeVector().end(); geneEdge++ ){
    outputStream  <<"\t" << *( (**geneEdge).edge() );
  }

  vector<SpeciesEdge*>::const_iterator speciesEdge;

  for( speciesEdge = state.speciesEdgeVector().begin(); speciesEdge != state.speciesEdgeVector().end(); speciesEdge++ ){
    if( !(**speciesEdge).edge()->isArtificial() ){
      outputStream<<"\n" << *( (**speciesEdge).edge() );
      for( geneEdge = state.geneEdgeVector().begin(); geneEdge != state.geneEdgeVector().end(); geneEdge++ ){
	outputStream <<"\t"; 
	if( (**geneEdge).isCoveredBy( *speciesEdge ) )
	  outputStream<< "Y";
      }
    }
  }
  outputStream<<"\n\n";
}

void printSpeciesAncestorRelationship( std::ostream& outputStream, const State& state ){
  outputStream<<"\n\n";
  outputStream<<"Species ancestor relationship (<)";
  outputStream<<"\n\n";

  for( vector<Species*>::const_iterator species = state.speciesVector().begin(); species != state.speciesVector().end(); species++ )
    outputStream<< "\t" << (**species).nodeId();

  for( vector<Species*>::const_iterator species1 = state.speciesVector().begin(); species1 != state.speciesVector().end(); species1++ ){
    outputStream<< "\n" << (**species1).nodeId();
    for( vector<Species*>::const_iterator species2 = state.speciesVector().begin(); species2 != state.speciesVector().end(); species2++ ){
      outputStream<< "\t";
      if( **species1 < **species2 )
	    outputStream<< "Y";
    }
  }
}

void printSpeciesVector( std::ostream& output , const State& state ){
  output <<"\n\nSpecies Vector";
  for( vector<Species*>::const_iterator iter = state.speciesVector().begin(); iter != state.speciesVector().end(); iter++ ){
    if( (**iter).isCladeSpecies() )
      cerr<<endl << (const CladeSpecies&) (**iter);
  }
}

void State::printRealGeneVector( std::ostream& output ) const{
  output <<"\n\nReal Genes Vector";
  output <<"\n\tId\tFather\tLeft\tRight";
  for( vector<RealGene*>::const_iterator iter = geneVector().begin(); iter != geneVector().end(); iter++ ){
    output<<"\n\t";
    output<< (**iter).nodeId();

    output<<"\t";
    if( !(**iter).isRootNode() )
      output<< (**iter).slicedFather()->geneString();
   
    if( !(**iter).isCoveredByASpecies() ){
      output<<"\t" << (**iter).slicedChild(LEFT)->geneString();
      output<<"\t" << (**iter).slicedChild(RIGHT)->geneString();
    }
  }
}

void State::printGermesList( std::ostream& output ){
  for( vector<Species*>::const_iterator species = speciesVector().begin(); species != speciesVector().end(); species++ )
    (**species).printGermesList( output );
}

void printEdgesVector( std::ostream& output , const State& state ){
  output <<"\n\nSpecies edge vector:";
  for( vector<SpeciesEdge*>::const_iterator iterEdge = state.speciesEdgeVector().begin(); iterEdge != state.speciesEdgeVector().end(); iterEdge++ )
    output<< "\n\n" << **iterEdge;
}


std::ostream& operator << (std::ostream& outputStream, const State& state ){

//   outputStream<< state.geneTree();
//   outputStream<< state.speciesTree();

  printRecNodeFct( outputStream , state );

//  printRealGeneVector( outputStream );

//  printOrthoRelationship( outputStream , state );

//  printSlicedSubtrees( outputStream , state );
//   printCoveringGeneEdgeRelationship( outputStream , state );

//   printSpeciesAncestorRelationship( outputStream , state );

//   printCoveringRelationship( outputStream , state );

//  printSpeciesVector( outputStream , state );
//  printEdgesVector  ( outputStream , state );

  outputStream << setprecision(2) << "\n\nLik = " << exp( state.logL() );

  return outputStream;
}


