/***************************************************************************
			State.h
		-------------------
	begin	 : Sun Jan  7 18:42:22 EST 2007
	copyright: (C) 2005 by Universite De Montreal
	email	 :  doyonjea@iro.umontreal.ca
***************************************************************************/
#ifndef STATE_H
#define STATE_H

//! This class defines the base structure for a State object.
/**
*@author Jean-Philippe Doyon
*/

#include "Species.h"
#include "SpeciesEdge.h"
#include "RealGene.h"
#include "GeneEdge.h"
#include "TreePair.h"

#include "SlideClasses.h"
#include "BirthDeathParam.h"

#include "Reconciliation.h"

class State:public Reconciliation{

  friend std::ostream& operator << (std::ostream& outputStream, const State& state );

 private:
  static int _classCounter;

  double _logL;
  vector<Species*>     _speciesVector;     // Bijection: Species <----> Node of the species tree
  vector<SpeciesEdge*> _speciesEdgeVector; // Bijection: SpeciesEdge <----> Edge of the species tree
  vector<RealGene*>    _geneVector;        // Bijection: Gene    <----> Node of the gene tree 
  vector<GeneEdge*>    _geneEdgeVector;

  // For a pair of leaf genes, indicates if they are orthologous or not
  // Upper triangular matrix; i < j is an antecedent
  vector< vector<bool> > _orthoStateVector;

 public:

  static int classCounter(){ return _classCounter; }

  // Constructor and destructor

  // Used when the recVector is defined
  State( const TreePair& treePair , const vector<uint>& recVector , const BirthDeathParam& birthDeathParam );

  // Used to generate a random gene tree
  State( const TreePair& treePair , const BirthDeathParam& birthDeathParam );

  ~State();

  // Delete methods
  void deleteSlicedSubtrees();
  void deleteArtifGenes();

  // Used when a swap, between two chains (Metropolis Coupling), is made
  void redefine( const vector<uint>& recVector );

  // Access methods (Class attributes)
  const vector<Species*>&     speciesVector()     const { return _speciesVector; }
  const vector<SpeciesEdge*>& speciesEdgeVector() const { return _speciesEdgeVector; }
  const vector<RealGene*>&    geneVector()        const { return _geneVector; }
  const vector<GeneEdge*>&    geneEdgeVector()    const { return _geneEdgeVector; }

  // Species/Genes relationship
  Species*     getSpecies( const Node* node )     const{ return _speciesVector[ node->nodeId() ]; }
  SpeciesEdge* getSpeciesEdge( const Edge* edge ) const{ return _speciesEdgeVector[ edge->edgeId() ]; }
  RealGene*    getGene( const Node* node )        const{ return _geneVector[ node->nodeId() ]; }
  GeneEdge*    getGeneEdge( const Edge* edge )    const{ return _geneEdgeVector[ edge->edgeId() ]; }

  // Genes orthologous relationship
  bool orthologousGenes( const RealGene* gene1 , const RealGene* gene2 ) const;

  // Likelihood methods
  // Clear means that it has to clear previous preatment for H_Phi
  double computeLogL( bool clear , bool debug = false );
  double logL() const { return _logL; }

  //MCMC operator (4 possibilities, se below)
  bool slideIsPossible( uint nodeId , DIRECTION dir );
  void performSlide( const Slide& slide , bool updateLogL );

/*   // RecVector methods */
/*   void updateRecVector( const Slide& slide  ); */

  // Output methods
  void writeToPrimeFile( const string& geneFileName , const string& speciesFileName ) const;
  void printRealGeneVector( std::ostream& output ) const;
  void printGermesList    ( std::ostream& output );

  // Shortcut methods
  const SpeciesTree& speciesTree() const{
    return (const SpeciesTree&) ( _speciesVector[0]->node()->tree() ); 
  }
  const GeneTree& geneTree() const{ 
    return (const GeneTree& )   ( _geneVector[0]->node()->tree() ); 
  }
  const RealGene* getRealGene( uint id ) const{
    return _geneVector[id]; 
  }

  const TreePair& treePair() const{
    return speciesTree().treePair();
  }

  const Species* mappingFunction( const RealGene* realGene ) const{ 
    return getSpecies( speciesTree().treePair().mappingFunction( realGene->node() ) );
  }

  // Antecedent methods
  // True iff all leaf species have at least one gene
  bool allLeafSpeciesAreCovered() const;

  // Compute all treeString canonical representation for the whole gene tree
  void pretreatmentFor_H_Phi();
  void clear_PretreatmentFor_H_Phi();

  // Debug methods
  void checkAllNodes( std::ostream& output ) const; // Verify that all nodes attributes are OK
  void checkConsistancy( std::ostream& output ) const;
  bool stateIsValid( std::ostream& output ) const;

  // Gene tree random generation methods
  //        Recursive Birth-and-death (Arvestad et al.)
  void randomGeneTree( string& geneTreeString );
  //        Pure Birth-and-death
  void randomGeneTree2( string& geneTreeString );

  // Compute the number of lost (nb. of ArtifGene)
  uint computeNbOfLost() const;



 private:

  // Methods used by both constructor
  void initializeVectors( const TreePair& treePair , const BirthDeathParam& birthDeathParam );

  // Species/Genes relationship
  void computeOrthologousStateVector();

  // Likelihood methods
  void compute_All_Dy_and_B_Values();

  void updateLogL_SpecDup_1( RealGene* B , SpeciesEdge*& edge23 , SpeciesEdge*& edge24 , bool updateLogL );
  void updateLogL_SpecDup_2( RealGene* B , SpeciesEdge* edge23 , SpeciesEdge* edge24 , bool updateLogL );

  void updateLogL_DupSpec_1( RealGene* B , ArtifGene*& A1 , ArtifGene*& A2 , bool updateLogL );
  void updateLogL_DupSpec_2( RealGene* B , ArtifGene* A1 , ArtifGene* A2  , bool updateLogL );

  void updateLogL_DupDup ( const int sign , DIRECTION dir , RealGene* B , bool updateLogL );
  void constructTreeData_BottomUp( const Gene* firstGene );
  void compute_eV_BottomUp( const Gene* firstGene );

  // Define the state according to the recVector 
  // Used when a reconciliation is used as the first state of the MCMC (see Parameters)
  void computeReconciliation( const vector<uint>& recVector );
  
  void addArtifNodes( const SlicedSubtree& slicedSubtree );
  void computeAllSlicedSubtrees();

  //MCMC operator
  void spec2Dup   ( RealGene* realGene );
  void dup2Spec   ( RealGene* realGene );
  void dup2UpDup  ( RealGene* realGene , ArtifGene*& A );
  void dup2DownDup( RealGene* realGene , ArtifGene*& A1 , ArtifGene*& A2 );
  
  SlideName getSlideName( RealGene* realGene , const Slide& slide ) const;
};

#endif
