#include "StateSet.h"
#include "MathClass.h"
#include "OrthoParam.h"
#include "SlideSet.h"

void StateSet::initialize( uint speciesTreeSize , uint nbOfGenes ){
  _layerSize = speciesTreeSize;
  _rootLayer = new StateLayer( _layerSize , true , 0 );
  _layerList.push_back( _rootLayer );
  
  _nbOfGenes = nbOfGenes;
}

void StateSet::computePostProb(){
  double totalL = 0;
  for( list< const StateInfo* >::const_iterator state = stateInfoList().begin(); state != stateInfoList().end(); state++ )
    totalL += exp( (**state).logL() );

  for( list< const StateInfo* >::iterator state = stateInfoList().begin(); state != stateInfoList().end(); state++ )
    ( const_cast<StateInfo&> (**state) ).setPostProb( exp( (**state).logL() ) / totalL );
}

void StateSet::computeApproxPostProb(){

  _totalCount = 0;
  for( list< const StateInfo* >::const_iterator iter = stateInfoList().begin(); iter != stateInfoList().end(); iter++ )
    _totalCount += (**iter).count();

  for( list< const StateInfo* >::iterator state = stateInfoList().begin(); state != stateInfoList().end(); state++ ){
    if( totalCount() != 0 )
      ( const_cast<StateInfo&> (**state) ).setApproxPostProb( (double) (**state).count() / totalCount() );
  }
}


uint StateSet::nbOfNonNullState() const{
  uint N = 0;

  for( list< const StateInfo* >::const_iterator iter = stateInfoList().begin(); iter != stateInfoList().end(); iter++ )
    if( (**iter).count() > 0 )
      N++;

  return N;
}

StateInfo* StateSet::addState( const vector<uint>& recVector , const string& stringId ){

//  Exception( !state.stateIsValid( cerr ) , "StateSet::addState ---> state is not valid" );

  const StateLayer* leafLayer = (const StateLayer*) ( retrieveLeafLayer( recVector ) );

  const StateInfo* stateInfo;

  // Case A: first time the state is visited
  if( !leafLayer->objExists( recVector.back() ) ){
    // Create StateInfo on the corresponding leaf layer
    stateInfo = ( const_cast<StateLayer*> (leafLayer) )->createStateInfo( recVector , stringId );
    stateInfoList().push_back( stateInfo );
  }
  // Case B: state already exists (has been previously visited)
  else
    stateInfo = leafLayer->getObjInfo( recVector.back() );

  return const_cast<StateInfo*> (stateInfo);
}

const StateInfo& StateSet::getStateInfo( const State& state ) const{
  return ( const StateInfo& ) ( getObjInfo( state.recVector() ) );
}

const StateInfo& StateSet::getStateInfo( const vector<uint>& recVector ) const{
  return ( const StateInfo& ) ( getObjInfo( recVector ) );
}


void StateSet::checkStateConsistency( const State& state ) const{
  if( objExists( state.recVector() ) ){
    const StateInfo& stateInfo = getStateInfo( state );
    Exception( stateInfo.recVector() != state.recVector()                                , "StateSet::checkStateConsistency ----> recVector differs" );
    double diffRatio = MathClass::diffRatio( stateInfo.logL() , state.logL() );
    if( diffRatio >= MAX_DIFF_RATIO ){
      cerr<< state;
      cerr<<"\nDiff ratio: " << diffRatio;
      cerr<<"\nCount: " << stateInfo.count();
      cerr<<"\nState id: " << stateInfo.id();
      Exception( true , "StateSet::checkStateConsistency ----> logL differs " );
    }
  }
}

void StateSet::orderStateInfoList(){
  // Order list in increasing order of likelihood
  list< const StateInfo* > tmpList;

  while( !stateInfoList().empty() ){
    const StateInfo* state = stateInfoList().front();
    stateInfoList().pop_front();

    list< const StateInfo* >::iterator it = tmpList.begin();
    while( it != tmpList.end() && (**it).postProb() < (*state).postProb() )
      it++;
      
    tmpList.insert( it , state );
  }
  stateInfoList().splice( stateInfoList().end() , tmpList );
}


void StateSet::writeAllState( std::ostream& output , const TreePair& treePair , const BirthDeathParam& birthDeathParam , double minPostProb ) const{
  for( list< const StateInfo* >::const_iterator stateInfo = stateInfoList().begin(); stateInfo != stateInfoList().end(); stateInfo++ ){
    State state( treePair , (**stateInfo).recVector() , birthDeathParam );
    //    output<<"\nReconciliation: " << (**stateInfo).stringId();
    if( (**stateInfo).postProb() >= minPostProb ){
      output<< state;
      output<< "\nPost prob:" << (**stateInfo).postProb();
      output<< endl;
    }
  }
}

void StateSet::print( std::ostream& output , const OrthoParam& orthoParam ) const{
  // Compute Diff total
  double diffTotal = 0;
  for( list< const StateInfo* >::const_iterator state = stateInfoList().begin(); state != stateInfoList().end(); state++ )
    diffTotal += (**state).diffPostProb();

  if( orthoParam.paramIntValue( OUTPUT_POST_PROB ) ){
    // Print all StateInfo
    if( orthoParam.paramIntValue( OUTPUT_POST_PROB ) )
      output<< "\n\nRecVector\t\t\tPostProb\t\tLogL\t\tCount\t\t\tApprox\t\t\tEcart";

    output<< setprecision(2);

    for( list< const StateInfo* >::const_iterator state = stateInfoList().begin(); state != stateInfoList().end(); state++ ){
      output <<"\n";
      (**state).print( orthoParam , output );
      if( _treePair.extremeRec( Lowest ) == **state )
	output <<"\t\t Lowest";
      if( _treePair.extremeRec( Highest ) == **state )
	output <<"\t\t Highest";
    }
  }


  output<<"\n\nSummation of (P - P') = " << diffTotal;

  output<< "\n\nState set";
  output<< "\n\tSize set        = " << size();
  output<< "\n\tTotal count = " << totalCount();
  output<< "\n\tNb of non null state = " << nbOfNonNullState();
}
