#include "Tree.h"
#include "CladeNode.h" 
#include "Node.h"
#include "LeafNode.h"
#include "ArtifGene.h"
#include "TreePair.h"
#include "State.h"
#include "Reconciliation.h"

const Edge* Tree::getEdge( uint edgeId ) const{
  Exception( vectorIdOutOfBound( edgeId , _edgeVector.size() ) , "Tree::getEdge ----> edge Id out of bound" );
  return _edgeVector[ edgeId ];
}

const Node* Tree::getNode( uint nodeId ) const{
  Exception( vectorIdOutOfBound( nodeId , _nodeVector.size() ) , "Tree::getNode ---> node id out of bound" );
  return _nodeVector[ nodeId ];
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// mainBuild sub functions
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Tree::constructTreeStructure( Edge*& artificialEdge ){
  list<Node*> stackNode;
  uint depth = 0;
  uint i = 0;

  // Construct the tree data structure
  while( i < _treeString.size() ){
    if( _treeString[i] == '(' ){
      CladeNode* cladeNode;
      if( i == 0 ){
	cladeNode = _root = new CladeNode( depth , *this );

	// Only the species tree has an (Species) edge
	// It is important that the edge is a "SpeciesEdge", otherwise they would be a "glibc detected free()" error
	if( isASpeciesTree() )
	  artificialEdge = new Edge( 0 , _root ); 
      }
      else{
	cladeNode = new CladeNode ( depth , *this , (CladeNode*) stackNode.back() );
	( (CladeNode*) (stackNode.back()) )->addChild( cladeNode );
      }
      stackNode.push_back( cladeNode );
      _nodeList.push_back( cladeNode );
      depth++;
    }

    if( _treeString[i] == ',' ){
      stackNode.pop_back();
      depth--;
    }

    if( _treeString[i] == ')' ){
      stackNode.pop_back();
      depth--;
    }    

    if( isdigit( _treeString[i] ) ){
      uint taxaId = readTaxaId( _treeString , i );

      CladeNode* father = ( stackNode.empty()? 0: (CladeNode*) stackNode.back() );
      LeafNode* leafNode = new LeafNode( depth , taxaId , *this , father );

      if( _treeString[i] == '*' ){
	i++;
	leafNode->setFictive( true );
      }

      _leafList.push_back( leafNode );

      if( !stackNode.empty() ) 
	( (CladeNode*) (stackNode.back()) )->addChild( leafNode );

      stackNode.push_back( leafNode );
      _nodeList.push_back( leafNode );
      depth++;
      i--; // Put the reading position before the last ',' or ')'
    }
    i++;
  }
}

void Tree::computeNode_And_EdgeVectors( Edge* artificialEdge ){
  _nodeVector.resize( Node::currentNodeId() );
  _edgeVector.resize( Edge::currentEdgeId() );

  if( isASpeciesTree() )
    _edgeVector[ artificialEdge->edgeId() ] = artificialEdge;

  list<Node*>::const_iterator iterNode;

  for( iterNode = _nodeList.begin(); iterNode != _nodeList.end(); iterNode++ ){
    _nodeVector[ (**iterNode).nodeId() ] = *iterNode;
    if( !(**iterNode).isLeafNode() ){
      CladeNode* cladeNode = (CladeNode*) (*iterNode);
      _edgeVector[ cladeNode->edge( LEFT )->edgeId()  ] = const_cast<Edge*> ( cladeNode->edge( LEFT ) );
      _edgeVector[ cladeNode->edge( RIGHT )->edgeId() ] = const_cast<Edge*> ( cladeNode->edge( RIGHT ) );
    }
  }
}

void Tree::compute_Ancestors_And_Dist_Matrices(){
  // Define matrices size and default values

  _ancestorsBoolMatrix.resize( nbOfNodes() );
  _distMatrix.resize( nbOfNodes() );

  for( int i = 0; i < nbOfNodes(); i++ ){
    _ancestorsBoolMatrix[i].resize( nbOfNodes() , false );
    _distMatrix[i].resize( nbOfNodes() , 0 );
  }

  for( vector<Node*>::iterator iterNode = _nodeVector.begin(); iterNode != _nodeVector.end(); iterNode++ ){
    if( !(**iterNode).isRootNode() ){
      // Set ancestor relationship for all nodes along the path between the root and this node
      const Node* tmpNode = *iterNode;
      uint dist = 0;
      do{
	tmpNode = tmpNode->father();
	_ancestorsBoolMatrix[ tmpNode->nodeId() ] [ (**iterNode).nodeId() ] = true;
	_distMatrix[ tmpNode->nodeId() ] [ (**iterNode).nodeId() ] = dist++;
      }while( !tmpNode->isRootNode() );
    }
  }
}

void Tree::computeLCAFunctionVector(){
  // Construct the lcaVector and set all value to 0
  _lcaFctVector.resize( nbOfNodes() );
  for( uint i = 0; i < _lcaFctVector.size(); i++ )
    _lcaFctVector[i].resize( nbOfNodes() , 0 );

  // Compute the lca values with a trivial NxN loop
  // This step can be optimized by completing only the upper triangular matrix
  vector<Node*>::iterator iterNode1,iterNode2;
  for( iterNode1 = _nodeVector.begin(); iterNode1 != _nodeVector.end(); iterNode1++ )
    for( iterNode2 = _nodeVector.begin(); iterNode2 != _nodeVector.end(); iterNode2++ )
      _lcaFctVector[ (**iterNode1).nodeId() ] [ (**iterNode2).nodeId() ] = computeLCAFct( *iterNode1 , *iterNode2 );
}

void Tree::computePostfixNodeIds(){
  // Depth First Search
  restartAll_DFS_Visited();

  list<CladeNode*> stackNode;
  stackNode.push_back( _root );


  while( !stackNode.empty() ){
    CladeNode* currentNode = stackNode.back();
    Node* nextNode = currentNode->nextDFSChild();

    if( nextNode != 0 ){
      nextNode->setDFS_Visited( true );

      if( !nextNode->isLeafNode() )
	stackNode.push_back( (CladeNode*) nextNode );
      else
	nextNode->setPostfixNodeId();

    }
    else{
      stackNode.pop_back();
      currentNode->setPostfixNodeId();
    }
  }
}

void Tree::computePrefixNodeIds(){
  list<Node*> queueNode;
  queueNode.push_front( _root );

  uint id = 0;
  while( !queueNode.empty() ){
    Node* node = queueNode.front();
    queueNode.pop_front();
    node->setPrefixId( id++ );
    _prefixList.push_back( node );

    if( !node->isLeafNode() ){
      queueNode.push_front( const_cast<Node*> ( ( (CladeNode*) (node) )->child( RIGHT ) ) );
      queueNode.push_front( const_cast<Node*> ( ( (CladeNode*) (node) )->child( LEFT  ) ) );
    }
  }
}


Tree::Tree( const TreePair& treePair , const ParamTree& paramTree , bool isASpeciesTree ):
  _treePair( treePair ) , _paramTree(paramTree) , _isASpeciesTree( isASpeciesTree ){

  string treeString = _paramTree.treeString2();

  if( treeString.length() != 0 ){
    // Reset CurrentIds for LeafNode, Edge and Node classes
    LeafNode::resetCurrentLeafId();
    Edge::resetEdgeId();
    Node::resetNodeId();

    _treeString = treeString;

    Edge* artificialEdge = 0; // This is an artificial edge where child node is the root (FOR SPECIES TREE ONLY)

    constructTreeStructure( artificialEdge );

    transformInNormalForm();


    computePostfixNodeIds();
    computePrefixNodeIds();
    computeNode_And_EdgeVectors( artificialEdge );


    compute_Ancestors_And_Dist_Matrices();
    computeLCAFunctionVector();
    computeAllCladeSet();
  }
}

Tree::~Tree(){
  // Delete all nodes
  for( vector<Node*>::iterator iterNode = _nodeVector.begin(); iterNode != _nodeVector.end(); iterNode++ )
    delete *iterNode;

  // Delete all edges
  for( vector<Edge*>::iterator iterEdge = _edgeVector.begin(); iterEdge != _edgeVector.end(); iterEdge++ )
    delete *iterEdge;
}

void Tree::writeToOfstream( ofstream& treeStream , const string& treeFigName ){
  // Depth First Search
  restartAll_DFS_Visited();

  list<CladeNode*> stackNode;

  if( !treeFigName.empty() )
    treeStream << treeFigName <<":\n";

  treeStream <<"(";
  stackNode.push_back( _root );

  while( !stackNode.empty() ){
    CladeNode* currentNode = stackNode.back();

    Node* nextNode = currentNode->nextDFSChild();

    if( nextNode != 0 ){
      nextNode->setDFS_Visited( true );
      if( !nextNode->isLeftChild() )
	treeStream <<",";

      if( !nextNode->isLeafNode() ){
	treeStream <<"(";
	stackNode.push_back( (CladeNode*) nextNode  );
      }
      else
	treeStream << nextNode->nodeId();
    }
    else{
      treeStream <<")";
      treeStream << currentNode->nodeId();

      stackNode.pop_back();
    }
  }
  treeStream <<";\n";
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Structure functions
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

const Node* Tree::lcaFunction( const Node* node1 , const Node* node2 ) const{
  Exception( !containsNode( node1 )                                    , "Tree::lcaFunction ---> node1 is not in the tree" );
  Exception( !containsNode( node2 )                                    , "Tree::lcaFunction ---> node2 is not in the tree" );
  Exception( _lcaFctVector[ node1->nodeId() ] [ node2->nodeId() ] == 0 , "Tree::lcaFunction ---> lca is null" );

  return _lcaFctVector[ node1->nodeId() ] [ node2->nodeId() ];
}

Node* Tree::computeLCAFct( const Node* node1 , const Node* node2 ) const{

  Node* node = const_cast<Node*> (node1);
  while( !( *node <= *node2 ) )
    node = const_cast<CladeNode*> ( node->father() );

  return node;
}

bool Tree::containsNode( const Node* node ) const{
  Exception( node == 0 , "Tree::containsNode ---> node == 0" );

  if( node->nodeId() >= _nodeVector.size() || _nodeVector[ node->nodeId() ] != node )
    return false;

  return true;
}

bool Tree::isBinary() const{
  for( vector<Node*>::const_iterator node = _nodeVector.begin(); node != _nodeVector.end(); node++ )
    if( !(**node).isLeafNode() && !( (const CladeNode&) (**node) ).isBinary() )
      return false;
  return true;
}

///////////////////
// Distance methods

// Node , Node
uint Tree::dist( const Node* node1 , const Node* node2 ) const{
  Exception( !( *node1 <= *node2 ) , "Tree::dist1 ----> node1 !<= node2" );
  return _distMatrix[ node1->nodeId() ] [ node2->nodeId() ];
}

// Node , Edge
uint Tree::dist( const Node* node1 , const Edge* edge2 ) const{
  const Node* node2 = edge2->childNode();
  Exception( !( *node1 < *node2 ) , "Tree::dist2 ----> node1 !< node2" );
  return _distMatrix[ node1->nodeId() ] [ node2->nodeId() ];
}

// Edge , Node
uint Tree::dist( const Edge* edge1 , const Node* node2 ) const{
  Exception( !( *( edge1->fatherNode() ) < *node2 ) , "Tree::dist3 ----> fatherNode !< node2" );
  Exception( !( *( edge1->childNode() ) <= *node2 ) , "Tree::dist3 ----> fatherNode !<= node2" );

  const Node* node1 = edge1->fatherNode();
  return _distMatrix[ node1->nodeId() ] [ node2->nodeId() ];
}

// Edge , Edge
uint Tree::dist( const Edge* edge1 , const Edge* edge2 ) const{
  const Node* edge1a = edge1->fatherNode();
  const Node* edge1b = edge1->childNode();

  const Node* edge2a = edge2->fatherNode();
  const Node* edge2b = edge2->childNode();

  Exception( !( *edge1a <= *edge2a ) , "Tree::dist ---> edge1a !<= edge2a" );
  Exception( !( *edge1b <= *edge2b ) , "Tree::dist ---> edge1b !<= edge2b" );

  if( edge1 == edge2 )
    return 0;
  else
    return dist( edge1a , edge2b );
}

void Tree::restartAll_DFS_Visited(){
  for( int i = 0; i < _nodeVector.size(); i++ )
    _nodeVector[i]->setDFS_Visited( false );
}

bool Tree::ancestorMatrixValue( const Node* node1 , const Node* node2 ) const{
  return _ancestorsBoolMatrix[ node1->nodeId() ] [ node2->nodeId() ];
}

// Perform a postfix tree traversal order
void Tree::transformInNormalForm() const{
  for( vector<Node*>::const_iterator node = _nodeVector.begin(); node != _nodeVector.end(); node++ )
    if( !(**node).isLeafNode() ){
      CladeNode& cladeNode = const_cast<CladeNode&> ( (const CladeNode&) (**node) );
      if( !cladeNode.child( LEFT )->isFirstInNormForm( cladeNode.child( RIGHT ) ) )
	cladeNode.interchangeChildren();
    }
}

// Perform a postfix tree traversal order
string Tree::computeNormalFormString( const Reconciliation* reconciliation ) const{
  for( vector<Node*>::const_iterator node = _nodeVector.begin(); node != _nodeVector.end(); node++ )
    if( !(**node).isLeafNode() ){
      CladeNode& cladeNode = const_cast<CladeNode&> ( (const CladeNode&) (**node) );
      cladeNode.computeNormalFormString( reconciliation );
    }

  return _root->subtreeNormForm() + ";";
}

string Tree::treeTypeString() const{
  if( isASpeciesTree() )
    return "SPECIES_TREE";
  else
    return "GENE_TREE";
}

// This computation is not usefull for this project (see Naiara relating projects)
void Tree::computeAllCladeSet(){
  vector<Node*>::iterator iterNode;
  for( iterNode = _nodeVector.begin(); iterNode != _nodeVector.end(); iterNode++ )
    if( !(**iterNode).isLeafNode() ){
      CladeNode* cladeNode = (CladeNode*) *iterNode;
      cladeNode->addToCladeSet( cladeNode->child(LEFT)->cladeSet() );
      cladeNode->addToCladeSet( cladeNode->child(RIGHT)->cladeSet() );
    }
}

void outputDFSVisited( std::ostream& outputStream, const Tree& tree){
  outputStream<<"\n\nDFS Visited status";
  vector<Node*>::const_iterator iterNode;
  for( iterNode = tree.nodeVector().begin(); iterNode != tree.nodeVector().end(); iterNode++ )
    cerr<<"\nNode: " <<**iterNode <<" --> " << (**iterNode).dfsVisited();
}

std::ostream& operator << (std::ostream& output, const Tree& tree){

  output << tree.treeString();

//   output <<"\n\nDist matrix: "  << tree.distMatrix().size();
//   output <<"\n\nAncestor bool matrix: "  << tree.ancestorsBoolMatrix();

//   outputDFSVisited( output , tree );

//  tree.outputDistMatrix( output );
//  tree.outputAncestorMatrix( output );

//  tree.outputNodes( output );
//   tree.outputEdges( output );

//  tree.outputLCAFunction( output );

  return output;
}


void Tree::outputLCAFunction( ostream& output ) const{
  for( uint id1 = 0; id1 < _nodeVector.size() - 1; id1++ )
    for( uint id2 = id1 + 1; id2 < _nodeVector.size(); id2++ ){
      output<<endl;
      output<< id1 << " , " << id2 << ": ";
      output<< lcaFunction( getNode(id1) , getNode(id2) )->nodeId();
    }
}

void Tree::outputDistMatrix( ostream& output ) const{
  output<<"\n\nDistances between nodes\n";
  for( uint i = 0; i < _distMatrix.size(); i++ )
    cerr<< "\t" << i;

  for( uint i = 0; i < _distMatrix.size(); i++ ){
    cerr<< "\n\n" << i;
    for( uint j = 0; j < _distMatrix.size(); j++ ){
      cerr<<"\t";
      if( *( _nodeVector[i] ) <= *( _nodeVector[j] ) )
	cerr<< _distMatrix[i][j];
    }
  }
}

void Tree::outputAncestorMatrix( ostream& output ) const{
  output <<"\n\nNode ancestor relationship";
  output <<"\nNode\tAncestors";

  vector<Node*>::const_iterator iterNode1, iterNode2;
  for( iterNode2 = nodeVector().begin(); iterNode2 != nodeVector().end(); iterNode2++ ){
    output <<"\n" << (**iterNode2).nodeId() <<"\t";
    for( iterNode1 = nodeVector().begin(); iterNode1 != nodeVector().end(); iterNode1++ )
      if( ancestorMatrixValue( *iterNode1 , *iterNode2 ) )
	output << (**iterNode1).nodeId() <<" ";
  }
}

void Tree::outputNodes( ostream& output ) const{
  output <<"\n\nPostfix\tPrefix\tDepth\tNodeType\tSpecieId\tSide\tCladeSet\n";
  
  vector<Node*>::const_iterator iterNode;
  //  for( iterNode = nodeVector().begin(); iterNode != nodeVector().end(); iterNode++ ){
  for( list<Node*>::const_iterator iterNode = _prefixList.begin(); iterNode != _prefixList.end(); iterNode++ ){
    output <<"\n" << (**iterNode).nodeId();
    output <<"\t" << (**iterNode).prefixId();
    output <<"\t" << (**iterNode).depth();
    output <<"\t" << (**iterNode).isLeafNode();

    output <<"\t";
    if( (**iterNode).isLeafNode() )
      output << ( (LeafNode*) (*iterNode) )->specieId();

    if( !(**iterNode).isRootNode() )
      output<<"\t" << ( (**iterNode).isLeftChild()? "L": "R" );

    output <<"\t\t" << (**iterNode).cladeSet();
  }

  output<<"\nPrefix list:";
  for( list<Node*>::const_iterator node = _prefixList.begin(); node != _prefixList.end(); node++ )
    output<<" " << (**node).prefixId();
}

void Tree::outputEdges( ostream& output ) const{
  output <<"\n\nEdge Id\t\tEdge";
  vector<Edge*>::const_iterator iterEdge;
  for( iterEdge = edgeVector().begin(); iterEdge != edgeVector().end(); iterEdge++ ){
    output <<"\n"  << (**iterEdge).edgeId();
    int nodeId1 = (**iterEdge).isArtificial()? ARTIFICIAL_NODE_ID: (int) (**iterEdge).fatherNode()->nodeId();
    
    output <<"\t\t(" << nodeId1 <<" , " << (**iterEdge).childNode()->nodeId() << ")";
  }
}
