/***************************************************************************
			Tree.h
		-------------------
	begin	 : Thu Jun  8 15:13:51 EDT 2006
	copyright: (C) 2005 by Universite De Montreal
	email	 :  doyonjea@iro.umontreal.ca
***************************************************************************/
#ifndef TREE_H
#define TREE_H

//! This class defines the base structure for a Tree object.
/**
*@author Jean-Philippe Doyon
*/

#include "CladeNode.h"
#include "LeafNode.h"
#include "Edge.h"


class TreePair;
class State;
class Reconciliation;
#include "ParamTree.h"

/* Binary relation between a pair of nodes / edges:
For nodes:
   x is Less_Eq to y iff ( x is an ancestor of y or x == y)
*/

class Tree{

  friend std::ostream& operator << (std::ostream& outputStream, const Tree& tree);

 private:
  vector< vector<Node*> > _lcaFctVector;

  // This variable is used in the Tree constructor
  // It is necessary because a virtual function "isASpeciesTree" can't be called in the constructor
  const bool _isASpeciesTree; 
  
  // List ordered accorded to the order of nodes induced by the input tree string
  // Used to define the reconciliation induced by the Birth-And-Death (randomGeneTree)
  list<Node*> _nodeList;

  // The node of the list are ordered according to the prefix (u,u_1,u_2) traversal of T
  list<Node*> _prefixList;

 protected:

  const ParamTree& _paramTree;

  CladeNode* _root;
  string _treeString;

  // The node are ordered according to a postfix order of the tree
  vector<Node*> _nodeVector;
  vector<Edge*> _edgeVector;
  list<LeafNode*> _leafList;
  
  // M[x][y] = true iff x is an ancestor of y
  vector< vector< bool > > _ancestorsBoolMatrix;

  // For x,y \in V(G) s.t. x <= y only
  // D[x][y] is the number of nodes z \in V(G) s.t. x < z < y
  vector< vector< uint > > _distMatrix;

  const TreePair& _treePair;

  // Public methods
 public: 

  // This BIDON method is defined to define Tree as an abstract class
  virtual void bidon() const = 0;

  // Constructor and destructor
  Tree( const TreePair& treePair , const ParamTree& paramTree , bool isASpeciesTree );
  virtual ~Tree();

  // Access methods (Class attributes)
  CladeNode* root()                 const { return _root; }
  string treeString()               const { return _treeString; }
  const vector<Node*>& nodeVector() const { return _nodeVector; }
  const vector<Edge*>& edgeVector() const { return _edgeVector; }
  const list<LeafNode*>& leafList() const { return _leafList; }

  const list<Node*>& nodeList()     const { return _nodeList; }

  const list<Node*>& prefixList()   const { return _prefixList; }

  const TreePair& treePair()        const { return _treePair; }
  
  const vector< vector< bool > >& ancestorsBoolMatrix() const{ return _ancestorsBoolMatrix; }
  const vector< vector< uint > >& distMatrix()          const{ return _distMatrix; }

  virtual const ParamTree& paramTree() const = 0;

  // Boolean methods
  bool isASpeciesTree() const{ return _isASpeciesTree; }

  // Get Edge/Node
  const Edge* getEdge( uint edgeId ) const;
  const Node* getNode( uint nodeId ) const;

  // Tree informations
  uint nbOfNodes()  const { return _nodeVector.size(); }
  uint nbOfEdges()  const { return _edgeVector.size(); }
  uint nbOfLeaves() const { return _leafList.size(); }
  uint nbOfClades() const { return nbOfNodes() - nbOfLeaves(); }
  bool empty()      const { return nbOfNodes() == 0; }

  bool containsNode( const Node* node ) const;
  bool isBinary() const;

  // Distance methods
  uint dist( const Node* node1 , const Node* node2 ) const;
  uint dist( const Node* node1 , const Edge* edge2 ) const;
  uint dist( const Edge* edge1 , const Node* node2 ) const;
  uint dist( const Edge* edge1 , const Edge* edge2 ) const;

  // Others methods
  const Node* lcaFunction( const Node* node1 , const Node* node2 ) const;
  void restartAll_DFS_Visited();
  bool ancestorMatrixValue( const Node* node1 , const Node* node2 ) const;

  const Edge* rootEdge( Side side ) const{ return _root->edge(side); }

  // Compute the tree string in normal form
  void transformInNormalForm() const;
  string computeNormalFormString( const Reconciliation* reconciliation = 0 ) const;


  // Print methods
  void writeToOfstream( ofstream& treeStream , const string& treeFigName = "" );
  string treeTypeString() const;

  void outputLCAFunction( ostream& output ) const;
  void outputDistMatrix( ostream& output ) const;
  void outputAncestorMatrix( ostream& output ) const;
  void outputNodes( ostream& output ) const;
  void outputEdges( ostream& output ) const;

 protected:
  void computeAllCladeSet();

  // Private methods
 private:

  // Build sub functions
  void constructTreeStructure( Edge*& artificialEdge );
  void computeNode_And_EdgeVectors( Edge* artificialEdge );
  void compute_Ancestors_And_Dist_Matrices();
  void computeLCAFunctionVector();

  Node* computeLCAFct( const Node* node1 , const Node* node2 ) const;
  void computePostfixNodeIds();
  void computePrefixNodeIds();

  void computeDistMatrix();

};

#endif
