#include "TreePair.h"
#include "Reconciliation.h"
#include "StateSet.h"
#include "Cell.h"
#include "Types.h"

TreePair::TreePair( const ParamSpeciesTree& paramSpeciesTree , const ParamGeneTree& paramGeneTree , bool performsAlgo )
  :_speciesTree( *this , paramSpeciesTree ) , _diameter(0) , _geneTree( *this , paramGeneTree ){

  /////////////////
  // Define Cell Set
  _cellSet.resize( _speciesTree.nbOfNodes() + _speciesTree.nbOfEdges() );

  // Define Node Cell
  for( uint i = 0; i < _speciesTree.nodeVector().size(); i++ ){
    const Node* node = _speciesTree.nodeVector()[ i ];
    _cellSet[ cellId( node ) ] = new Cell( *this , node );
  }

  // Define Edge Cell
  for( uint i = 0; i < _speciesTree.edgeVector().size(); i++ ){
    const Edge* edge = _speciesTree.edgeVector()[i];
    _cellSet[ cellId( edge ) ] = new Cell( *this , edge );
  }


  /////////////////

  if( paramGeneTree.isNull() )
    return;

  compute_Mapping_And_SpeciesIdCount();

  if( !performsAlgo )
    return;

  // Both reconciliations will be defined latter
  _extremeRecVector.resize(2,0);

  // Define Cell Set for each node of the Gene Tree
  for( vector<Node*>::const_iterator node = _geneTree.nodeVector().begin(); node != _geneTree.nodeVector().end(); node++ )
    (**node).defineCellSet();

  vector<bool> nodeSet; // Size == 0 so that it is not considered into computeNbOfRec
  _nbOfRec = computeNbOfRec( nodeSet , _T );
  if( nbOfRec() == 0 )
    return;

  /////////////////////////////////////////
  // Temporary code in comments (2008 May 22): error in defineRandomTables for geneTree.252
  /////////////////////////////////////////
  // Define Random Tables T1, T2, and T3 for each u in V(G)
//  for( vector<Node*>::const_iterator node = _geneTree.nodeVector().begin(); node != _geneTree.nodeVector().end(); node++ )
//    (**node).defineRandomTables();

  // Define Lowest / Highest reconciliations
  defineExtremeRec( Lowest );
  defineExtremeRec( Highest );

  _diameter = distNMC( extremeRec( Lowest ) , extremeRec( Highest ) );
}

TreePair::~TreePair(){
  // Delete all cell of _cellSet
  for( vector< Cell* >::iterator cell = _cellSet.begin(); cell != _cellSet.end(); cell++ )
    delete *cell;

  // Delete extreme reconciliations
  for( vector<Reconciliation*>::iterator rec = _extremeRecVector.begin(); rec != _extremeRecVector.end(); rec++ )
    delete *rec;
}

const Node* TreePair::mappingFunction( const Node* gene ) const{
  Exception( vectorIdOutOfBound( gene->nodeId() , _mappingVector.size() ) , "TreePair::mappingFunction ---> nodeId out of bound" );

  const Node* specieNode = _mappingVector[ gene->nodeId() ];
  Exception( specieNode == 0 , "TreePair::mappingFunction ---> mapping function not initialized for node id " + itos( gene->nodeId() ) );

  return specieNode;
}


void TreePair::compute_Mapping_And_SpeciesIdCount(){
  _mappingVector.resize( _geneTree.nbOfNodes() , 0 );
  _speciesIdCount.resize( _speciesTree.nbOfLeaves() , 0 );

  vector<Node*>::const_iterator iterNode;

  for(  iterNode = _geneTree.nodeVector().begin(); iterNode != _geneTree.nodeVector().end(); iterNode++ ){
    if( (**iterNode).isLeafNode() ){
      uint specieId = ( (const LeafNode&) (**iterNode) ).specieId();
      _mappingVector[ (**iterNode).nodeId() ] = const_cast<LeafNode*> ( _speciesTree.specieId_2_LeafNode( specieId ) );
      _speciesIdCount[ specieId ]++;
    }
    else{
      const Node* specieNode1 = mappingFunction( ( (CladeNode*)(*iterNode) )->child(LEFT) );
      const Node* specieNode2 = mappingFunction( ( (CladeNode*)(*iterNode) )->child(RIGHT) );
      _mappingVector[ (**iterNode).nodeId() ] = const_cast<Node*> ( _speciesTree.lcaFunction( specieNode1 , specieNode2 ) );
    }
  }
}

// Compute the parsimonious reconciliation
// Perform a Prefix order tree traversal
void TreePair::defineExtremeRec( ExtremeType extremeType ){
  _extremeRecVector[ extremeType ] = new Reconciliation( *this );
  for( vector<Node*>::const_reverse_iterator node = _geneTree.nodeVector().rbegin(); node != _geneTree.nodeVector().rend(); node++)
    _extremeRecVector[ extremeType ]->setExtremeMapping( extremeType , *node );
}

vector<uint> TreePair::computeFirstRecVector( const OrthoParam& orthoParam ) const{
  // Define First Rec
  vector<uint> recVector;

  // Lowest
  if( orthoParam.firstRecIsLowest() ) 
    recVector = extremeRec( Lowest ).recVector();

  // Random
  if( orthoParam.firstRecIsRandom() )
    recVector = randomRec();
      
  // Given
  if( orthoParam.firstRecIsGiven() ){
    _geneTree.stringId_To_RecVector( orthoParam.paramString( FIRST_REC ) , recVector );
  }

  return recVector;
}


// Construct a Random reconciliation by a prefix traversal of G
vector<uint> TreePair::randomRec() const{
  Exception( nbOfRec() == 0 , "TreePair::randomRec() ----> nb of rec is 0 " );
  Exception( _geneTree.nodeVector().front()->T1().empty() ,"TreePair::randomRec() ----> random tables have to be defined (cf constructor)" ); 

  vector<Cell*> rec( _geneTree.nbOfNodes() , 0 );

  for( vector<Node*>::const_reverse_iterator node = _geneTree.nodeVector().rbegin(); node != _geneTree.nodeVector().rend(); node++ ){
    // u = r(G) or u in L(G)
    if( (**node).isRootNode() || (**node).isLeafNode() )
      rec[ (**node).nodeId() ] = getCell( mappingFunction( *node ) );
    else{
      Cell* fatherCell = rec[ (**node).father()->nodeId() ];
      rec[ (**node).nodeId() ] = (**node).randomCell( fatherCell );
    }
  }

  vector<uint> recVector( _geneTree.nbOfNodes() , 0 );
  for( vector<Node*>::const_iterator node = _geneTree.nodeVector().begin(); node != _geneTree.nodeVector().end(); node++ )
    recVector[ (**node).nodeId() ] = rec[ (**node).nodeId() ]->layerId();

  return recVector;
}

void TreePair::readRecVector( const string& inputFile , vector<uint>& recVector ) const{
  recVector.resize( _geneTree.nbOfNodes() , 0 );
  ifstream stream( inputFile.c_str() , std::ios::in );

  skipUntilString( stream , "Gene" );

  while( !stream.eof() ){
    string buffer;
    getline( stream  , buffer );
    vector<string> stringVector;
    partitionString( buffer , stringVector , BLANK );
    recVector[ stoi( stringVector[0] ) ] = stoi( stringVector[1] );
  }

  stream.close();
}

uint TreePair::T( const Node* gene , const Cell* cell ) const{
  return _T[ gene->nodeId() ] [ cell->id() ];
}

////////////////////////////////
// Genes family information

bool TreePair::coversAllSpecies() const{
  for( int i = 0; i < _speciesIdCount.size(); i++ )
    if( _speciesIdCount[i] == 0 )
      return false;
  return true;
}
  


////////////////////////////////
// Combinatorial algorithms
////////////////////////////////

uint TreePair::computeNbOfRec( const vector< bool >& nodeSet , vector< vector<uint> >& T ) const{
  // Define the matrix T[u,c]
  T.resize( _geneTree.nbOfNodes() , vector<uint>( _cellSet.size() , 0 ) );

  // Define the matrix T(u,c) by a postfix order of tree G
  for( vector<Node*>::const_iterator N = _geneTree.nodeVector().begin(); N != _geneTree.nodeVector().end(); N++ ){
    // Base Case
    if( (**N).isLeafNode() )
      T[ (**N).nodeId() ] [ cellId( mappingFunction( *N ) ) ] = 1;
    // Recursive case
    else{
      for( Cell* C = (**N).lowestCell(); C != 0; C = (**N).nextCell( C ) ){

// 	if( distCell( C , (**N).lowestCell() ) != 0 )
// 	  continue;

	// u \in K(G) & u \in nodeSet
	if( !nodeSet.empty() && C->isANode() && nodeSet[ (**N).nodeId() ] )
	  continue;
	
	const Node* N1 = ( (const CladeNode&) (**N) ).child( LEFT );
	const Node* N2 = ( (const CladeNode&) (**N) ).child( RIGHT );
	for( Cell* C1 = (*N1).lowestCell(); C1 != 0; C1 = (*N1).nextCell( C1 , C ) )
	  for( Cell* C2 = (*N2).lowestCell(); C2 != 0; C2 = (*N2).nextCell( C2 , C ) )
	    T[ (**N).nodeId() ] [ C->id() ] += T[ (*N1).nodeId() ] [ C1->id() ] * T[ (*N2).nodeId() ] [ C2->id() ];

	// u \in K(G) & u \notin nodeSet
	if( !nodeSet.empty() && C->isANode() && !nodeSet[ (**N).nodeId() ] )
	  break;
      }
    }
  }

  // Compute the number of rec for r(G)
  uint nbOfRec = 0;
  // If r(G) is a forced dup AND L(G) = L(S) ----> there is no reconciliation
  const CladeNode* rootG = _geneTree.root();
  if( !( rootG->isAForcedDup() && coversAllSpecies() ) )
    for( Cell* C = rootG->lowestCell(); C != 0; C = rootG->nextCell( C ) )
      nbOfRec += T[ rootG->nodeId() ] [C->id()];

  return nbOfRec;
}


uint TreePair::computeNbOfSubDupOpt() const{
  vector< bool > nodeSet( _geneTree.nbOfNodes() , false );
  vector< vector<uint> > T;
  uint nbOfRec = computeNbOfRec( nodeSet , T );
  return nbOfRec;
}

void TreePair::generateConsistantGeneTree( const string& originRealRecString , string& geneTreeString , string& realRecString ) const{
  Node* species = const_cast<Node*> ( mappingFunction( geneTree().root() ) );
  Exception( species == speciesTree().root() , "TreePair::generateConsistantGeneTree ---> M( r(G) ) = r(S) " );

  // Compute the child of r(S) that is the ancestor of Species
  while( !species->father()->isRootNode() ){
    species = const_cast<CladeNode*> ( species->father() );
  }

  // Any speciesId of the sibling is not in G
  string speciesName = _speciesTree.paramTree().id2SpeciesName( itos( *( species->sibling()->cladeSet().begin() ) ) );


  //////////////////////
  // Define geneTreeString
  geneTreeString  = "(";
  geneTreeString += _geneTree.paramTree().treeString1();

  // Erase ";"
  geneTreeString.erase( geneTreeString.size() - 1 );

  geneTreeString += ",";
  geneTreeString +=  speciesName;
  geneTreeString += "*);";

  //////////////////////
  // Define realRecString
  realRecString  = "[" + itos( speciesTree().root()->nodeId() ) + "]";
  realRecString += "(" + originRealRecString;

  // Erase ";"
  realRecString.erase( realRecString.size() - 1 );

  realRecString += "," + speciesName    + ");";
}

//////////////////////////////////////////////
// Output methods

std::ostream& operator << (std::ostream& outputStream, const TreePair& treePair){
  outputStream<< "\n\nSpecies tree\n" << treePair.speciesTree();//.treeString();
  outputStream<< "\n\nGenes tree\n"  << treePair.geneTree();//.treeString();

  outputStream<< "\nCovers all species: " << treePair.coversAllSpecies();

  outputStream<<"\nNb of reconciliatons: " << treePair.nbOfRec();
//     outputStream<<"\n\nLowest reconciliations: " << treePair.extremeRec( Lowest );
//     outputStream<<"\nHighest reconciliations: " << treePair.extremeRec( Highest );
  if( treePair.nbOfRec() != 0 )
    outputStream<<"\nSpace diameter: " << treePair.diameter();


  //  treePair.printCellAncestorship( outputStream );

//   treePair.print_Mapping_Vector( outputStream );
//   treePair.printTMatrix( outputStream );

  return outputStream;
}

// Reading methods
const Cell* TreePair::retrieveCell( const Node* geneNode , const string& cellStringId ) const{
  Exception( _cellSet.empty() , "TreePair::retrieveCell ---> cellSet is empty" );

  // Species Edge
  if( cellStringId[0] == EDGE_TAG ){
    Exception( geneNode->isLeafNode() , "TreePair::retrieveCell ---> geneNode is a Leaf node" );
    uint i = 1;
    const Cell* nodeCell = _cellSet[ readTaxaId( cellStringId , i ) ];
    return ( (const CladeNode*) (geneNode) )->getChildEdgeCell( nodeCell );
  }

  // Species node case
  uint i = 0;
  uint taxaId = readTaxaId( cellStringId , i );
  
  return _cellSet[ taxaId ];
}

void TreePair::print_Mapping_Vector( std::ostream& outputStream ) const{
  outputStream<<"\n\nMapping function";
  outputStream<<"\nGenes node\tSpecies\n";

  uint geneId;
  for( geneId = 0; geneId < _mappingVector.size(); geneId++ )
    outputStream << geneTree().getNode( geneId )->nodeId() <<"\t\t" << _mappingVector[ geneId ]->nodeId() <<endl;
}

void TreePair::printTMatrix( std::ostream& outputStream ) const{
  for( vector<Node*>::const_iterator node = _geneTree.nodeVector().begin(); node != _geneTree.nodeVector().end(); node++ )
    for( uint c = 0; c < _cellSet.size(); c++ ){
      uint t = _T[ (**node).nodeId() ] [ c ];
      if( t != 0 )
	cerr<< "\n( " << (**node).nodeId() << " , " << *getCell( c ) << " ) :" << t;
    }
  
}

void TreePair::printCellAncestorship( ostream& output ) const{
  output<<"\nCell ancestorship\n";

  for( vector< Cell* >::const_iterator cell1 = _cellSet.begin(); cell1 != _cellSet.end(); cell1++ ){
    output<< "\n\n" << **cell1 << " : ";
    for( vector< Cell* >::const_iterator cell2 = _cellSet.begin(); cell2 != _cellSet.end(); cell2++ )
      if( **cell1 <= **cell2 )
	output<< **cell2 << " ";
    }
}
