/***************************************************************************
			TreePair.h
		-------------------
	begin	 : Fri Jun  9 14:52:31 EDT 2006
	copyright: (C) 2005 by Universite De Montreal
	email	 :  doyonjea@iro.umontreal.ca
***************************************************************************/
#ifndef TREEPAIR_H
#define TREEPAIR_H

//! This class defines the base structure for a TreePair object.
/**
*@author Jean-Philippe Doyon
*/

#include "SpeciesTree.h"
#include "GeneTree.h"
#include "ParamGeneTree.h"
#include "ParamSpeciesTree.h"
#include "OrthoParam.h"

class Reconciliation;
class StateSet;

// For Lowest and Highest reconciliations
enum ExtremeType{Lowest,Highest};

const uint POURC_STEP = 1;
// For the monitoring
const string LOG_FILE = "exploration.log";

class TreePair{

  friend std::ostream& operator << (std::ostream& outputStream, const TreePair& treePair);

 private:

  const SpeciesTree _speciesTree;
  const GeneTree _geneTree;

  // M: V(G) --> V(S)
  // Function Mapping: vector[ gene node Id ] ---> species node
  vector<Node*> _mappingVector;

  // Set of all cell (Node/Edge of the species tree)
  // Bijection: Cell id <----> cell
  // A node : i in [0 ... n - 1]
  // An edge: i in [n ... n + m - 1]
  // n is |V(S)| and m is |V(E)|
  vector< Cell* > _cellSet;

  // Matrix T(u,c) s.t.
  //    - u \in V(G) and c \in V(S) + E(S)
  //    - T[u,c] = 0
  vector< vector<uint> > _T;

  // Matrix: Species id ----> nb of leaves that have this id
  vector<uint> _speciesIdCount;

  // Contains Minima and maxima reconciliations
  vector< Reconciliation* > _extremeRecVector;

  // Attributes of the reconciliation space
  uint _nbOfRec;
  uint _diameter;

 public:

  TreePair( const ParamSpeciesTree& paramSpeciesTree , const ParamGeneTree& paramGeneTree , bool performsAlgo = true );

  ~TreePair();

  const SpeciesTree& speciesTree() const { return _speciesTree; }
  const GeneTree& geneTree()          const { return _geneTree; }

  const Node* mappingFunction( const Node* gene ) const;

  uint T( const Node* gene , const Cell* cell ) const;

  // Node/Edge ----> cell id
  uint cellId( const Node* node ) const{ return node->nodeId(); }
  uint cellId( const Edge* edge ) const{ return _speciesTree.nbOfNodes() + edge->edgeId(); }

  // Node/Edge ----> Cell
  Cell* getCell( const Node* node ) const { return _cellSet[ cellId( node ) ]; }
  Cell* getCell( const Edge* edge ) const { return _cellSet[ cellId( edge ) ]; }

  Cell* getCell( uint cellId ) const { return _cellSet[ cellId ]; }

  const vector< Cell* >& cellSet() const { return _cellSet; }

  const vector<uint>& speciesIdCount() const { return _speciesIdCount; }

  uint nbOfRec()  const { return _nbOfRec; }
  uint diameter() const { return _diameter; }

  ////////////////////////////////
  // Genes family information
  bool coversAllSpecies() const;

  ////////////////////////////////
  // Combinatorial algorithms
  ////////////////////////////////

  // Count the nb. of reconciliations
  // nodeSet \subseteq K(G)
  // Consider only the rec s.t. u \in nodeSet --> rec(u) \neq M(u)
  uint computeNbOfRec( const vector< bool >& nodeSet , vector< vector<uint> >& T ) const;

  // Compute the number of rec s.t. #dup(rec) == #dup(recmin) + delta
  // For the moment, delta == 0
  uint computeNbOfSubDupOpt() const;

  // Return a randomize reconciliation in the recVector format
  vector<uint> randomRec() const;

  // Either minima or maxima reconciliations
  void defineExtremeRec( ExtremeType extremeType );
  const Reconciliation& extremeRec( ExtremeType extremeType ) const{
    Exception( nbOfRec() == 0 , "TreePair::randomRec() ----> nb of rec is 0 " );
    return *( _extremeRecVector[ extremeType ] );
  }

  vector<uint> computeFirstRecVector( const OrthoParam& orthoParam ) const;

  // Read recVector from InputFile
  void readRecVector( const string& inputFile , vector<uint>& recVector ) const;

  // Reading methods
  const Cell* retrieveCell( const Node* geneNode , const string& cellStringId ) const;

  // Consistency between G and S: M( r(G) ) = r(S)
  // If not, insert in G a fictive extant gene
  bool consistantRootMapping() const{
    return mappingFunction( geneTree().root() ) == speciesTree().root();
  }

  void generateConsistantGeneTree( const string& originRealRecString , string& geneTreeString , string& realRecString ) const;

  // Output methods
  void printCellAncestorship( ostream& output ) const;

 private:

  // Algorithm method
  void compute_Mapping_And_SpeciesIdCount();

  // Debug function
  void print_Mapping_Vector( std::ostream& outputStream ) const;
  void printTMatrix( std::ostream& outputStream ) const;

};

#endif
