#include "Gene.h"
#include "State.h"
#include "SpeciesEdge.h"
#include "CladeSpecies.h"
#include "GeneEdge.h"
#include "ArtifGene.h"

int Gene::_classCounter = 0;

Gene::Gene( const State& state ):
  _state( state ) {
  _classCounter++;
  _species   = 0;
  _leftSlicedSubtree  = 0;
  _rightSlicedSubtree = 0;
  _treeData = 0;
}

Gene::~Gene(){
  _classCounter--;
  
  // The left/right sliced subtrees are deleted either in ~State or in State::performSlide
  _leftSlicedSubtree = 0;
  _rightSlicedSubtree = 0;
  
  if( _treeData != 0 ){
    delete _treeData;
    _treeData = 0;
  }
}

const SlicedSubtree& Gene::getSlicedSubtree( Side side ) const{
  Exception( _leftSlicedSubtree == 0 || _rightSlicedSubtree == 0 , "Gene::getSlicedSubtree ---> sliced subtree not constructed: " + geneString() );
  const SlicedSubtree& slicedSubtree = ( side==LEFT? *_leftSlicedSubtree: *_rightSlicedSubtree );
  return slicedSubtree;
}

const TreeData& Gene::treeData() const{
  Exception( _treeData == 0 , "Gene::treeData ---> treeData is not defined: " + geneString() );
  return *_treeData;
}

void Gene::setCovering( const Species* species ){
  Exception( species == 0 , "Gene::setCovering ---> species == 0" );
  _species = species;
  const_cast<Species*> (_species)->addCoveredGene( this );
}

bool Gene::isCoveredByASpecies() const{
  return _species != 0;
}

bool Gene::isCoveredBy( const Species* species ) const{
  return isCoveredByASpecies() && (*_species) == (*species);
}


const Species* Gene::coveringSpecies() const{
  Exception( _species == 0 , " :Gene::coveringSpecies  ---> _species == 0" );
  return _species;
}

const SpeciesEdge* Gene::coveringSpeciesEdgeSide( Side side ) const{
  Exception( !isCoveredByASpecies() , "Gene::coveringSpeciesEdge ---> this is not covered by a species node" );
  Exception( isLeafNode()           , "Gene::coveringSpeciesEdge ---> this is a leaf gene" );
  return ( (CladeSpecies*) (_species) )->edge(side);
}


void Gene::buildSlicedSubtrees( bool empty ){
  Exception( _leftSlicedSubtree  != 0 , "Gene::deleteSlicedSubtrees --->  left =0 " );
  Exception( _rightSlicedSubtree != 0 , "Gene::deleteSlicedSubtrees ---> right =0 " );

  _leftSlicedSubtree  = new SlicedSubtree( coveringSpeciesEdgeSide( LEFT )  , this , empty );
  _rightSlicedSubtree = new SlicedSubtree( coveringSpeciesEdgeSide( RIGHT ) , this , empty );
}

void Gene::deleteSlicedSubtrees(){
  Exception( _leftSlicedSubtree == 0 , "Gene::deleteSlicedSubtrees --->  left =0 " );
  Exception( _rightSlicedSubtree == 0 , "Gene::deleteSlicedSubtrees ---> right =0 " );
  delete _leftSlicedSubtree;
  delete _rightSlicedSubtree;

  _leftSlicedSubtree  = 0;
  _rightSlicedSubtree = 0;
}


double Gene::compute_eV( bool debug ){
  Exception( !isCoveredByASpecies() , "Gene::compute_eV ---> this is not covered by a species node" );

  double eV;

  if( isLeafNode() )
    _eV = 0;
  else{
    eV  = compute_eA( LEFT , debug ) + compute_eA( RIGHT , debug );
    if( !debug )
      _eV = eV;
  }

  return eV;
}

void Gene::define_eV( double eAL , double eAR ){
  _eAL = eAL;
  _eAR = eAR;
  _eV = eAL + eAR;
}

void Gene::debug_eA( Side side , double new_eA ) const{
  string S = "\nGene::debug_eA ---> eA differ:";
  S += "\n\tSide:" + sideToString( side );
  S += "\n\tEdge:" + ( (const CladeSpecies*) ( coveringSpecies() ) )->edge( side )->toString();
  S += "\n\tOld: " + dtos( eA(side) );
  S += "\n\tNew: " + dtos( new_eA );
  Exception( eA(side) != new_eA , S );
 }

double Gene::compute_eA( Side side , bool debug ){

  const SpeciesEdge*   speciesEdge = coveringSpeciesEdgeSide( side );
  SlicedSubtree& slicedSubtree     = const_cast<SlicedSubtree&> ( getSlicedSubtree( side ) );

  double eA;

  // Case: current gene u = r(G)
  //       child u' of u (side) is fictive
  // ( Cf LeafNode.h )
  if( !isArtifNode() && ( (RealGene*) (this) )->isRootNode() && 
      ( (RealGene*) (this) )->retrieveChild( speciesEdge )->fictiveLeaf() ){
    eA = speciesEdge->P_Fun( 0 );
  }

  else{
    eA = speciesEdge->P_Fun( slicedSubtree.nbOfLeaves() );

    if( slicedSubtree.nbOfLeaves() != 0 ){
      eA += slicedSubtree.H_fun( true );
      eA += slicedSubtree.Phi_fun( true );

      list<Gene*>::const_iterator iterNode;
      for( iterNode = slicedSubtree.leafNodeList().begin(); iterNode != slicedSubtree.leafNodeList().end(); iterNode++ )
	eA += (**iterNode ).eV();
    }
  }

  if( debug )
    debug_eA( side , eA );
  else{
    if( side == LEFT )
      _eAL = eA;
    else
      _eAR = eA;
  }

  return eA;
}

void Gene::setTreeData( uint classId ){
//  Exception( _treeData != 0 , "Gene::setTreeData ---> treeData has already been defined: " + geneString() );
  if( _treeData != 0 )
    delete _treeData;
  _treeData = new TreeData( classId );
}

void Gene::setTreeData( const TreeData& treeData1 , const TreeData& treeData2 ){
//  Exception( _treeData != 0 , "Gene::setTreeData ---> treeData has already been defined: " + geneString() );
  if( _treeData != 0 )
    delete _treeData;
  _treeData = new TreeData( treeData1 , treeData2 );
}

void Gene::unsetTreeData(){
//  Exception( _treeData == 0 , "Gene::unsetTreeData ---> treeData has not been defined: " + geneString() );

  if( _treeData != 0 )
    delete _treeData;

  _treeData = 0;
}


std::ostream& operator << (std::ostream& outputStream, const Gene& gene ){
  if( gene.isArtifNode() )
    outputStream << (const ArtifGene&) (gene);
  else
    outputStream << (const RealGene&) (gene);

  return outputStream;
}
