/***************************************************************************
			Gene.h
		-------------------
	begin	 : Tue Jan  9 11:40:17 EST 2007
	copyright: (C) 2005 by Universite De Montreal
	email	 :  doyonjea@iro.umontreal.ca
***************************************************************************/
#ifndef GENENODE_H
#define GENENODE_H

//! This class defines the base structure for a Gene object.
/**
*@author Jean-Philippe Doyon
*/

// This is an ABSTRACT CLASS

#include "StdClass.h"
#include "TreeData.h"
#include "StringInfo.h"

class State;
class SpeciesEdge;
class Species;
class GeneEdge;
class SlicedSubtree;



class Gene{

  friend std::ostream& operator << (std::ostream& outputStream, const Gene& gene );

 private:

  static int _classCounter;

  double _eV, _eAL, _eAR;

  // Only for genes covered by a species
  StringInfo* _stringInfo;

 protected:
  // Reconciliation attributes: 
  // This Gene is either reconcilied to a SpeciesNode or a SpeciesEdge
  // SpeciesEdge attribute is in RealGene only 
  const Species* _species;

  const State& _state; 

  SlicedSubtree* _fatherSlicedSubtree;
  SlicedSubtree* _leftSlicedSubtree;
  SlicedSubtree* _rightSlicedSubtree;

  uint _nbOfLeaves;
  uint _betaFct;

  // Only for Gene u reconcilied on species edge.
  // The canonical representation of the sliced subtree rooted at gene u
  TreeData* _treeData;

  // Set <---> Gene \in L(T), for a Sliced Subtree T
  list<Gene*>::iterator _leafIterator;

  // The iterator in the _prefixOrderList of the Sliced Subtree
  list< const Gene* >::iterator _orderIterator;

 public:

  static int classCounter(){ return _classCounter; };

  // Constructor and Destructor
  Gene( const State& state );
  virtual ~Gene();

  virtual void clearForRedefine(){
    _species = 0;
    _leftSlicedSubtree = _rightSlicedSubtree = _fatherSlicedSubtree = 0;

    if( _treeData != 0 ) 
      delete _treeData;
    _treeData = 0;
  }

  // Access methods (Class attributes)
  const State& state() const { return _state; }
  virtual uint nodeId() const = 0;

  const TreeData& treeData() const;

  void setStringInfo( StringInfo* stringInfo ){ _stringInfo = stringInfo; }
  StringInfo* stringInfo() const{ return _stringInfo; }

  // Sliced subtree methods
  const SlicedSubtree& getSlicedSubtree( Side side ) const;
  void setFatherSlicedSubtree( SlicedSubtree* tree ) { _fatherSlicedSubtree = tree; }
  const SlicedSubtree* fatherSlicedSubtree() const   { return _fatherSlicedSubtree; }

  // Empty is used in two cases:
  //   Case 1: ArtifGene: spec2Dup, etc.
  //   Case 2: RealGene: in dup2Spec 
  //           - when both sliced subtrees had not been previously constructed
  //           - when the gene is not covered by a species in the first rec (most parsimonious one)
  void buildSlicedSubtrees( bool empty );

  void deleteSlicedSubtrees();

  bool childrenSlicedSubtree_AreDefined() const{ return _leftSlicedSubtree != 0; }


  // Node type methods (Pure virtual)
  virtual bool isLeafNode()  const  = 0;
  virtual bool isArtifNode() const  = 0;

  // DFS visited status
  virtual bool dfsVisited() const = 0;
  virtual void setDFS_Visited(bool b) = 0;
  virtual bool fictiveLeaf()    const = 0;

  // Reconciliation methods
  void setCovering( const Species* species );
  bool isCoveredByASpecies() const;
  bool isCoveredBy( const Species* species ) const;
  const Species* coveringSpecies() const;

  const SpeciesEdge* coveringSpeciesEdgeSide( Side side ) const;

  // Sliced subtree methods 
  virtual const Gene* slicedFather() const  = 0;

  virtual const Gene* nextDFS_SlicedChild( const SlicedSubtree& slicedSubtree ) const = 0;
  virtual bool isFirstSlicedChild        ( const SlicedSubtree& slicedSubtree  , const Gene* childGene ) const = 0;
  
  // Likelihood methods
  double compute_eV( bool debug = false );
  void define_eV( double eAL , double eAR );
  double eV() const { return _eV; }

  double compute_eA( Side side , bool debug = false );
  double eA(Side side) const { return ( side==LEFT? _eAL: _eAR ); }



  // Recompute eV by recomputing eA for the specified side
  void set_eA( Side side , double eA );

  // Methods for the computation of H and Phi functions (see Phd report)
  void setTreeData( uint classId );
  void setTreeData( const TreeData& treeData1 , const TreeData& treeData2 );
  void unsetTreeData();

  uint nbOfLeaves() const { return _nbOfLeaves; }
  uint betaFct() const    { return _betaFct; }

  // Debug methods
  virtual void checkNode( std::ostream& output ) const = 0; // Verify that all node attributes are OK

  // Print methods
  virtual string geneString() const = 0;

 private:
  void debug_eA( Side side , double new_eA ) const;

};

#endif
